/******************************************************************************
 *             D A T A    E N C R Y P T I O N    S T A N D A R D
 *-----------------------------------------------------------------------------
 *   INRIA - Unite de Recherche Rhone-Alpes
 *   655, avenue de l'Europe
 *   38330 Montbonnot Saint Martin
 *   FRANCE
 *-----------------------------------------------------------------------------
 *   Module		:	gate_functions.c
 *   Auteur		:	Wendelin SERWE
 *   Version		:	1.9
 *   Date		: 	06/03/01 14:03:16
 *****************************************************************************/

/*
 * Gate functions for the DES prototype generated using  EXEC/CAESAR
 * 
 * The DES prototype repeatedly reads the inputs on gates CRYPT, DATA, and KEY
 * from file stdin, and outputs the (de)ciphered value via gate OUTPUT to
 * file stdout.  The emulator also logs its execution trace to file
 * CAESAR_LOG_FILE.
 */

/* ------------------------------------------------------------------------- */

#include <ctype.h>		/* for tolower() */

#include "caesar_kernel.h"

/* ------------------------------------------------------------------------- */

#define CAESAR_ADT_INTERFACE
#include "full_des.h"

/* ------------------------------------------------------------------------- */

/* auxiliary macro to access a single bit of a 64-bit word */

#define CAESAR_GET_BIT(CAESAR_WORD,CAESAR_BIT) \
	(((CAESAR_WORD) & (((unsigned long long int) 1) << (CAESAR_BIT))) ? BIT_ONE () : BIT_ZERO ())

/* ------------------------------------------------------------------------- */

/* auxiliary function to input a 64-bit vector */

void READ_64_BIT_VECTOR (CAESAR_DATA)
ADT_BIT64 *CAESAR_DATA;
{
     unsigned long long int CAESAR_WORD;

     fscanf (stdin, " %llx", &CAESAR_WORD);

     *CAESAR_DATA = MK_64 (
	   CAESAR_GET_BIT (CAESAR_WORD, 0), CAESAR_GET_BIT (CAESAR_WORD, 1),
	   CAESAR_GET_BIT (CAESAR_WORD, 2), CAESAR_GET_BIT (CAESAR_WORD, 3),
	   CAESAR_GET_BIT (CAESAR_WORD, 4), CAESAR_GET_BIT (CAESAR_WORD, 5),
	   CAESAR_GET_BIT (CAESAR_WORD, 6), CAESAR_GET_BIT (CAESAR_WORD, 7),
	   CAESAR_GET_BIT (CAESAR_WORD, 8), CAESAR_GET_BIT (CAESAR_WORD, 9),
	 CAESAR_GET_BIT (CAESAR_WORD, 10), CAESAR_GET_BIT (CAESAR_WORD, 11),
	 CAESAR_GET_BIT (CAESAR_WORD, 12), CAESAR_GET_BIT (CAESAR_WORD, 13),
	 CAESAR_GET_BIT (CAESAR_WORD, 14), CAESAR_GET_BIT (CAESAR_WORD, 15),
	 CAESAR_GET_BIT (CAESAR_WORD, 16), CAESAR_GET_BIT (CAESAR_WORD, 17),
	 CAESAR_GET_BIT (CAESAR_WORD, 18), CAESAR_GET_BIT (CAESAR_WORD, 19),
	 CAESAR_GET_BIT (CAESAR_WORD, 20), CAESAR_GET_BIT (CAESAR_WORD, 21),
	 CAESAR_GET_BIT (CAESAR_WORD, 22), CAESAR_GET_BIT (CAESAR_WORD, 23),
	 CAESAR_GET_BIT (CAESAR_WORD, 24), CAESAR_GET_BIT (CAESAR_WORD, 25),
	 CAESAR_GET_BIT (CAESAR_WORD, 26), CAESAR_GET_BIT (CAESAR_WORD, 27),
	 CAESAR_GET_BIT (CAESAR_WORD, 28), CAESAR_GET_BIT (CAESAR_WORD, 29),
	 CAESAR_GET_BIT (CAESAR_WORD, 30), CAESAR_GET_BIT (CAESAR_WORD, 31),
	 CAESAR_GET_BIT (CAESAR_WORD, 32), CAESAR_GET_BIT (CAESAR_WORD, 33),
	 CAESAR_GET_BIT (CAESAR_WORD, 34), CAESAR_GET_BIT (CAESAR_WORD, 35),
	 CAESAR_GET_BIT (CAESAR_WORD, 36), CAESAR_GET_BIT (CAESAR_WORD, 37),
	 CAESAR_GET_BIT (CAESAR_WORD, 38), CAESAR_GET_BIT (CAESAR_WORD, 39),
	 CAESAR_GET_BIT (CAESAR_WORD, 40), CAESAR_GET_BIT (CAESAR_WORD, 41),
	 CAESAR_GET_BIT (CAESAR_WORD, 42), CAESAR_GET_BIT (CAESAR_WORD, 43),
	 CAESAR_GET_BIT (CAESAR_WORD, 44), CAESAR_GET_BIT (CAESAR_WORD, 45),
	 CAESAR_GET_BIT (CAESAR_WORD, 46), CAESAR_GET_BIT (CAESAR_WORD, 47),
	 CAESAR_GET_BIT (CAESAR_WORD, 48), CAESAR_GET_BIT (CAESAR_WORD, 49),
	 CAESAR_GET_BIT (CAESAR_WORD, 50), CAESAR_GET_BIT (CAESAR_WORD, 51),
	 CAESAR_GET_BIT (CAESAR_WORD, 52), CAESAR_GET_BIT (CAESAR_WORD, 53),
	 CAESAR_GET_BIT (CAESAR_WORD, 54), CAESAR_GET_BIT (CAESAR_WORD, 55),
	 CAESAR_GET_BIT (CAESAR_WORD, 56), CAESAR_GET_BIT (CAESAR_WORD, 57),
	 CAESAR_GET_BIT (CAESAR_WORD, 58), CAESAR_GET_BIT (CAESAR_WORD, 59),
	 CAESAR_GET_BIT (CAESAR_WORD, 60), CAESAR_GET_BIT (CAESAR_WORD, 61),
	CAESAR_GET_BIT (CAESAR_WORD, 62), CAESAR_GET_BIT (CAESAR_WORD, 63));
}

/*---------------------------------------------------------------------------*/

int  CRYPT (CAESAR_MODE_0, CAESAR_TYPE_0, CAESAR_DATA_0, CAESAR_MODE_1)
CAESAR_KERNEL_OFFER CAESAR_MODE_0;
char *CAESAR_TYPE_0;
ADT_BOOL *CAESAR_DATA_0;
CAESAR_KERNEL_OFFER CAESAR_MODE_1;
{
     char CAESAR_C;

     CAESAR_KERNEL_ASSERT_INPUT (CAESAR_MODE_0);
     CAESAR_KERNEL_ASSERT_TYPE (CAESAR_TYPE_0, "ADT_BOOL");
     CAESAR_KERNEL_ASSERT_EOL (CAESAR_MODE_1);

     /* input CAESAR_DATA_0 */
     do {
	  fprintf (stdout, "ciphering or deciphering (c/d)? ");
	  fscanf (stdin, " %c", &CAESAR_C);
	  CAESAR_C = (char) tolower (CAESAR_C);
     } while ((CAESAR_C != 'c') && (CAESAR_C != 'd'));
     if (CAESAR_C == 'c')
	  *CAESAR_DATA_0 = ADT_TRUE ();
     else
	  *CAESAR_DATA_0 = ADT_FALSE ();

     CAESAR_KERNEL_LOG_GATE (__func__);
     /* LINTED */
     CAESAR_KERNEL_LOG_OFFER (CAESAR_MODE_0, CAESAR_TYPE_0, *CAESAR_DATA_0, ADT_PRINT_BOOL, 1);
     CAESAR_KERNEL_LOG_RESULT (1);

     return (1);
}

/* ------------------------------------------------------------------------- */

int DATA (CAESAR_MODE_0, CAESAR_TYPE_0, CAESAR_DATA_0, CAESAR_MODE_1)
CAESAR_KERNEL_OFFER CAESAR_MODE_0;
char *CAESAR_TYPE_0;
ADT_BIT64 *CAESAR_DATA_0;
CAESAR_KERNEL_OFFER CAESAR_MODE_1;
{
     CAESAR_KERNEL_ASSERT_INPUT (CAESAR_MODE_0);
     CAESAR_KERNEL_ASSERT_TYPE (CAESAR_TYPE_0, "ADT_BIT64");
     CAESAR_KERNEL_ASSERT_EOL (CAESAR_MODE_1);

     /* input CAESAR_DATA_0 */
     fprintf (stdout, "input data? ");
     READ_64_BIT_VECTOR (CAESAR_DATA_0);

     CAESAR_KERNEL_LOG_GATE (__func__);
     /* LINTED */
     CAESAR_KERNEL_LOG_OFFER (CAESAR_MODE_0, CAESAR_TYPE_0, *CAESAR_DATA_0, ADT_PRINT_BIT64, 1);
     CAESAR_KERNEL_LOG_RESULT (1);

     return (1);
}

/* ------------------------------------------------------------------------- */

int KEY (CAESAR_MODE_0, CAESAR_TYPE_0, CAESAR_DATA_0, CAESAR_MODE_1)
CAESAR_KERNEL_OFFER CAESAR_MODE_0;
char *CAESAR_TYPE_0;
ADT_BIT64 *CAESAR_DATA_0;
CAESAR_KERNEL_OFFER CAESAR_MODE_1;
{
     CAESAR_KERNEL_ASSERT_INPUT (CAESAR_MODE_0);
     CAESAR_KERNEL_ASSERT_TYPE (CAESAR_TYPE_0, "ADT_BIT64");
     CAESAR_KERNEL_ASSERT_EOL (CAESAR_MODE_1);

     /* input CAESAR_DATA_0 */
     fprintf (stdout, "DES key? ");
     READ_64_BIT_VECTOR (CAESAR_DATA_0);

     CAESAR_KERNEL_LOG_GATE (__func__);
     /* LINTED */
     CAESAR_KERNEL_LOG_OFFER (CAESAR_MODE_0, CAESAR_TYPE_0, *CAESAR_DATA_0, ADT_PRINT_BIT64, 1);
     CAESAR_KERNEL_LOG_RESULT (1);

     return (1);
}

/* ------------------------------------------------------------------------- */

int OUTPUT (CAESAR_MODE_0, CAESAR_TYPE_0, CAESAR_DATA_0, CAESAR_MODE_1)
CAESAR_KERNEL_OFFER CAESAR_MODE_0;
char *CAESAR_TYPE_0;
ADT_BIT64 CAESAR_DATA_0;
CAESAR_KERNEL_OFFER CAESAR_MODE_1;
{
     CAESAR_KERNEL_ASSERT_OUTPUT (CAESAR_MODE_0);
     CAESAR_KERNEL_ASSERT_TYPE (CAESAR_TYPE_0, "ADT_BIT64");
     CAESAR_KERNEL_ASSERT_EOL (CAESAR_MODE_1);

     /* output CAESAR_DATA_0 */
     fprintf (stdout, "output data = ");
     ADT_PRINT_BIT64 (stdout, CAESAR_DATA_0);
     fprintf (stdout, "\n");

     CAESAR_KERNEL_LOG_GATE (__func__);
     /* LINTED */
     CAESAR_KERNEL_LOG_OFFER (CAESAR_MODE_0, CAESAR_TYPE_0, CAESAR_DATA_0, ADT_PRINT_BIT64, 1);
     CAESAR_KERNEL_LOG_RESULT (1);

     return (1);
}

/* ------------------------------------------------------------------------- */

