from collections import deque
import itertools
import copy

# State space generator, input is network of LTSs, by Anton Wijs, 2012
# 2016: counter-example generation added

#user-defined class to pass state space search data to/from generateLTS-function in graphfunctions module
class Searchdata:
	def __init__(self, hder, trns, hasht):
		self.h = hder
		self.t = trns
		self.ht = hasht

hashtable = {}
systemstatecounter = 0
systemtranscounter = 0
keep_quiet = False
outputobject = ""

def actionStripParams(action):
	"""Strip the parameters and quotes off the action string"""
	# We strip both LOTOS style and mCRL/mCRL2 style, i.e. everything following
	# '(' and ' !', respectively
	l1 = action.split('!')
	l2 = l1[0].split('#')
	l1 = l2[0].split('(')
	#newa = l1[0].rstrip(' ')
	newa = l1[0].rstrip('"')
	newa = newa.lstrip('"')
	return newa

def actionGetParams(action):
	"""Strip the action off the params"""
	params = ''
	l1 = action.split('!')
	if len(l1) > 1:
		for i in range(1,len(l1)):
			params += '!' + l1[i]
	else:
		l1 = action.split('#')
		if len(l1) > 1:
			params = '#' + l1[1]
		else:
			l1 = action.split('(')
			if len(l1) > 1:
				for i in range(1,len(l1)):
					params += '(' + l1[i]
	params = params.rstrip('"')
	return params

def writeAut(header, trans, autfile, writetrans):
	"""Write the trans LTS to file autfile"""
	outfile = open(autfile + ".aut",'w')

	#print header of .aut file
	outfile.write('des(' + header[0] + ', ' + header[1] + ', ' + header[2] + ')\n')
	if writetrans:
		for i in range(0,int(header[2])):
			outgoing = trans.get(str(i))
			if outgoing != None:
				for k, v in outgoing.items():
					for t in v:
						outfile.write('(' + str(i) + ', ' + k + ', ' + t + ')\n')
	outfile.close()

def writeMap(filebase):
	"""Write the mapping between system state indices and process state vectors, as established in the hashtable"""
	global hashtable
	outfile = open(filebase + ".map", 'w')
#	pickle.dump(hashtable, outfile)
#	print hashtable
	# print content of hashtable, in reversed order
	for v, k in hashtable.items():
		outfile.write(k + ', ' + v + '\n')
	outfile.close()

def getHash(current):
	hash = ''
	for i in range(0,len(current)):
		hash += ',' + current[i]
	return hash

def getIndex(current):
	"""Get the state index from the hash table (given state vector is a list!)"""
	global hashtable, systemstatecounter, keep_quiet, outputobject
	tup = getHash(current)
	#print current
	#tup = current
	index = hashtable.get(tup)
	if index == None:
		# new state
		index = str(systemstatecounter)
		hashtable[tup] = index
		#print "new state " + str(systemstatecounter) + " " + str(current)
		systemstatecounter = systemstatecounter + 1
		if not keep_quiet:
			if systemstatecounter % 10000 == 0:
				if outputobject == "":
					print("generated " + str(systemstatecounter) + " states")
				else:
					outputobject.log("generated " + str(systemstatecounter) + " states\n")
		return ['1', index]
	return ['0', index]

def getOutgoing(current, procs, scope):
	"""Get all outgoing transitions of the procstates in scope in current"""
	trans = []

	# j points to current index in scope
	j = 0
	nextinscope = scope[j]
	for i in range(0,len(current)):
		if i == nextinscope:
			j += 1
			if j < len(scope):
				nextinscope = scope[j]
			else:
				nextinscope = len(procs)
			out = (procs[i]).get(str(current[i]),{})
		else:
			out = {}
		trans.append(out)
	return trans

#def getSyncrules(a, i, rules):
#	"""Produce the rule vectors applicable for a"""
#	vect = []
#	ak = 'a'
#	v = []
#	for ak, v in (rules[i]).items():
#		if actionMatches(a, ak):
#			vect += v
#	return vect

def getSyncrules(a, i, rules, get_previous):
	"""Produce the rule vectors applicable for a"""
	vect = []
	if get_previous:
		for j in range(0,i):
			vect += (rules[j]).get(a,[])
	vect += (rules[i]).get(a,[])
	return vect

def is_accepting_state(s, outgoing, ids):
	"""check if given state vector satisfies criteria (i.e. all processes in ids have an outgoing accept-transition"""
	for id in ids:
		if outgoing[id].get('"accept"',set([])) == set([]):
			return False
	return True

#def actionMatches(action, ruleact):
#	"""Checks whether the action is matched on by the rule action"""
#	result = action.find(ruleact, 0, len(ruleact)+1)
#	if result != -1:
#		return True
#	else:
#		return False
#
#def actionRename(action, oldact, newact):
#	"""Renames the prefix of action from oldact to newact"""
#	return action.replace(oldact, newact, 1)

# scope: which processes should be taken into account when generating (list of indices)
# check_for_accept: for which processes should we check whether their state has an accept selfloop? If all those process states have one, the state vector is considered to be accepting (set of indices)
# report_cmc: for compositional model checking, if an accepting state vector was not encountered, report which part
# of the trace at position 'report_cmc' in the network has been explored (index, -1 if not used)
def generateLTS(procsheaders, procs, procsactsets, syncrules, syncactions, filebase, quiet, output, genbound, scope, check_for_accept, cmc, report_cmc):
	"""Generate the system LTS + state vector mapping of the loaded network description"""
	global hashtable, systemstatecounter, systemtranscounter, keep_quiet, outputobject
	accepting_state = -1
	systemstatecounter = 0
	systemtranscounter = 0
	hashtable.clear()
	systemtrans = {}
	keep_quiet = quiet
	if not keep_quiet:
		outputobject = output
	# to help determine type by shedskin
	a = 'a'
	actres = 'a'
	an = 'a'
	vect = ['a']
	v = set(['i'])
	tgts = set(['i'])
	targets = [set(['0'])]

	# BFS open queue
	openset = deque([['0']])
	openset.popleft()
	# BFS next queue
	nextstates = deque([['0']])
	nextstates.popleft()
	#create initial state
	current = [procsheaders[i][0] for i in range(0,len(procs))]

	[nw, index] = getIndex(current)
	nextstates.append(current)

	# dictionary to traverse backwards for counter-example generation
	preddict = {}

	# iteration counter
	iters = 0

	# sort the scope
	sorted_scope = sorted(scope)
	# explore scope
	explore_scope = sorted_scope
	if cmc:
		explore_scope = explore_scope[:-1]
	
	report_cmc_highest_state_ids = []
	report_cmc_transition_blocked = []
	scope_prefix_set = set([])
	process_checked = 0
	# highest position in scope (not equal to process checked) that has blocked at least one transition of process checked
	highest_blocking_pos = -1
	
	if report_cmc:
		# for compositional model checking: keep track of the highest state indices reached in process LTSs in the scope (except for the last process). These process LTSs are single traces.
		report_cmc_highest_state_ids = [0 for j in range(0,len(procs)-2)]
		# keep track of whether the outgoing transition of the highest states have been blocked once by the process LTS being checked
		report_cmc_transition_blocked = [False for j in range(0,len(procs)-2)]
		# set version of first part of scope
		scope_prefix_set = set(scope[:-3])
		# the process being checked
		process_checked = scope[len(scope)-3]

	abort = False
	while len(nextstates) > 0 and (genbound == 0 or iters < genbound) and not abort:
		openset = nextstates
		nextstates = deque([['0']])
		nextstates.popleft()
		while len(openset) > 0 and not abort:
			current = openset.popleft()
			#print current
			[nw, index] = getIndex(current)
			#get all outgoing transitions of the procstates in current
			trans = getOutgoing(current, procs, sorted_scope)
			#iterate over all the transitions to find suitable system transitions
			for i in explore_scope:
				for a, v in (trans[i]).items():
					# must the action synchronise?
					a_stripped = actionStripParams(a)
					if a_stripped not in syncactions[i]:
						# no sync, can be fired
						for js in v:
							target = copy.deepcopy(current)
							target[i] = js
							# get state index
							[nw, tgt_index] = getIndex(target)
							# add to nextstates?
							if nw == '1':
								#print 'add1 ' + str(target)
								nextstates.append(target)
								# check for accepting state
								if check_for_accept != set([]):
									# store pred info
									preddict[tgt_index] = index
									# possibly update highest state index reached
									if report_cmc:
										if i in scope_prefix_set:
											ijs = int(js)
											if report_cmc_highest_state_ids[i] < ijs:
												report_cmc_highest_state_ids[i] = ijs
												# update transition blocking info
												report_cmc_transition_blocked[i] = False
									# is state accepting?
									if is_accepting_state(target, getOutgoing(target, procs, sorted_scope), check_for_accept):
										accepting_state = tgt_index
										abort = True
							# add transition to systemtrans
							systemout = systemtrans.get(index)
							if systemout == None:
								systemtrans[index] = {a: set([tgt_index])}
								systemtranscounter = systemtranscounter + 1
							else:
								systemoutlabel = systemout.get(a)
								if systemoutlabel == None:
									systemout[a] = set([tgt_index])
									systemtrans[index] = systemout
									systemtranscounter = systemtranscounter + 1
								else:
									if tgt_index not in systemoutlabel:
										systemoutlabel.add(tgt_index)
										systemout[a] = systemoutlabel
										systemtrans[index] = systemout
										systemtranscounter = systemtranscounter + 1
					else:
						# there are possibly rules for this action
						rules = getSyncrules(a_stripped, i, syncrules, ((len(scope) != len(procs)) and i < len(procs)-2))
						if rules != []:
							for (vect, actres) in rules:
								if vect[i] != '*':
									# record first position relevant for this vector
									first_relevant_pos = -1
									targets = []
									# for each position in the rule that is in the scope, check if the sync can happen
									# k is current index in scope
									k = 0
									indexinscope = sorted_scope[k]
									# in which part of the scope is the transition enabled? and does the process being checked uniquely disable the transition?
									enabled_in_scope = []
									disabled_in_scope = []
									disabled_outside_scope = False
									for j in range(0,len(vect)):
										if vect[j] == '*':
											targets.append(set([current[j]]))
										else:
											if j == indexinscope and first_relevant_pos == -1:
												first_relevant_pos = j
											if j == i:
												targets.append(v)
												if report_cmc:
													if j == indexinscope and j < len(procs)-2:
														enabled_in_scope.append(j)
											else:
												# get the required action label
												#print "one party " + a
												#an = actionRename(a, vect[i], vect[j])
												#print "the other " + str(an)
												an = '"' + vect[j] + actionGetParams(a) + '"'												
												if j < indexinscope:
													# check if action is in procsactionset
													if an not in procsactsets[j]:
														# disabling not due to process_checked
														if report_cmc:
															disabled_outside_scope = True
														break
													targets.append(set([current[j]]))
												else:
													tgts = trans[j].get(an)
													if tgts == None:
														# disabled by process in scope
														if report_cmc and j < len(procs)-2:
															disabled_in_scope.append(j)
														else:
															break
													else:
														targets.append(tgts)
														if report_cmc and j < len(procs)-2:
															enabled_in_scope.append(j)
										if j == indexinscope:
											k += 1
											if k < len(sorted_scope):
												indexinscope = sorted_scope[k]
											else:
												indexinscope = len(procs)
									# is sync successful?
									if len(vect) == len(targets):
										# yes, get all combinations of new process states, if required
										if first_relevant_pos == i:
											# create a list of all possible combinations of target states
											#newcomb = list(itertools.product(*targets))
											#newcomb = list(list(x) for x in newcomb)
											newcomb = [[]]
											for j in range(0,len(targets)):
												if len(targets[j]) > 1:
													tmplist = copy.deepcopy(newcomb)
													newcomb = []
													for s in targets[j]:
														tlist = copy.deepcopy(tmplist)
														for k in tlist:
															k.append(s)
														newcomb += tlist
												else:
													for s in targets[j]:
														for c in newcomb:
															c.append(s)
											#result = a.find('GBUSRESET', 0, 10)
											#if result != -1:
											#	print current
											#	print newcomb
											# add the transitions to systemtrans
											newstates = set([])
											for tgt in newcomb:
												# get state index
												[nw, tgt_index] = getIndex(tgt)
												newstates.add(tgt_index)
												# add to nextstates?
												if nw == '1':
													#print 'add2 ' + str(tgt)
													nextstates.append(tgt)
													# check for accepting state
													if check_for_accept != set([]):
														# store pred info
														preddict[tgt_index] = index
														# possibly update highest state index reached
														if report_cmc:
															for j in enabled_in_scope:
																if report_cmc_highest_state_ids[j] < int(tgt[j]):
																	report_cmc_highest_state_ids[j] = int(tgt[j])
																	report_cmc_transition_blocked[j] = False
														# is state accepting?
														if is_accepting_state(tgt, getOutgoing(tgt, procs, sorted_scope), check_for_accept):
															accepting_state = tgt_index
															abort = True
											# add transitions to systemtrans
											#an = actionRename(a, vect[i], actres)
											an = '"' + actres + actionGetParams(a) + '"'
											systemout = systemtrans.get(index)
											if systemout == None:
												systemtrans[index] = {an: newstates}
												systemtranscounter = systemtranscounter + len(newstates)
											else:
												systemoutlabel = systemout.get(an)
												if systemoutlabel == None:
													systemout[an] = newstates
													systemtrans[index] = systemout
													systemtranscounter = systemtranscounter + len(newstates)
												else:
													old = len(systemoutlabel)
													systemout[an] = systemoutlabel | newstates
													systemtranscounter = systemtranscounter + (len(systemout[an]) - old)
									elif report_cmc:
										# if we are reporting for cmc, check whether a transition has just been blocked exclusively by process process_checked										
										if disabled_in_scope == [process_checked] and not disabled_outside_scope:
											for j in enabled_in_scope:
												report_cmc_transition_blocked[j] = True
										# alternatively, if process checked was enabled, record highest (in terms of scope position) process that disabled this transition
										elif process_checked in enabled_in_scope:
											for j in reversed(range(0,len(scope)-3)):
												if scope[j] in disabled_in_scope:
													if j > highest_blocking_pos:
														highest_blocking_pos = j
													break
		iters += 1
	# adjust the header
	systemLTSheader = ['0', str(systemtranscounter), str(systemstatecounter)]
	# write the data to disk
	writeAut(systemLTSheader, systemtrans, filebase, check_for_accept == set([]))
	if check_for_accept == set([]):
		writeMap(filebase)
	# if required and applicable, write counter-example to disk
	if check_for_accept	!= set([]):
		if accepting_state != -1:
			ce = []
			tgt = accepting_state
			while tgt != '0':
				src = preddict[tgt]
				outgoing = systemtrans[src]
				for a, tgts in outgoing.items():
					if tgt in tgts:
						ce.append(a)
						break
				tgt = src
			ceLTSheader = ['0', str(len(ce)), str(len(ce)+1)]
			cetrans = {}
			for i in range(0,len(ce)):
				cetrans[str(i)] = {ce[len(ce)-1-i]: set([str(i+1)])}
			writeAut(ceLTSheader, cetrans, filebase + "-ce", True)
		elif report_cmc:
			# report which part of the trace of the single trace process LTS that has been blocked and has the highest id has been successfully explored
			# (and which action right after that therefore could not be explored)
			ppos = len(scope)-4
			# select process that has blocked process checked at least once, if present
			if highest_blocking_pos > -1:
				ppos = highest_blocking_pos
			else:
				for i in reversed(range(0,len(scope)-3)):
					if report_cmc_transition_blocked[scope[i]]:
						ppos = i
						break
			pid = scope[ppos]
			extrace = [str(ppos)]
			for i in range(0,report_cmc_highest_state_ids[pid]):
				outgoing = procs[pid].get(str(i),{})
				for a, v in outgoing.items():
					extrace.append(a)
			# next action
			if report_cmc_highest_state_ids[pid] < (int(procsheaders[pid][2])-1):
				outgoing = procs[pid].get(str(report_cmc_highest_state_ids[pid]),{})
				for a, v in outgoing.items():
					extrace.append(a)
				extrace.append("False")
			else:
				extrace.append("True")

			#print(str(process_checked) + ": " + str(extrace))
			#print(highest_blocking_pos)
			#for i in scope[:-3]:
			#	print(str(i) + ": " + str(procs[i]))
			#	print(report_cmc_transition_blocked[i])

			with open(filebase + "-extrace",'w') as outfile:
				for a in extrace:
					outfile.write(a + '\n')
			outfile.close()

#	result = Searchdata(systemLTSheader, systemtrans, hashtable)
#	return result

if __name__ == '__main__':
	generateLTS([{'a': {'a': set(['a'])}}], [{'a': [(['a'],'a')]}], [set(['a'])], 'a')
