from tkinter import *
import tkinter.filedialog as filedialog
import config

class _Buttons:
  def __init__( self, frame, parent ):
    self.parent = parent

    self.load_rule_system_button = Button( frame, text="Load rule system", command=parent.load_rule_system, state=NORMAL, font=self.parent.customfont, activebackground="gray93" )
    self.save_rule_system_button = Button( frame, text="Save rule system", command=parent.save_rule_system, state=DISABLED, font=self.parent.customfont )
    self.load_property_button = Button( frame, text="Load property", command=self.load_property, state=NORMAL, font=self.parent.customfont )
    self.discard_property_button = Button( frame, text="Discard property", command=self.discard_property, state=DISABLED, font=self.parent.customfont )
    self.load_model_button = Button( frame, text="Load network", command=self.load_model, state=NORMAL, font=self.parent.customfont )
    
    self.load_rule_system_button.grid( row=0, sticky=W+E )
    self.save_rule_system_button.grid( row=1, sticky=W+E )
    self.load_property_button.grid( row=2, sticky=W+E )
    self.discard_property_button.grid( row=3, sticky=W+E )
    self.load_model_button.grid( row=4, sticky=W+E )

  def update( self, no_rules_and_processes ):
    if no_rules_and_processes == True:
      self.save_rule_system_button.config( state=DISABLED )
    else:
      self.save_rule_system_button.config( state=NORMAL )
    if config.property_name == "":
      self.discard_property_button.config( state=DISABLED )
    else:
      self.discard_property_button.config( state=NORMAL )
      
  def load_model( self ):
    model = filedialog.askopenfilename( defaultextension='.exp', filetypes=[ ('Networks of LTSs', '.exp') ] )
    if model != None and model != '' and model != ():
      self.parent.set_model( model )

  def load_property( self ):
    property = filedialog.askopenfilename( defaultextension='.mcl', filetypes=[ ('Mu-calculus formulas', '.mcl') ] )
    if property != None and property != '' and property != ():
      self.parent.set_property( property )

  def discard_property( self ):
    self.parent.set_property( "" )

class LoadAndSaveFrame( Frame ):
  def __init__( self, frame, parent, customfont ):
    Frame.__init__( self, frame )
    self.parent = parent
    
    self.customfont = customfont
    self.button_frame = Frame( self )

    self.buttons = _Buttons( self.button_frame, self )
    self.button_frame.pack( side=LEFT )
 
  def update( self, no_rules_and_processes ):
    self.buttons.update( no_rules_and_processes )

  def save_rule_system( self ):
    file_name = filedialog.asksaveasfilename( defaultextension='.rs', filetypes=[ ( 'Rule systems', '.rs' ) ] )
    if file_name != None and file_name != '' and file_name != ():
      self.parent.save_rule_system( file_name )

  def load_rule_system( self ):
    file_name = filedialog.askopenfilename( defaultextension='.rs', filetypes=[ ( 'Rule systems', '.rs' ) ] )
    if file_name != None and file_name != '' and file_name != ():
      self.parent.load_rule_system( file_name )

  def set_model( self, model ):
    self.parent.set_model( model )

  def set_property( self, property ):
    self.parent.set_property( property )
