from tkinter import *
from console_frame import *
from load_and_save_frame import *
from verify_and_transform_frame import *
from rule_and_process_edit_frame import *
from rule import *
from sync_frame import *
import copy, os
import config
from refiner import *
import _thread
import tkinter.font
import pprint

class Editor( Frame ):
  def __init__( self, parent ):
    Frame.__init__( self, parent )
    self.parent = parent
    self.grid()
    self.customfont = tkinter.font.Font(family="Arial", size=11, weight="normal")

    self.model = ""
    self.property = ""
    self.selected_item = None
    self.item_type = "none"

    self.parent.title( 'REFINER-gui' )
    self.top_frame = Frame( self )
    self.middle_frame = Frame( self )
    self.bottom_frame = Frame( self )

    self.top_frame.grid( row=0, sticky=W+E, padx=10, pady=10 )
    self.middle_frame.grid( row=1, sticky=W+E, padx=10 )
    self.bottom_frame.grid( row=2, sticky=W+E, padx=10, pady=10 )

    self.load_and_save_frame = LoadAndSaveFrame( self.top_frame, self, self.customfont )
    self.verify_and_transform_frame = VerifyAndTransformFrame( self.top_frame, self, self.customfont )
    self.console_frame = ConsoleFrame( self.top_frame, self.customfont )

    self.load_and_save_frame.grid( row=0, column=0, padx=10, pady=10 )
    self.verify_and_transform_frame.grid( row=0, column=1, sticky=N, padx=10, pady=10 )
    self.console_frame.grid( row=0, column=2, sticky=N, padx=10, pady=10 )

    self.edit_frame = RuleAndProcessEditFrame( self.middle_frame, self, self.customfont )
    self.edit_frame.pack( fill=X )

    self.sync_frame = SyncFrame( self.bottom_frame, self, self.customfont )
    self.sync_frame.pack( fill=X, padx=10, pady=10 )

  def log( self, string ):
    self.console_frame.log( string )

  def update( self ):
    no_rules_and_processes = len( config.rules_and_processes ) == 0
    if self.selected_item == None:
      self.edit_frame.update( None, self.item_type, no_rules_and_processes )
    else:
      self.edit_frame.update( config.rules_and_processes[ self.selected_item ], self.item_type, no_rules_and_processes )
    self.load_and_save_frame.update( no_rules_and_processes )
    self.verify_and_transform_frame.update( self.model, not( no_rules_and_processes ), self.property )
    self.sync_frame.update( config.existing_synchronisation_rules, config.new_synchronisation_rules)

  def show_item( self, selected_item, item_type ):
    self.item_type = item_type
    self.selected_item = selected_item
    self.update()

  def add_rule( self, item ):
    rule = Rule()
    config.rules_and_processes[ item ] = rule
    self.update()

  def add_process( self, item ):
    process = Process()
    config.rules_and_processes[ item ] = process
    self.update()

  def clone_item( self, new, old ):
    config.rules_and_processes[ new ] = copy.deepcopy( config.rules_and_processes[ old ] )
    self.update()
  
  def delete_item( self, item ):
    del config.rules_and_processes[ item ]
    self.selected_item = None
    self.item_type = "none"
    self.update()

  def add_state( self, name, x, y ):
    selected_item = config.rules_and_processes[ self.selected_item ]
    if self.item_type == "process":
      selected_item.add_state( x, y )
    elif self.item_type == "rule":
      if name == "left":
        selected_item.left.add_state( x, y )
      elif name == "right":
        selected_item.right.add_state( x, y )
    self.update()

  def add_glue_state( self, name, x, y ):
    selected_item = config.rules_and_processes[ self.selected_item ]
    if self.item_type == "process":
      selected_item.add_glue_state( x, y )
    elif self.item_type == "rule":
      selected_item.left.add_glue_state( x, y )
      selected_item.right.add_glue_state( x, y )
    self.update()

  def add_exclusive_glue_state( self, name, x, y ):
    selected_item = config.rules_and_processes[ self.selected_item ]
    if self.item_type == "process":
      selected_item.add_exclusive_glue_state( x, y )
    elif self.item_type == "rule":
      selected_item.left.add_exclusive_glue_state( x, y )
      selected_item.right.add_exclusive_glue_state( x, y )
    self.update()

  def delete( self, name, item, item_type ):
    selected_item = config.rules_and_processes[ self.selected_item ]
    if self.item_type == "process":
      selected_item.delete( item )
    elif self.item_type == "rule":
      if item_type == "glue_state" or item_type == "exclusive_glue_state":
        selected_item.left.delete( item )
        selected_item.right.delete( item )
      else:
        if name == "left":
          selected_item.left.delete( item )
        else:
          selected_item.right.delete( item )
    self.update()

  def add_transition( self, name, source, target, label, c1x, c1y, c2x, c2y, lx, ly ):
    selected_item = config.rules_and_processes[ self.selected_item ]
    if self.item_type == "process":
      selected_item.add_transition( source, target, label, c1x, c1y, c2x, c2y, lx, ly )
    elif self.item_type == "rule":
      if name == "left":
        selected_item.left.add_transition( source, target, label, c1x, c1y, c2x, c2y, lx, ly )
      elif name == "right":
        selected_item.right.add_transition( source, target, label, c1x, c1y, c2x, c2y, lx, ly )
    self.update()

  def update_synchronisation_rules( self ):
    ( config.existing_synchronisation_rules, config.new_synchronisation_rules ) = self.sync_frame.get_rules()

  def output_process( self, process, file ):
    number_of_exclusive_glue_states = len( process.exclusive_glue_states )
    number_of_glue_states = len( process.glue_states) + number_of_exclusive_glue_states
    number_of_states = len( process.states ) + number_of_glue_states
    
    file.write( "des(" + str( len( process.transitions ) ) + "," + str( number_of_states ) + "," + str( number_of_glue_states ) + "," + str( number_of_exclusive_glue_states ) + ")\n" )
    for index in process.exclusive_glue_states:
      state = process.exclusive_glue_states[ index ]
      file.write( "[" + str( state[ "x" ] ) + "," + str( state[ "y" ] ) + "]" )
    for index in process.glue_states:
      state = process.glue_states[ index ]
      file.write( "[" + str( state[ "x" ] ) + "," + str( state[ "y" ] ) + "]" )
    for index in process.states:
      state = process.states[ index ]
      file.write( "[" + str( state[ "x" ] ) + "," + str( state[ "y" ] ) + "]" )
    file.write( "\n" )
    for index in process.transitions:
      transition = process.transitions[ index ]
      trans_str = transition[ "label" ]
      if trans_str[0] != '"':
        trans_str = '"' + trans_str + '"'
      file.write(
        "(" + str( transition[ "source" ][ "id" ] ) + "," + trans_str + "," + str( transition[ "target" ][ "id" ] ) + ")" +
        "[" + str( transition[ "c1x"] ) + "," + str( transition[ "c1y" ] ) + "]" +
        "[" + str( transition[ "c2x" ] ) + "," + str( transition[ "c2y" ] ) + "]" +
        "[" + str( transition[ "lx" ] ) + "," + str( transition[ "ly" ] ) + "]" + "\n"
      )

  def save_rule_system( self, file_name ):
    self.update_synchronisation_rules()
    rp_list = config.rules_and_processes
    rules = [ rp_list[ item ] for item in rp_list if isinstance( rp_list[ item ], Rule ) ]
    processes = [ rp_list[ item ] for item in rp_list if isinstance( rp_list[ item ], Process ) ]
    file = open( file_name, "w" )
    file.write( "transform\n" )
    for rule in rules:
      self.output_process( rule.left, file )
      file.write( "->\n" )
      self.output_process( rule.right, file )
      file.write( "\n" )
    file.write( "add\n" )
    for process in processes:
      self.output_process( process, file )
      file.write( "\n" )
    file.write( "synch left\n" )
    file.write( config.existing_synchronisation_rules )
    file.write( "synch right\n" )
    file.write( config.new_synchronisation_rules )
    file.close()
    file_name_split = file_name.split('/')
    file_name_split = file_name_split[len(file_name_split)-1][:-3]
    self.log("finished writing rule system " + file_name_split + "\n")
    self.log("---\n")

  def load_rule_system( self, file_name ):
    config.rules_and_processes = {}
    config.existing_synchronisation_rules = ""
    config.new_synchronisation_rules = ""
    config.rs_folder, config.rs_name = os.path.split(file_name)
    config.rs_name = config.rs_name[:-3]
    refiner_readRulesystem()
    self.selected_item = None
    self.item_type = "none"
    self.edit_frame.selection_frame.next_id = len(config.rules_and_processes)
    self.update()

  def verify( self, divergency, fairness ):
    self.update_synchronisation_rules()
    _thread.start_new_thread(refiner_verify, (divergency, fairness))

  def transform( self, divergency ):
    self.update_synchronisation_rules()
    _thread.start_new_thread(refiner_transform, (divergency, ))

  def set_model( self, model ):
    self.model = model
	# detect folder of model
    config.network_folder, config.network_name = os.path.split(model)
    config.network_name = config.network_name[:-4]
    _thread.start_new_thread(refiner_readNetwork, (False, ))
    self.update()

  def set_property( self, property ):
    self.property = property
    if property != "":
      config.property_folder, config.property_name = os.path.split(property)
      config.property_name = config.property_name[:-4]
    else:
      config.property_folder = ""
      config.property_name = ""
    refiner_constructhidingset()
    self.update()

def main(args):
  for i in range(0,len(args)):
    if args[i] == '-t':
      i += 1
      config.nrthreads = int(args[i])
  root = Tk()
  root.option_add("*background", "gray93")
  root.tk_setPalette("background","gray93","activeBackground","gray93")
  #root.wm_resizable(0,0)
  # linux
  #root.geometry( '1170x800+0+0' )
  root.geometry( '1160x780+0+0' )
  editor = Editor( root )
  config.outputobject = editor
  editor.log("REFINER started. Multi-threading = " + str(config.nrthreads) + "\n")
  editor.log("---\n")
  root.mainloop()

if __name__ == '__main__':
  args = []
  for i in range(1,len(sys.argv)):
    args.append(sys.argv[i])
  main(args)
