# The REFINER tool, v.0.92 by Anton Wijs, 2013

# Added with v.0.92:
# - fixed synchronisation condition: required link between patterns of a rule is now checked
# - support for new rulesystem file format; all info now in one format
# - rulesystem description now contains abstract sync rules, before and after transformation;
#   these are mapped to concrete sync rules of input network

# Added with v.0.91:
# - interfaces to isolate parts of a system / rule vector for analysis purposes
# - new weakened application conditions (removal completeness, new synchronisation condition)
# - exclusive glue-states

# for timing measurements
from __future__ import with_statement
import time
import sys, string, collections, itertools, copy, pickle
import os
import subprocess
#from shutil import copyfile
from graphfunctions import generateLTS
from collections import deque, Counter
import re
import multiprocessing
import config
import glob
from rule import *

# command line options
opt_prop = ""
opt_generate = False
opt_safety_onthefly = False
opt_divergency = False
opt_transform = ""
opt_check = ""
opt_network = ""
opt_fairness = False
opt_expopenoutput = False
opt_quiet = False
opt_constructhidingset = False
opt_compositional_mc = False
opt_genbound = 0

#the processes
proccounter = 0
procfiles = []
# the process headers
procsheaders = []
# the standard process LTSs, structurally ordered by source states
procs = []
# the number of process LTSs, before transformation
procs_orig_size = 0
# the reversed process LTSs, structurally ordered by target states
revprocs = []
# the action process LTSs, structurally ordered by action (transition) labels
actprocs = []
# the process LTSs after a transformation
transformed_procs = []
new_procfiles = []
# the process headers after a transformation
transformed_procsheaders = []
# a mapping from new rule states (right hand patterns) to states in procs
transformed_procs_new_states = []
# the action sets of the processes
procsactsets = []
# the action prefix sets of the processes
procsactprefixsets = []
# set of action patterns to be hidden (-a option)
hideactions = ['none']
# set of concrete actions hidden
hiddenactions = set([])
# currently, three types of sync descriptions are supported.
# these are, together with their encoding in inputsync:
# 1. single action syncing with action(s) of same name, e.g. A #2 -> (1, A, 2)
# 2. multiple actions A_0, ..., A_n syncing, leading to action B, order of actions important
#    (uses * for don't care values), e.g. (<A_0, *, A_1>,B) -> (2, (A_0, *, A_1), B)
# 3. multiset of actions A_0, ..., A_n, leading to action B. These actions must all take
#    part in the sync, but there is no order imposed (no * used), e.g. ({A_0, A_1},B) -> (3, (A_0, A_1), B)
inputsync = {}
#the final, rewritten sync rules
syncrules = {}
#the sync actions
syncactions = []
tmpsynclist = []
tmpsynccounter = -1
#the structure of the network
inputnetwork = []
tmplists = [[]]
tmpcounter = 0
lastsyncaction = ""
acttmplist = []
tmprule = ""

#data for state space searches
systemLTSheader = []
systemLTSheader.append('0')
systemLTSheader.append('0')
systemLTSheader.append('0')
systemtrans = {}
systemactset = set([])
hashtable = {}
revhashtable = {}
systemstatecounter = 0
systemtranscounter = 0

# set of tau-diverging states
divergencyset = set([])
# sets of non-diverging process states
procndivsets = []

# a rule system is a finite set of LTS-pairs
RS = []
# add rules of rule system
RS_add = []
# tuples of sets of prefixes of actions in left and right patterns, respectively
RS_actprefixsets = []
# description (i.e. list of rule id's) how to construct vectors of (dependent) transformation rules
RS_vector = []
# number of rule instances in RS_vector that are transform rules (as opposed to add rules)
RS_vector_nr_transrules = 0
# functions assigning rule id (owner) to actions; two functions in list, first for left patterns, second for right
RS_ownsaction = [{},{}]
# syncrules in the rulesystem w.r.t. the left-hand patterns
RS_left_syncrules = {}
RS_left_syncactions = []
# syncrules in the rulesystem w.r.t. the left-hand patterns, optimised for state space generation
RS_gen_left_syncrules = []
# new syncrules introduced by the rulesystem
RS_right_syncrules = {}
RS_right_syncactions = []
# syncrules in the rulesystem w.r.t. the right-hand patterns, optimised for state space generation
RS_gen_right_syncrules = []
# full list of syncrules for network after transformation
newsyncrules = {}
# full sets of actions for network after transformation
newsyncactions = []
# the concrete actions in the process action sets after transformation subject to synchronisation
#newprocssyncactsets = []

# on which processes does a rule match?
rulematcheson = []
# matches information for all processes, and all rules, from rule state (left pattern) to process states
procsrulematches = []
# COMBINED matches information for all processes, and all rules, from rule state (left pattern) to process states
procsrulestatemaps = []

# a list of checks for property preservation. A check consists of two sets of dependent rule id's (transformation rule id's and add rule id's)
checks = []
# a list of action exclusion sets. For each check, the corresponding exclusion set should be excluded from hiding
exclsets = []

# global variables for compositional model checking
# dependency relation of process LTSs w.r.t. the property LTS (directly dependent LTSs are in first set in list, one step indirectly dependent in the next, etc.
prop_dependency = []

# cmc: compositional model checking structures
initial_constraintsLTSs = []
constraintsLTSs = []
initial_min_constraintsLTSs = []
min_constraintsLTSs = []
propertyLTSs = []
cmc_syncrules = []
cmc_syncactions = []
cmc_constraintsdep = []
cmc_checktree = ({},set([]))
cmc_checklist = []
checkprocs = []
checkheaders = []
# list to keep track of constraints updates to be made
cmc_constraints_updates = []
constraintsLTS_is_extended = []

class Timer(object):
		def __enter__(self):
				self.__start = time.time()

		def __exit__(self, type, value, traceback):
				# Error handling here
				self.__finish = time.time()

		def duration_in_seconds(self):
				return self.__finish - self.__start

timer = Timer()

def print_error(s):
	"""Print message to screen and exit"""
	if not opt_quiet:
		if config.outputobject == "":
			print("ERROR: " + str(s))
		else:
				config.outputobject.log("ERROR: " + str(s) + "\n")
	if config.outputobject == "":
		sys.exit(1)

def print_message(s):
	"""Print message to screen"""
	if not opt_quiet:
		if config.outputobject == "":
			print(s)
		else:
			config.outputobject.log(str(s) + "\n")

def print_separator():
	"""Print separator to screen"""
	if not opt_quiet and config.outputobject == "":
		print('---')

def toTupleOfTuples(l):
	"""Convert list of lists (depth unknown) to tuple of tuples"""
	if hasattr(l, "__iter__") and not isinstance(l, str):
		return tuple(toTupleOfTuples(x) for x in l)
	else:
		return l

# START PERMUTATIONS WITH UNIQUE VALUES

class unique_element:
	def __init__(self,value,occurrences):
		self.value = value
		self.occurrences = occurrences

def perm_unique(elements):
	eset=set(elements)
	listunique = [unique_element(i,elements.count(i)) for i in eset]
	u=len(elements)
	return perm_unique_helper(listunique,[0]*u,u-1)

def perm_unique_helper(listunique,result_list,d):
	if d < 0:
		yield tuple(result_list)
	else:
		for i in listunique:
			if i.occurrences > 0:
				result_list[d]=i.value
				i.occurrences-=1
				for g in  perm_unique_helper(listunique,result_list,d-1):
					yield g
				i.occurrences+=1

# END PERMUTATIONS WITH UNIQUE VALUES

# BEGIN I/O FUNCTIONS

def incrtmpcounter():
	global tmpcounter
	tmpcounter = tmpcounter + 1
	tmplists.append([])

def getNewRule(type, acts, an, structure):
	"""Create a new rule with structure (vector of actions, resulting action)"""
	if type == 1:
		# simple rule
		vect = []
		for i in range(0,int(an)):
			vect.append(acts)
		for i in range(len(vect), len(structure)):
			vect.append('*')
		allvects = list(perm_unique(vect))
		return (allvects, acts)
	if type == 2:
		# vector rule
		#print 'vector rule' + str(acts)
		return ([acts], an)
	if type == 3:
		# set rule
		lacts = list(acts)
		for i in range(len(lacts), len(structure)):
			lacts.append('*')
		allvects = list(perm_unique(lacts))
		#print allvects
		return (allvects, an)

def dectmpcounter():
	global tmpcounter, tmpsynccounter, tmplists, inputsync
	#first add the sync actions
	structure = tmplists[tmpcounter]
	tupkey = toTupleOfTuples(structure)
	for (type, acts, an) in tmpsynclist[tmpsynccounter]:
		newrule = getNewRule(type, acts, an, structure)
		result = inputsync.get(tupkey)
		if result == None:
			inputsync[tupkey] = [newrule]
		else:
			result.append(newrule)
			inputsync[tupkey] = result
	tmplists[tmpcounter-1].append(tmplists[tmpcounter])
	del tmpsynclist[tmpsynccounter]
	tmpsynccounter = tmpsynccounter - 1
	del tmplists[tmpcounter]
	tmpcounter = tmpcounter - 1

def addProcess(a):
	global proccounter
	file = (a[1:])[:-1]
	procfiles.append(file)
	(tmplists[tmpcounter]).append(proccounter)
	proccounter = proccounter + 1

def addSyncAction(alist):
	if len(alist) > 1:
		mult = alist[1]
	else:
		mult = '2'
	(tmpsynclist[tmpsynccounter]).append((1, alist[0], mult))

def addActionToList(a):
	global acttmplist
	acttmplist.append(a)

def addVectResult(a):
	global acttmplist, tmpsynclist, tmpsynccounter
	(tmpsynclist[tmpsynccounter]).append((2, toTupleOfTuples(acttmplist), a))
	acttmplist = []

def addSetResult(a):
	global acttmplist, tmpsynclist, tmpsynccounter
	(tmpsynclist[tmpsynccounter]).append((3, toTupleOfTuples(acttmplist), a))
	acttmplist = []

def setNewSyncList():
	global tmpsynccounter
	tmpsynccounter = tmpsynccounter + 1
	tmpsynclist.append([])

def addRule(alist):
	global rulefiles
	rulefiles.append((alist[0][1:][:-1], alist[1][1:][:-1], alist[2]))

def parseSyncRule(a):
	"""Parse sync action"""
	if a[0] == '(':
		if a[1] == '<':
			i = a.rfind('>')
			t = a[2:][:-(len(a)-i)]
			tlist = t.split(',')
			for ta in tlist:
				addActionToList(ta)
			ta = a[:-1][i+2:]
			addVectResult(ta)
		else:
			i = a.rfind('}')
			t = a[2:][:-(len(a)-i)]
			tlist = t.split(',')
			for ta in tlist:
				addActionToList(ta)
			ta = a[:-1][i+2:]
			addSetResult(ta)			
	else:
		tlist = a.split(' ')
		if len(tlist) > 1:
			tlist[1] = tlist[1][1:]
		addSyncAction(tlist)

def parseNetwork(inFile):
	"""Parse the given network file"""
	tokens = []
	lines = inFile.readlines()
	for l in lines:
		tlist = l.split(' ')
		for t in tlist:
			t = t.replace(' ','')
			t = t.replace('\t','')
			t = t.rstrip()
			if t != '':
				tokens.append(t)
	pos = 0
	skip = 0
	while pos < len(tokens):
		if tokens[pos] == 'par':
			setNewSyncList()
			pos += 1
			if tokens[pos] == 'using':
				pos += 1
			while tokens[pos] != 'in':
				tokens[pos] = tokens[pos].rstrip(',')
				skip = 0
				# a syncrule of type 1? (e.g. a #2)
				if tokens[pos+1][0] == '#':
					tokens[pos] += ' ' + tokens[pos+1]
					tokens[pos] = tokens[pos].rstrip(',')
					skip = 1
				# a syncrule of type 2 in exp.open style? (e.g. a * _ * b -> c or a || _ || b -> c)
				# if yes, we convert this rule to our format (e.g. (<a,*,b>,c))
				elif tokens[pos+1][0] == '*' or tokens[pos+1][0] == '||':
					if tokens[pos] == '_':
						tokens[pos] = '*'
					tokens[pos] = '(<' + tokens[pos]
					skip = 1
					while tokens[pos+skip] != '->':
						if tokens[pos+skip] == '*' or tokens[pos+skip] == '||':
							tokens[pos] += ','
						elif tokens[pos+skip] == '_':
							tokens[pos] += '*'
						else:
							tokens[pos] += tokens[pos+skip]
						skip += 1
					tokens[pos] += '>,'
					skip += 1
					tokens[pos] += tokens[pos+skip].rstrip(',')
					tokens[pos] += ')'
				parseSyncRule(tokens[pos])
				pos += skip + 1
			incrtmpcounter()
		elif tokens[pos] == '||':
			pos += 1
			continue
		elif tokens[pos][0] == '"':
			addProcess(tokens[pos])
		elif tokens[pos] == 'end':
			dectmpcounter()
			pos += 1
		else:
			print_error("unexpected input in network description!")
		pos += 1

def parseSet(S):
	"""Parse the given application scope into a set"""
	S = S[1:][:-1]
	S.replace(' ', '')
	slist = S.split(',')
	rset = set([])
	for s in slist:
		rset.add(int(s))
	return rset

def readNetwork(networkfile, quiet):
	"""Read a network (.exp) file"""
	global syncrules, syncactions, procs, procsactsets, procsheaders, inputnetwork, procs_orig_size, procfiles
	try:
		inFile = open(os.path.join(config.network_folder, config.network_name + ".exp"),'r')
	except IOError:
		if not quiet:
			print_error("network file does not exist!")
		return []

	parseNetwork(inFile)
	inputnetwork = tmplists[0]
	inputnetwork = inputnetwork[0]

	#read the .aut files
	for i in range(0,len(procfiles)):
		[h,v,a] = readAut(config.network_folder, procfiles[i])
		procs.append(v)
		procsactsets.append(a)
		procsheaders.append(h)

	procs_orig_size = len(procs)
	if not quiet:
		print_message("Process LTSs have been read from disk")

	#flatten the input sync rules
	#if not opt_noflatten:
	# construct the procsactprefixsets
	storeProcsactprefixsets()
	syncrulestmp = rewriteAllSync(inputnetwork)
	#else:
	#	syncrulestmp = {}
	#	for st, rs in inputsync.items():
	#		for (vects, actres) in rs:
	#			syncrulestmp[actres] = vects
	#	syncrulestmp = [syncrulestmp]
	for a, v in syncrulestmp[0].items():
		newv = []
		for vec in v:
			if countVectorElements(vec) > 1:
				newv.append(vec)
		if newv != []:
			syncrulestmp[0][a] = newv
	# build the syncaction sets for each process
	syncactions = buildSyncactionsets(syncrulestmp[0], len(procs))
	# order the rules in a list of dictionaries for the generation of the system LTS
	syncrules = []
	for i in range(0,len(procs)):
		syncrules.append({})
	syncrules = orderSyncrules(syncrulestmp[0], syncrules)
	#concrete_syncrules = computeConcreteSyncrules(syncrules, procssyncactsets, [])
	syncblocked = syncrulestmp[1]

def writeSyncrules(syncrules, outFile, expopenformat):
	"""Write the syncrules to the network file"""
	global newsyncactions, transformed_procs
	outFile.write("par")
	j = 0
	tmpstring = "-1"
	for i in range(0,len(syncrules)):
		for a, v in syncrules[i].items():
			for k1 in range(0,len(v)):
				if tmpstring != "-1":
					outFile.write(",")
				outFile.write("\n")
				if v[k1][0][0] == '*':
					tmpstring = "  " + "_"
				else:
					tmpstring = "  " + v[k1][0][0]
				for k2 in range(1,len(v[k1][0])):
					tmpstring += " * "
					if v[k1][0][k2] == '*':
						tmpstring += "_"
					else:
						tmpstring += v[k1][0][k2]
				tmpstring += " -> " + v[k1][1]
				outFile.write(tmpstring)
		j += 1
	# if we need to write in exp.open format, we must also include
	# rules for all one-party syncs (i.e. individual action firing)
	# NOTE 1: In that case, we assume we are writing the transformed network
	# NOTE 2: We strip the actions off any substring following ' !' (get rid of gate parameters)
	if expopenformat:
		for i in range(0,len(transformed_procs)):
			newacts = set([])
			atrans = getActTrans(transformed_procs[i])
			for a, v in atrans.items():
				a_stripped = actionStripParams(a)
				if a != 'i' and a != 'tau':
					newacts.add(a_stripped)
			for a in newacts:
				if a not in newsyncactions[i]:
					outFile.write(",\n")
					tmpstring = "  "
					for j in range(0,i):
						tmpstring += "_ * "
					tmpstring += str(a)
					for j in range(i+1,len(transformed_procs)):
						tmpstring += " * _"
					tmpstring += " -> " + str(a)
					outFile.write(tmpstring)
	outFile.write("\nin\n")

def writeNetwork(folder, procs, headers, procfiles, syncrules, networkfile, expopenformat):
	"""Write the transformed network in a file"""
	global newprocfiles
	# open a new file
	i = networkfile.rfind(".")
	if i == -1:
		newfile = networkfile + ".1"
		version = 0
	else:
		version = int(networkfile[i+1:])
		newfile = networkfile[:i] + "." + str(version+1)
	config.transformed_network_name = newfile
	outFile = open(os.path.join(folder, config.transformed_network_name + ".exp"), 'w')
	writeSyncrules(syncrules, outFile, expopenformat)
	newprocfiles = []
	for j in range(0,len(procfiles)):
		# create new .aut file name
		i = (procfiles[j]).find(".")
		iaut = (procfiles[j]).find(".aut")
		if i == iaut:
			tmp = (procfiles[j])[:-4] + ".1"
		else:
			version = int(((procfiles[j])[i+1:])[:-4])
			tmp = (procfiles[j])[:i] + "." + str(version+1)
		newprocfiles.append(tmp)
	# create new names for newly introduced processes
	i = 0
	for j in range(len(procfiles),len(procs)):
		tmp = "p" + str(i) + "_" + str(version) + "." + str(version+1)
		newprocfiles.append(tmp)
		i += 1
	for j in range(0,len(newprocfiles)-1):
		outFile.write("  \"" + newprocfiles[j] + ".aut\"\n  ||\n")
	j = len(newprocfiles)-1
	outFile.write("  \"" + newprocfiles[j] + ".aut\"\n" + "end par\n")

	# finally, write the .aut files
	for j in range(0,len(newprocfiles)):
		writeAut(folder, headers[j], procs[j], newprocfiles[j])

autscanner=re.Scanner([
	(r"des",															lambda scanner,token:("HEADER", token)),
	(r"[0-9]+",														lambda scanner,token:("NAT", token)),
	(r"[\s\(\),]+", None), # None == skip token.
	(r"[A-Za-z0-9!\_\-?.\*]+\([A-Za-z0-9!\_\-?,.\*\(\)]+\)|[A-Za-z0-9!\_\-?.\*]+|[\"\'][A-Za-z0-9!?\_\-\#\s.,\*\(\)]+[\"\']",	lambda scanner,token:("ACTION", token)),
])

def readAut(folder, autfile):
	"""Read the .aut file and place the data in a dictionary"""
	trans = {}

	if autfile == "oneglue.aup":
		return [['1', '0', '1', '1'], {}, set([])]
	if autfile == "empty.aut":
		return [['0', '0', '0'], {}, set([])]
	try:
		inFile = open(os.path.join(folder, autfile),'r')
	except IOError:
		print_error(".aut file does not exist!")
		return [['0','0','0'], {}, set([])]
	# scan the first line
	line = inFile.readline()
	global token, next
	results, remainder = autscanner.scan(line)
	inFile.close()

	# read the header
	if results[0][0] != 'HEADER':
		print_error("unexpected start of .aut description!")
	transheader = []
	# an aut file has three values in the header
	for index in range(1,4):
		transheader.append(results[index][1])

	# count to remove double entries
	ntrans = 0
	actset = set([])

	#store transitions in memory
	currentsrc = ''
	newentry = {}
	first = True
	for transline in open(os.path.join(folder, autfile)):
		tl, remainder = autscanner.scan(transline)
		src = str(tl[0][1])
		act = str(tl[1][1])
		tgt = str(tl[2][1])
		if not first:
			# add quotes if not present
			act = act.replace('\'', '"')
			if act[0] != '"':
				act = '"' + act + '"'
			# remove spaces
			act = act.replace(' ', '')
			# put transition label in actset
			actset.add(act)
			# are we building a new dictionary entry?
			if src != currentsrc:
				# possibly add old entry if not empty
				if currentsrc != '':
					trans[currentsrc] = newentry
				currentsrc = src
				# obtain entry built so far
				newentry = trans.get(currentsrc,{})
			outgoingwithlabel = newentry.get(act)
			if outgoingwithlabel == None:
				newentry[act] = set([tgt])
				ntrans += 1
			else:
				if tgt not in outgoingwithlabel:
					# insert new element
					outgoingwithlabel.add(tgt)
					newentry[act] = outgoingwithlabel
					ntrans += 1
		else:
			first = False
	# add final entry
	trans[currentsrc] = newentry
	transheader[1] = str(ntrans)
	return [transheader, trans, actset]

def writeAut(folder, header, trans, autfile):
	"""Write the trans LTS to file autfile"""
	outfile = open(os.path.join(folder, autfile + ".aut"),'w')

	#print header of .aut file
	if header[0] == '':
		header[0] = '0'
	outfile.write('des (' + header[0] + ', ' + header[1] + ', ' + header[2] + ')\n')
	for i in range(0,int(header[2])):
		outgoing = trans.get(str(i))
		if outgoing != None:
			for k, v in outgoing.items():
				label = k
				if label[0] != '"':
					label = '"' + str(k) + '"'
				for t in v:
					outfile.write('(' + str(i) + ', ' + k + ', ' + t + ')\n')
	outfile.close()

def writeMap(folder, filebase):
	"""Write the mapping between system state indices and process state vectors, as established in the hashtable"""
	global hashtable
	
	outfile = open(os.path.join(folder, filebase + ".map"), 'w')
	for v, k in hashtable.items():
		outfile.write(k + ', ' + v + '\n')
	outfile.close()

def readMap(folder, filebase):
	"""Read the mapping between system state indices and process state vectors"""
	global hashtable
	result = {}
	try:
		inFile = open(os.path.join(folder, filebase + ".map"), 'r')
	except IOError:
		print_error("the mapping file does not exist!")
	inFile.close()
	#lines = inFile.readlines()
	for line in open(os.path.join(folder, filebase + ".map")):
		linelist = line.split(', ')
		vlist = linelist[1].split(',')
		vlist[len(vlist)-1] = vlist[len(vlist)-1][:-1]
		result[linelist[0]] = vlist[1:]
	return result

def reverseHashtable():
	"""Reverse the contents of hashtable, and put it in revhashtable"""
	global hashtable, revhashtable
	revhashtable.clear()
	for k, v in hashtable.items():
		revhashtable[v] = k

def readProcndivsets(filebase, propbase):
	"""Read the non-diverging process states from disk"""
	global procndivsets
	try:
		inFile = open(os.path.join(config.network_folder, filebase + "_" + propbase + ".ndv"), 'rb')
	except IOError:
		print_error("the file with non-diverging process states does not exist!")
	procndivsets = pickle.load(inFile)
	inFile.close()

def writeProcndivsets(filebase, propbase):
	"""Write the non-diverging process states to disk"""
	global procndivsets
	outFile = open(os.path.join(config.network_folder, filebase + "_" + propbase + ".ndv"), 'wb')
	pickle.dump(procndivsets, outFile)
	outFile.close()

def writeNewProcndivsets(filebase, propbase):
	"""Write the updated non-diverging process states to disk in new file"""
	i = filebase.rfind(".")
	if i == -1:
		newfile = filebase + ".1"
	else:
		version = int(filebase[i+1:])
		newfile = filebase[:i] + "." + str(version+1)
	writeProcndivsets(newfile, propbase)

def readProcsrulestatemaps(filebase, rulebase):
	"""Read procsrulestatemaps info"""
	global procsrulestatemaps
	try:
		inFile = open(os.path.join(config.network_folder, filebase + "_" + rulebase + ".pmp"), 'rb')
	except IOError:
		print_error("Checks information does not exist! First check for well-formedness and completeness!")
	procsrulestatemaps = pickle.load(inFile)
	inFile.close()

def writeProcsrulestatemaps(filebase, rulebase):
	"""Write procsrulestatemaps info"""
	global procsrulestatemaps
	outFile = open(os.path.join(config.network_folder, filebase + "_" + rulebase + ".pmp"), 'wb')
	pickle.dump(procsrulestatemaps, outFile)
	outFile.close()

RSscanner=re.Scanner([
	(r"transform",															lambda scanner,token:("SECTIONSTART", 0)),
	(r"add",																	lambda scanner,token:("SECTIONSTART", 1)),
	(r"des",																	lambda scanner,token:("HEADER", token)),
	(r"\[[0-9]+[.[0-9]+]?,[0-9]+[.[0-9]+]?\]",								lambda scanner,token:("COORDS", token)),
	(r"[0-9]+",																lambda scanner,token:("NAT", token)),
	(r"[\s,\(\)]+",															None), # None == skip token.
	(r"->",																	lambda scanner,token:("TO", token)),
	(r"[A-Za-z0-9!\_\-\?.\*]+\([A-Za-z0-9!\_\-\?,.\*\(\)]+\)|[A-Za-z0-9!\_\-\?.\*]+|[\"\'][A-Za-z0-9!\?\_\-\#\s.,\*\(\)]+[\"\']",	lambda scanner,token:("ACTION", token)),
])

syncrulescanner=re.Scanner([
	(r"synch\sleft",															lambda scanner,token:("SECTIONSTART", 2)),
	(r"synch\sright",															lambda scanner,token:("SECTIONSTART", 3)),
	(r"->",																	lambda scanner,token:("TO", token)),
	(r"[A-Za-z0-9!\_\-?.]+|\*|\_",	lambda scanner,token:("ACTION", token)),
	(r"[\{\}\<\>]",					lambda scanner,token:("BRACKET", token)),
	(r"#[0-9]+",						lambda scanner,token:("ARITY", token)),
	(r"[\s,\(\)]+",					None), # None == skip token.
])

# START PERMUTATIONS WITH UNIQUE VALUES ON SPECIFIC POSITIONS

# Permutation function for constructing valid synchronisation rules

# f is a function indicating for each action which positions are valid
# size is the requested size of the vectors (depends on whether we want network or rule system sync rules)
def perm_pos_unique(elements, f, size):
	eset=set(elements)
	# compute possible elements for every vector position
	dom = []
	for i in range(0,size):
		dom.append(set([]))
	for e in eset:
		for i in f.get(e):
			dom[i].add(e)
	# determine where to add *'s
	for i in range(0,len(dom)):
		S = set([])
		totalcount = 0
		for a in dom[i]:
			S |= f.get(a)
			totalcount += elements.count(a)
		if len(S) > totalcount or S == set([]):
			dom[i].add('*')
	listposunique = [unique_element(i,elements.count(i)) for i in eset]
	listposunique.append(unique_element('*',size-len(elements)))
	return perm_pos_unique_helper(listposunique,['*']*size,size-1,dom)

def perm_pos_unique_helper(listposunique,result_list,d,dom):
	if d < 0:
		yield tuple(result_list)
	else:
		for i in listposunique:
			if i.value in dom[d]:
				if i.occurrences > 0:
					result_list[d]=i.value
					i.occurrences-=1
					for g in perm_pos_unique_helper(listposunique,result_list,d-1,dom):
						yield g
					i.occurrences+=1

# END PERMUTATIONS WITH UNIQUE VALUES ON SPECIFIC POSITIONS

def extractPattern(p):
	"""Extract a pattern description (for verification/transformation) from a given process object p"""
	transheader = []
	trans = {}
	actset = set([])
	# we rearrange the header for backwards compatibility
#	htmp = transheader[2]
	transheader.append(str(len(p.glue_states) + len(p.exclusive_glue_states)))
	transheader.append(str(len(p.transitions)))
	transheader.append(str(len(p.states) + len(p.glue_states) + len(p.exclusive_glue_states)))
	transheader.append(str(len(p.exclusive_glue_states)))
	# count to remove double entries
	ntrans = 0
	#store transitions in memory
	for (tid, t) in p.transitions.items():
		src = str(t["source"]["id"])
		act = str(t["label"])
		tgt = str(t["target"]["id"])
		# add quotes if not present
		act.replace('\'', '"')
		if act[0] != '"':
			act = '"' + act + '"'
		# put transition label in actset
		actset.add(act)
		# add transition
		outgoing = trans.get(src)
		if outgoing == None:
			outgoing = {}
		outgoingwithlabel = outgoing.get(act)
		if outgoingwithlabel == None:
			outgoing[act] = set([tgt])
			ntrans += 1
		else:
			if tgt not in outgoingwithlabel:
				# insert new element
				outgoingwithlabel.add(tgt)
				outgoing[act] = outgoingwithlabel
				ntrans += 1
		trans[src] = outgoing
	transheader[1] = str(ntrans)
	return [transheader, trans, actset]

def readProcess(l, index):
	"""Read the LTS pattern in list l at index"""
	if l[index][0] != 'HEADER':
		print_error("unexpected start of pattern description!")
	newproc = Process()
	transheader = []
	# a pattern has four values in the header
	index += 1
	for j in range(0,4):
		transheader.append(l[index][1])
		index += 1
	# sanity check
	if int(transheader[1]) < int(transheader[2]):
		print_error("the number of states must be >= the number of glue-states!")
	if int(transheader[2]) < int(transheader[3]):
		print_error("the number of glue-states must be >= the number of exclusive glue-states!")
	# possibly read the coords of the states
	read_coords = False
	if l[index][0] == 'COORDS':
		for i in range(0,int(transheader[1])):
			coords = l[index][1].strip('\[\]').split(',')
			if i < int(transheader[3]):
				newproc.add_exclusive_glue_state(float(coords[0]), float(coords[1]))
			elif i < int(transheader[2]):
				newproc.add_glue_state(float(coords[0]), float(coords[1]))
			else:
				newproc.add_state(float(coords[0]), float(coords[1]))
			index += 1
		# read coords in remainder of file
		read_coords = True
	else:
		# add states without coordinates
		for i in range(0,int(transheader[1])):
			if i < int(transheader[3]):
				newproc.add_exclusive_glue_state(0.0, 0.0)
			elif i < int(transheader[2]):
				newproc.add_glue_state(0.0, 0.0)
			else:
				newproc.add_state(0.0, 0.0)
	#store transitions in memory
	currentsrc = ''
	newentry = {}
	while index < len(l):
		if l[index][0] != 'NAT':
			break
		src = l[index][1]
		act = l[index+1][1]
		tgt = l[index+2][1]
		index += 3
		if read_coords:
			coords = []
			coords.append(tuple(l[index][1].strip('\[\]').split(',')))
			coords.append(tuple(l[index+1][1].strip('\[\]').split(',')))
			coords.append(tuple(l[index+2][1].strip('\[\]').split(',')))
			index += 3
		# add quotes if not present
		act.replace('\'', '"')
		if act[0] != '"':
			act = '"' + act + '"'
		if read_coords:
			newproc.add_transition(int(src), int(tgt), act, float(coords[0][0]), float(coords[0][1]), float(coords[1][0]), float(coords[1][1]), float(coords[2][0]), float(coords[2][1]))
		else:
			newproc.add_transition(int(src), int(tgt), act, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
	return index, newproc

def RS_getRulesWithAction(action, side):
	"""Lookup the rule containing the given action in the left (side==False) or right (side==True) pattern"""
	global RS, RS_add, RS_actprefixsets, RS_ownsaction
	elside = 0
	if side:
		elside = 1
	if action == '*':
		return set([])
	indices = RS_ownsaction[elside].get(action)
	if indices != None:
		return indices
	indices = set([])
	for i in range(0,len(RS)):
		if action in RS_actprefixsets[i][elside]:
			indices.add(i)
	for i in range(0,len(RS_add)):
		if action in RS_actprefixsets[i+len(RS)][elside]:
			indices.add(len(RS)+i)
	RS_ownsaction[elside][action] = indices
	return indices

def RS_getNewRules(side, type, acts, an):
	"""Create new sync rules for rule system (vector of actions, resulting action, multiset of actions from outside the rule system)"""
	global RS_vector
	posforaction = {}
	interface = Counter()
	elside = 0
	if side:
		elside = 1
	if type == 1:
		# simple rule
		S = RS_getRulesWithAction(acts,side)
		if S == set([]):
			interface[acts] += 1
			return ([], '', interface)
		vect = []
		for i in range(0,int(an)):
			vect.append(acts)
		rltact = acts
	if type == 2 or type == 3:
		# vector rule
		vect = []
		for act in acts:
			if act != '*':
				S = RS_getRulesWithAction(act,side)
				if S == set([]):
					interface[act] += 1
				else:
					vect.append(act)
		rltact = an
	for action in vect:
		positions = posforaction.get(action)
		if positions == None:
			positions = set([])
			indices = RS_ownsaction[elside].get(action)
			for i in range(0,len(RS_vector)):
				if RS_vector[i] in indices:
					positions.add(i)
			posforaction[action] = positions
	tmplist = list(perm_pos_unique(vect, posforaction, len(RS_vector)))
	return (tmplist, rltact, interface)

def preprocessRulesystem():
	"""Convert the 'visual' representation of a rule system to the one used for transformation and verification"""
	global RS, RS_add, RS_vector, procsheaders, procs, RS_left_syncactions, RS_right_syncactions, syncrules, RS_left_syncrules, RS_right_syncrules, RS_gen_left_syncrules, RS_gen_right_syncrules, newsyncrules, newsyncactions, acttmplist, tmpsynclist, tmpsynccounter
	global token, next, inputnetwork, procndivsets

	# clear previous contents
	RS = []
	RS_add = []
	RS_vector = []
	RS_left_syncactions = []
	RS_right_syncactions = []
	RS_left_syncrules = []
	RS_right_syncrules = []
	RS_gen_left_syncrules = []
	RS_gen_right_syncrules = []
	procndivsets = []

	for rid, pats in config.rules_and_processes.items():
		if rid.find('Transform') != -1:
			# a transformation rule to process
			lpattern = extractPattern(pats.left)
			rpattern = extractPattern(pats.right)
			# add the rule to the rule system in memory
			# THIRD ELEMENT IS NOW AN EMPTY SET! PLACEHOLDER FOR APPLICATION SCOPES!!!
			RS.append((lpattern, rpattern, set([])))
		else:
			# an add rule to read
			lpattern = [['1', '0', '1', '1'], {}, set([])]
			rpattern = extractPattern(pats)
			# add the rule to the rule system in memory
			# application scope is chosen later to add appropriate number of new processes (w.r.t. new synch rules)
			RS_add.append((lpattern, rpattern, set([])))
	# per rule, store the action prefixes of left and right patterns in sets
	for i in range(0,len(RS)):
		RS_actprefixsets.append([set([]),set([])])
		for a in RS[i][0][2]:
			RS_actprefixsets[i][0].add(actionStripParams(a))
		for a in RS[i][1][2]:
			RS_actprefixsets[i][1].add(actionStripParams(a))
	for i in range(0,len(RS_add)):
		RS_actprefixsets.append([set([]),set([])])
		for a in RS_add[i][1][2]:
			RS_actprefixsets[i+len(RS)][1].add(actionStripParams(a))

	# read the remainder of the input, containing sync rules
	l = []
	input = 'synch left\n' + config.existing_synchronisation_rules + 'synch right\n' + config.new_synchronisation_rules
	lines = input.split('\n')
	for lx2 in range(0,len(lines)):
		results, remainder = syncrulescanner.scan(lines[lx2])
		l += results + [('NEWLINE',0)]
	index = 0
	section = l[index][1]
	index += 2
	process_rules = False

	# process the new syncrules
	# as we do this, we keep track of how many elements we will need to have for the verification of the rule system
	# (depends on the syncrules)
	required_rulevector_elements = Counter()
	for i in range(0,len(RS)):
		required_rulevector_elements[i] += 1
	for i in range(0,len(RS_add)):
		required_rulevector_elements[len(RS)+i] += 1
	tmp_rs_syncrules = set([])
	for j in [0,1]:
		acttmplist = []
		tmpsynclist = [[]]
		tmpsynccounter = 0
		while index < len(l):
			skipline = False
			if l[index][0] == 'NEWLINE':
				skipline = True
				index += 1
			else:
				if l[index][0] == 'SECTIONSTART':
					index += 1
					break
				rules_seen = copy.deepcopy(required_rulevector_elements)
				if l[index][0] == 'BRACKET':
					vect = '(' + l[index][1]
					index += 1
					while l[index][0] != 'BRACKET':
						act = l[index][1]
						vect += act
						# check rules_seen
						rids = RS_getRulesWithAction(act, section==3)
						for rid in rids:
							if rules_seen[rid] > 0:
								rules_seen[rid] -= 1
							else:
								required_rulevector_elements[rid] += 1
						if l[index+1][0] != 'BRACKET':
							vect += ','
						index += 1
					vect += l[index][1] + ',' + l[index+1][1] + ')'
					index += 2
				else:
					# we have a rule of type a #n OR a rule in the format a b c -> d
					act = l[index][1]
					index += 1
					arity_rule = False
					if index == len(l):
						arity = 2
						arity_rule = True
					else:
						if l[index][0] == 'ARITY':
							arity = int(l[index][1][1:])
							arity_rule = True
						else:
							if l[index][0] == 'NEWLINE':
								arity = 2
								arity_rule = True
							else:
								# format a b c -> d
								vect = '({' + act
								# check rules_seen
								rids = RS_getRulesWithAction(act, section==3)
								for rid in rids:
									if rules_seen[rid] > 0:
										rules_seen[rid] -= 1
									else:
										required_rulevector_elements[rid] += 1
								while l[index][0] != 'TO':
									act = l[index][1]
									vect += ',' + act
									rids = RS_getRulesWithAction(act, section==3)
									for rid in rids:
										if rules_seen[rid] > 0:
											rules_seen[rid] -= 1
										else:
											required_rulevector_elements[rid] += 1
									index += 1
								index += 1
								# add result of rule
								vect += '},' + l[index][1] + ')'
								index += 1
					if arity_rule:
						rids = RS_getRulesWithAction(act, section==3)
						if rids != set([]):
							for rid in rids:
								if rules_seen[rid] >= arity:
									rules_seen[rid] -= arity
								else:
									required_rulevector_elements[rid] += (arity-rules_seen[rid])
									rules_seen[rid] = 0
						vect = act + ' #' + str(arity)
						index += 1
			if not skipline:
				parseSyncRule(vect)
		# place the parsed syncrules in temporary set of rules
		for (type, acts, an) in tmpsynclist[tmpsynccounter]:
			tmp_rs_syncrules.add((section==3, type, acts, an))
		if section == 2:
			section = 3
	# all syncrules have been obtained. rewrite them to required vectors
	# construct RS_vector
	for i in range(0,len(RS)):
		for j in range(0,required_rulevector_elements[i]):
			RS_vector.append(i)
	for i in range(0,len(RS_add)):
		for j in range(0,required_rulevector_elements[len(RS)+i]):
			RS_vector.append(len(RS)+i)

	RS_left_syncrules = []
	RS_right_syncrules = []
	for (side, type, acts, an) in tmp_rs_syncrules:
		newrules = RS_getNewRules(side, type, acts, an)
		if not side:
			RS_left_syncrules.append(newrules)
		else:
			RS_right_syncrules.append(newrules)

	# if the interface of a syncrule is non-empty, we rename the action resulting from that rule to make it unique
	counter = 0
	tmp = []
	for (vects, rltact, interface) in RS_left_syncrules:
		if len(list(interface.elements())) > 0:
			tmp.append((vects, rltact + "_itf" + str(counter), interface))
			counter += 1
		else:
			tmp.append((vects, rltact, interface))
	RS_left_syncrules = tmp
	tmp = []
	for (vects, rltact, interface) in RS_right_syncrules:
		if len(list(interface.elements())) > 0:
			tmp.append((vects, rltact + "_itf" + str(counter), interface))
			counter += 1
		else:
			tmp.append((vects, rltact, interface))

	tmp_rs_syncrules = storeRulesInDict(RS_left_syncrules)
	RS_left_syncactions = buildSyncactionsets(tmp_rs_syncrules, len(RS_vector))

	# order the rules in a list of dictionaries for the generation of LTSs of right hand rule patterns
	RS_gen_left_syncrules = []
	for i in range(0,len(RS_vector)):
		RS_gen_left_syncrules.append({})
	RS_gen_left_syncrules = orderSyncrules(tmp_rs_syncrules, RS_gen_left_syncrules)

	tmp_rs_syncrules = storeRulesInDict(RS_right_syncrules)
	RS_right_syncactions = buildSyncactionsets(tmp_rs_syncrules, len(RS_vector))

	RS_gen_right_syncrules = copy.deepcopy(RS_gen_left_syncrules)
	RS_gen_right_syncrules = orderSyncrules(tmp_rs_syncrules, RS_gen_right_syncrules)

	# CHECK MODEL INDEPENDENT WELL-FORMEDNESS OF SYNC RULES
	# Check if existing sync actions occurring in right patterns also appear in corresponding left ones
	for i in range(0,len(RS)):
		for a in RS[i][1][2]:
			if mustSync(a, i, RS_left_syncactions):
				# does the same action also exist in left pattern?
				if a not in RS[i][0][2]:
					print_error("well-formedness violated! Transformation rule " + str(i) + " uses an existing sync action in the right pattern that is not in the left pattern!")
	for i in range(0,len(RS_add)):
		for a in RS_add[i][1][2]:
			if mustSync(a, i, RS_left_syncactions):
				print_error("well-formedness violated! Add rule " + str(i) + " uses existing sync action " + str(a) + " in its pattern!")

	for i in range(0,len(RS_vector)):
		RS_right_syncactions[i] |= RS_left_syncactions[i]

	# Possibly extend the network with positions for newly added processes
	if config.network_name != "":
		# convert the input network to the new size
		# add application scopes to the add rules and merge them with RS
		current = len(procs)
		inputnetwork += [i for i in range(len(procs),len(procs)+len(RS_add))]
		RS_add_tmp = []
		for i in range(0,len(RS_add)):
			RS_add_tmp.append((RS_add[i][0], RS_add[i][1], set([current+j for j in range(0,required_rulevector_elements[len(RS)+i])])))
			current += required_rulevector_elements[len(RS)+i]
		RS_add = RS_add_tmp
		for i in range(0,len(RS_add)):
			procsheaders.append(['0','0','1'])
			procsactsets.append(set([]))
			procsactprefixsets.append(set([]))
			syncactions.append(set([]))
			procs.append({})
		# extend the syncrules
		extendedrules = []
		for rsi in range(0,len(syncrules)):
			extendedrules.append({})
			for a, v in syncrules[rsi].items():
				out = []
				for (vect,an) in v:
					out.append((extendVector(vect, 0, inputnetwork),an))
				extendedrules[rsi][a] = out
		for rsi in range(0,len(RS_add)):
			extendedrules.append({})
		syncrules = extendedrules

def readRulesystem(filename):
	"""Read a rule system, which contains pairs of LTSs, newly to be added LTSs, and synchronisation rules"""
	try:
		inFile = open(filename + ".rs",'r')
	except IOError:
		print_error("rule system file does not exist!")
	
	lines = inFile.readlines()
	l = []
	for lx in range(0,len(lines)):
		if lines[lx].find('synch left') != -1 or lines[lx].find('synch right') != -1:
			break
		results, remainder = RSscanner.scan(lines[lx])
		l += results
	index = 0
	section = 0
	# counter to id rules and processes
	idcounter = 0
	while index < len(l):
		if l[index][0] == 'SECTIONSTART':
			section = l[index][1]
			index += 1
		else:
			if section == 0:
				# a transformation rule to read
				index, lProcess = readProcess(l, index)
				if l[index][0] != 'TO':
					print_error("unexpected format in rule system file!")
				index += 1
				index, rProcess = readProcess(l, index)
				# add the rule to the rule system in memory
				r = Rule()
				r.setPatterns(lProcess, rProcess)
				config.rules_and_processes["Transformation " + str(idcounter)] = r
				idcounter += 1
			if section == 1:
				# an add rule to read (i.e. process)
				index, rProcess = readProcess(l, index)
				# add the rule to the rule system in memory
				config.rules_and_processes["Process " + str(idcounter)] = rProcess
				idcounter += 1
	# read the remainder of the input, containing sync rules
	l = []
	gui_syncrules = ["",""]
	mode = 0
	for lx2 in range(lx,len(lines)):
		if lines[lx2] == 'synch left\n':
			mode = 0
		elif lines[lx2] == 'synch right\n':
			mode = 1
		else:
			gui_syncrules[mode] += lines[lx2]
	config.existing_synchronisation_rules = gui_syncrules[0]
	config.new_synchronisation_rules = gui_syncrules[1]

# END I/O FUNCTIONS

# START CONSTRUCT HIDING SET FROM MCL PROPERTY

scanner=re.Scanner([
	(r"[\'\"][A-Za-z\s0-9!.\*\(\),]+[\'\"]", lambda scanner,token:("ACTION", token)),
	(r"true|false",                          lambda scanner,token:("CONST", token)),
	(r"not|or|and|mu|nu|[.\?\*\@]|-\|",      lambda scanner,token:("OPERATOR", token)),
	(r"[A-Z]",                               lambda scanner,token:("VARIABLE", token)),
	(r"[\(\[\<]",                            lambda scanner,token:("OPENBR", token)),
	(r"[\)\]\>]",                            lambda scanner,token:("CLOSBR", token)),
	(r"\s+", None), # None == skip token.
])

def ActSetsUnion(sign1, set1, sign2, set2):
	"""Obtain the union of the two sets"""
	if sign1 == '+':
		if sign2 == '+':
			return ['+', set1 | set2]
		else:
			return ['-', set2 - set1]
	else:
		if sign2 == '+':
			return ['-', set1 - set2]
		else:
			return ['-', set1 & set2]

def combineActSets(l, r, op):
	"""Combine the given actsets"""
	#print 'combine ' + str(l) + ' ' + str(r) + ' with ' + str(op)
	if op == '':
		if l[0] == False:
			if r[0] == False:
				return l
			else:
				return r
		else:
			if r[0] == False:
				return l
			else:
				result = ActSetsUnion(l[1], l[2], r[1], r[2])
				return [True, result[0], result[1]]
	else:
		if op == 'or' or op == 'and':
			if l[0] == False:
				if r[0] == False:
					result = ActSetsUnion(l[1], l[2], r[1], r[2])
					return [False, result[0], result[1]]
				else:
					return r
			else:
				if r[0] == False:
					return l
				else:
					result = ActSetsUnion(l[1], l[2], r[1], r[2])
					return [True, result[0], result[1]]
		else:
			# op == '?'
			if l[0] == False:
				return r
			else:
				result = ActSetsUnion(l[1], l[2], r[1], r[2])
				return [True, result[0], result[1]]

def getNonHidingSet(l,mode):
	"""Extract non-hiding set from given list starting at position pos"""
	# if we encounter an opening bracket, do a recursive call for that subterm
	# break the given list l down into subterms
	subl = []
	i = 0
	while i < len(l):
		if l[i][0] == 'OPENBR':
			passl = []
			if l[i][1] == '[' or l[i][1] == '<':
				submode = True
			else:
				submode = False
			brc = 1
			while brc > 0:
				i += 1
				if l[i][0] == 'OPENBR':
					brc += 1
				else:
					if l[i][0] == 'CLOSBR':
						brc -= 1
				if brc > 0:
					passl.append(l[i])
			subl.append([submode,'nest',passl])
		else:
			while l[i][0] != 'OPENBR':
				subl.append([False,'leaf',l[i]])
				i += 1
				if i == len(l):
					break
			i -= 1
		i += 1
	# for each subterm in subl, get the result
	results = []
	for s in subl:
		if s[1] == 'nest':
			results.append(getNonHidingSet(s[2],s[0]))
		else:
			if s[2][0] == 'CONST' or s[2][0] == 'VARIABLE':
				results.append([False,'+',set([])])
			if s[2][0] == 'OPERATOR':
				results.append([False,'op',s[2][1]])
			else:
				if s[2][0] == 'ACTION':
					if s[2][1] == 'tau':
						results.append([False,'-',set(['tau'])])
					else:
						results.append([False,'+',set([s[2][1]])])
	# combine the subresults
	#print results
	tmpresult = []
	tmpop = ''
	for i in range(0,len(results)):
		if results[i][1] == 'op':
			if results[i][2] == 'or' or results[i][2] == 'and' or results[i][2] == '?' or results[i][2] == '.':
				tmpop = results[i][2]
		else:
			if tmpresult == []:
				tmpresult = results[i]
			else:
				tmpresult = combineActSets(tmpresult, results[i], tmpop)
				tmpop = ''
	if mode:
		return [True, tmpresult[1], tmpresult[2]]
	else:
		return tmpresult

def constructHidingSet():
	"""Read the property in propertyfile and construct hiding set"""
	global hideactions
	try:
		propertyfile = open(os.path.join(config.property_folder, config.property_name + ".mcl"), 'r')
	except IOError:
		print_error("property file does not exist!")

	lines = propertyfile.readlines()
	prop = ""
	for line in lines:
		prop += line.rstrip()
	global token, next
	results, remainder = scanner.scan(prop)
	nhset = getNonHidingSet(results,False)
	#print nhset
	if not nhset[0]:
		hideactions.append('all')
	else:
		if nhset[1] == '+':
			hideactions.append('all')
			if nhset[2] != set([]):
				hideactions.append('but')
				for a in nhset[2]:
					hideactions.append(a)
		else:
			if nhset[2] == set([]):
				hideactions.append('none')
			else:
				for a in nhset[2]:
					hideactions.append(a)
	writeHideActions()

# END CONSTRUCT HIDING SET FROM MCL PROPERTY

def getSubsystem(i, l):
	"""Get the subsystem at position i in l"""
	if hasattr(l[i], "__iter__") and not isinstance(l[i], str):
		return l[i]
	else:
		return [l[i]]

def actionStripInterfacecounter(action):
	"""Strip the interface counter off the given action"""
	index = action.find('_itf')
	if index == -1:
		return action
	else:
		return action[:-(len(action)-index)]

def actionStripParams(action):
	"""Strip the parameters and quotes off the action string"""
	# We strip both LOTOS style and mCRL/mCRL2 style, i.e. everything following
	# '(' and ' !', respectively
	l1 = action.split('!')
	l2 = l1[0].split('#')
	l1 = l2[0].split('(')
	#newa = l1[0].rstrip(' ')
	newa = l1[0].replace('"','')
	return newa

def actionStripPlaceholder(action):
	"""Strip the parameters and quotes off the action string"""
	# We strip placeholders
	l1 = action.split('#')
	#newa = l1[0].rstrip(' ')
	return newa.replace('"','')

def actionGetParams(action):
	"""Strip the action off the params"""
	params = ''
	l1 = action.split('!')
	if len(l1) > 1:
		for i in range(1,len(l1)):
			params += '!' + l1[i]
	else:
		l1 = action.split('#')
		if len(l1) > 1:
			params = '#' + l1[1]
		else:
			l1 = action.split('(')
			if len(l1) > 1:
				for i in range(1,len(l1)):
					params += '(' + l1[i]
	params = params.rstrip('"')
	return params

def actionGetPlaceholder(action):
	"""Return the placeholder in action"""
	params = ''
	l1 = action.split('#')
	if len(l1) > 1:
		params = '#' + l1[1]
	params = params.rstrip('"')
	return params

def actionMatches(action1, action2, placeholderdict):
	"""Does action1 match action2? Comparison is sensitive for placeholders in action1"""
	prefix = actionStripPlaceholder(action1)
	prefix2 = action2[:-(len(action2)-len(prefix))+1].replace('"','')
	if prefix[0] == '\'':
		areg = re.compile('"' + prefix + '"')
		if areg.match(prefix2) != None:
			prefixesEqual = True
		else:
			prefixesEqual = False
	else:
		prefixesEqual = (prefix == prefix2)
	if not prefixesEqual:
		return False
	params = actionGetPlaceholder(action1)
	if params != '':
		# try to obtain the placeholder value from placeholderdict
		value = placeholderdict.get(params)
		if value != None:
			return (prefix1 + value == action2)
		else:
			return True
	else:
		return actionGetParams(action2) == ''

def actionGetPlaceholdervalue(action1, action2):
	"""PRECONDITION: action1 matches action2. Return, if present, the value for the placeholder in action1"""
	params = actionGetPlaceholder(action1)
	prefix = actionStripPlaceholder(action1)
	#index = action1.find(params)
	if params == '':
		return ''
	else:
		# get the placeholder value
		#return action2[index:].rstrip('"')
		return action2[len(prefix)+1:].rstrip('"')

def actionRename(action, placeholderdict):
	"""Renames the action by replacing its placeholder, if present, by the concrete value in the given dict"""
	prefix = actionStripPlaceholder(action)
	if prefix == action.replace('"',''):
		return action
	params = actionGetPlaceholder(action)
	value = placeholderdict.get(params)
	if value == None:
		return action
	return '"' + prefix + value + '"'

def actionsetGetMatching(aset1, aset2):
	"""Return subset of actions in aset1 which have a match in aset2"""
	newset = set([])
	for a1 in aset1:
		for a2 in aset2:
			rlt = actionMatches(a1, a2, {})
			if rlt:
				newset.add(a1)
				break
	return newset

def getMatchedActions(a, statemap, i):
	"""get the set of concrete actions on which action a in transformation rule i matches; interprets placeholders"""
	global RS
	rset = set([])
	t = a.split('#')
	if len(t) > 1:
		for m in statemap:
			ac = actionRename(a, m[1])
			rset.add(ac)
	else:
		rset.add(a)
	return rset

def flatten(x):
	result = []
	if hasattr(x, "__iter__") and not isinstance(x, str):
		for el in x:
			if hasattr(el, "__iter__") and not isinstance(el, str):
				result.extend(flatten(el))
			else:
				result.append(el)
	else:
		result.append(x)
	return result

def extendVector(lst, pos, structure):
	"""Extend the given list (vector) with *'s"""
	newlst = []
	for i in range(0,pos):
		for j in range(0,len(flatten(structure[i]))):
			newlst.append('*')
	if not isinstance(lst, str):
		for i in range(0,len(lst)):
			newlst.append(lst[i])
	else:
		newlst.append(lst)
	for i in range(len(newlst),len(flatten(structure))):
		newlst.append('*')
	return newlst

def produceListOfVectors(a, struc, blockedsets):
	"""Produce a list of vectors containing only action a and *'s"""
	if hasattr(struc, "__iter__") and not isinstance(struc, str):
		rs = flatten(struc)
		result = []
		result.append(a)
		for i in range(1,len(rs)):
			result.append('*')
		#endresult = list(set(itertools.permutations(result)))
		endresult = list(perm_unique(result))
		# check the results for invalid ones
		newendresult = copy.deepcopy(endresult)
		for r in endresult:
			for i in range(0,len(r)):
				if r[i] == a:
					rlt = blockedsets.get(str(rs[i]))
					if a in rlt:
						newendresult.remove(r)
		endresult = newendresult
	else:
		rlt = blockedsets.get(str(struc))
		endresult = []
		if a not in rlt:
			endresult.append(a)
			endresult = [tuple(endresult)]
	return endresult

def storeProcsactprefixsets():
	"""derive the prefix sets from the action sets"""
	global procsactprefixsets, procs, procsactsets
	procsactprefixsets = []
	for i in range(0,len(procs)):
		procsactprefixsets.append(set([]))
		for a in procsactsets[i]:
			a_stripped = actionStripParams(a)
			procsactprefixsets[i].add(a_stripped)

def rewriteAllSync(structure):
	"""Flatten the sync rules"""
	global inputsync, procsactprefixsets
	# listofsubrules contains pairs of dicts (sync, block), with sync a dict of sync rules, and block a dict of actions to be blocked
	listofsubrules = []
	resultdict = {}
	subblockedsets = {}
	for i in range(0,len(procs)):
		subblockedsets[str(i)] = set([])
	if hasattr(structure, "__iter__") and not isinstance(structure, str):
		structlen = len(structure)
		for i in range(0,len(structure)):
			(r1, r2) = rewriteAllSync(structure[i])
			listofsubrules.append(r1)
			for k, v in r2.items():
				#print "addsub: " + str(k)
				subblockedsets[str(k)] = v
	else:
		structlen = 1
	# the blockedinsub sets serve to keep track of actions not to be performed by subsystems through syncs
	blockedinsub = []
	for i in range(0,structlen):
		blockedinsub.append(set([]))
	blockedsets = {}
	for i in range(0,len(procs)):
		blockedsets[str(i)] = set([])
	# process the rules at this stage
	rules = inputsync.get(toTupleOfTuples(structure))
	# set of actions resulting from syncs in structure
	syncresult_acts = set([])
	if rules != None:
		# first construct a dict with the actions to be blocked
		blocked = {}
		for (vects, actres) in rules:
			syncresult_acts.add(actres)
			for t in vects:
				# block all actions in the vector: add 1 to blocked
				minrange = min(structlen, len(t))
				for i in range(0,minrange):
					if t[i] != '*':
						rs = flatten(structure[i])
						for ind in rs:
							rlt = blocked.get(t[i])
							if rlt == None:
								if t[i] == actres:
									blocked[t[i]] = {ind: 0}
								else:
									blocked[t[i]] = {ind: 1}
							else:
								rltindex = rlt.get(ind)
								if rltindex == None:
									if t[i] == actres:
										rlt[ind] = 0
									else:
										rlt[ind] = 1
								elif rltindex != 0:
									if t[i] == actres:
										rlt[ind] = 0
								blocked[t[i]] = rlt
		# convert blocked to set
		for k, v in blocked.items():
			for id, n in v.items():
				if n == 1:
					(blockedsets[str(id)]).add(k)
		for (vects, actres) in rules:
				rlt = resultdict.get(actres)
				if rlt == None:
					resultdict[actres] = []
				for t in vects:
					if len(t) == structlen:
						lst = []
						for i in range(0,len(t)):
							rlt = (listofsubrules[i]).get(t[i])
							if rlt != None:
								lst.append(rlt)
								(blockedinsub[i]).add(t[i])
							else:
								newlsts = produceListOfVectors(t[i], structure[i], subblockedsets)
								filtered_newlsts = []
								flstruc = flatten(structure[i])
								for l in newlsts:
									store = True
									for i in range(0,len(flstruc)):
										if l[i] != '*':
											#print l[i]
											#print procsactprefixsets[int(flstruc[i])]
											if not l[i] in procsactprefixsets[int(flstruc[i])]:
												store = False
												break
									if store:
										filtered_newlsts.append(l)
								lst.append(filtered_newlsts)
						# get all combinations
						alllsts = list(itertools.product(*lst))
						alllsts = list(tuple(flatten(x)) for x in alllsts)
						# place them in resultdict
						resultdict[actres] += alllsts
	# next, pass all rules from subsystems to the next level, if not blockedinsub
	# produce set of all actions in substructures resulting from syncs
	subsyncresult_acts = set([])
	for i in range(0,len(listofsubrules)):
		for ar, v in (listofsubrules[i]).items():
			subsyncresult_acts.add(ar)
	#add 'single process' rule for syncresult actions if component i does not have a rule for it yet
	for i in range(0,len(listofsubrules)):
		for ar in subsyncresult_acts:
			if ar not in syncresult_acts:
				rlt = (listofsubrules[i]).get(ar)
				if rlt == None:
					if ar not in blockedinsub[i]:
						newlsts = produceListOfVectors(ar, structure[i], subblockedsets)
						filtered_newlsts = []
						flstruc = flatten(structure[i])
						for l in newlsts:
							store = True
							for j in range(0,len(flstruc)):
								if l[j] != '*':
									if not l[j] in procsactprefixsets[int(flstruc[j])]:
										store = False
										break
							if store:
								filtered_newlsts.append(l)
						listofsubrules[i][ar] = filtered_newlsts
	for i in range(0,len(listofsubrules)):
		for ar, v in (listofsubrules[i]).items():
			if ar not in blockedinsub[i]:
				rlt = resultdict.get(ar)
				if rlt == None:
					rlt = []
				resultdict[ar] = rlt + [tuple(extendVector(x, i, structure)) for x in v]
	# produce final blockedsets
	for id in flatten(structure):
		rlt = blockedsets.get(str(id))
		rlt2 = subblockedsets.get(str(id))
		blockedsets[str(id)] = rlt | rlt2
	return (resultdict, blockedsets)

def storeRulesInDict(rules):
	"""Put the list of rules in a dictionary sorted by resulting action"""
	result = {}
	for r in rules:
		rlt = result.get(r[1])
		if rlt == None:
			rlt = []
		for vect in r[0]:
			rlt.append(vect)
		if rlt != []:
			result[r[1]] = rlt
	return result

def countVectorElements(rule):
	"""Count the number of non-* elements in given vector"""
	result = 0
	for x in rule:
		if x != '*':
			result += 1
	return result

def buildSyncactionsets(rules, size):
	"""Build the sets of syncactions per vector element (process or rule, depending on size)"""
	syncactions = []
	# determine size of each vector
	for i in range(0,size):
		syncactions.append(set([]))
	for ac, v in rules.items():
		if v == []:
			for i in range(0,size):
				(syncactions[i]).add(ac)
		else:
			for vec in v:
				for i in range(0,len(vec)):
					if vec[i] != '*':
						(syncactions[i]).add(vec[i])
	return syncactions

def orderSyncrules(rules, ruledict):
	"""Order the rules in dictionaries for LTS generation"""
	# determine size of each vector
	length = 0
	for ac, v in rules.items():
		for vec in v:
			length = len(vec)
			if length > 0:
				break
		if length > 0:
			break
	for i in range(0,length):
		for ac, v in rules.items():
			newv = copy.deepcopy(v)
			for vec in v:
				lvec = list(vec)
				if vec[i] != '*':
					rlt = (ruledict[i]).get(vec[i])
					if rlt == None:
						(ruledict[i])[vec[i]] = [(lvec, ac)]
					else:
						if (lvec, ac) not in rlt:
							rlt.append((lvec, ac))
							(ruledict[i])[vec[i]] = rlt
					newv.remove(vec)
			rules[ac] = newv
	return ruledict

def getRevTrans(trans):
	"""store the reversed LTS"""
	revtrans = {}
	for k, v in trans.items():
		for a, t in v.items():
			for tgt in t:
				# reversed procs
				incoming = revtrans.get(tgt)
				if incoming == None:
					revtrans[tgt] = {a: set([k])}
				else:
					incomingwithlabel = incoming.get(a)
					if incomingwithlabel == None:
						incoming[a] = set([k])
					else:
						if k not in incomingwithlabel:
							incomingwithlabel.add(k)
						incoming[a] = incomingwithlabel
					revtrans[tgt] = incoming
	return revtrans

def getActTrans(trans):
	"""store the LTS sorted by transition labels"""
	atrans = {}
	for k, v in trans.items():
		for a, t in v.items():
			for tgt in t:
				# action procs
				tr = atrans.get(a)
				if tr == None:
					atrans[a] = {k: set([tgt])}
				else:
					trwithk = tr.get(k)
					if trwithk == None:
						tr[k] = set([tgt])
					else:
						if tgt not in trwithk:
							trwithk.add(tgt)
							tr[k] = trwithk
					atrans[a] = tr
	return atrans

# BEGIN functions for transformation rules / performing transformations on a network

def checkConfluence():
	"""Check confluence and termination of the rulesystem; simple check on action sets"""
	# for confluence, we check that the left hand action sets of the rules are disjoint
	# NOTE: this is check 1 for confluence. Check 2 is done when matches are determined.
	# For this, see function checkMatchesIntersect()
	#leftunion = set([])
	for i in range(0,len(RS)):
		#leftunion = leftunion | ((RS[i])[0])[2]
		for j in range(i+1,len(RS)):
			if len(((RS[i])[0])[2] & ((RS[j])[0])[2]) > 0:
				print_message(str(RS[i][0][2]))
				print_message(str(RS[j][0][2]))
				print_error("rules " + str(RS[i]) + "and " + str(RS[j]) + " are not confluent!")

def checkMatchesIntersect(rule, mapinfo, statemap):
	"""Perform check 2 for confluence: no two matches of a rule intersect"""
	# Check that no two states match on the same state
	# This is ok if the two states relate to a glue state, and
	# they have no conflicting transitions, i.e. they do not have
	# an incoming or outgoing transition in common
	intersected = set([])
	set1union = set([])
	set2union = set([])
	# function mapping process states to match ids. TD required to determine that no transitions are shared
	map_proc_to_mtch = {}	
	# get reversed LTS
	revtrans = getRevTrans(rule[0][1])
	for i in range(0,int(rule[0][0][2])):
		intertmp = set1union & mapinfo.get(str(i))
		intersected = intersected | intertmp
		set1union = set1union | mapinfo.get(str(i))
	# Now we have in intersected the set of process states that are matched on multiple times
	# Check per member if we ever have an overlap in ingoing and outgoing transitions
	set1union = set([])
	for s in intersected:
		for i in range(int(rule[0][0][0]),int(rule[0][0][2])):
			# TO FIX!
			if s in mapinfo.get(str(i)):
				# overlap?
				transset = rule[0][1].get(s)
				if transset == None:
					transset = {}
				for a, v in transset.items():
					if a in set1union:
						print_error("ERROR: Intersecting matches detected. Rule " + str(rule) + " is not confluent!")
					set1union.add(a)
				transset = revtrans.get(s)
				if transset == None:
					transset = {}
				for a, v in transset.items():
					if a in set2union:
						print_error("Intersecting matches detected. Rule " + str(rule) + " is not confluent!")
					set2union.add(a)

def getNumberOfOutgoing(trans):
	"""Retrieve the number of transitions in trans"""
	number = 0
	if trans != None:
		for (a,v) in trans:
			number += len(v)
	return number
			
def transformProcess(id):
	"""transform the given process according to the established matches"""
	global procs, procsheaders, transformed_procs, transformed_procsheaders, transformed_procs_new_states, procndivsets, RS, RS_add, procsrulematches
	transformed_procs[id] = copy.deepcopy(procs[id])
	transformed_procsheaders[id] = (procsheaders[id][0], procsheaders[id][1], procsheaders[id][2])
	nstates = int((procsheaders[id])[2])
	statecounter = nstates
	ntrans = int((procsheaders[id])[1])
	deleted = deque([])
	newmtch = {}
	transformed = False
	# in this list, we store for each process a mapping for each rule from new states to process states (right hand patterns)
	newstatesmaps = []
	count = 0
	for rid in range(0,len(procsrulematches[id])):
		mapdict = {}
		for mtch in procsrulematches[id][rid]:
			if rid >= len(RS):
				index = rid - len(RS)
				RSYS = RS_add
			else:
				index = rid
				RSYS = RS
			if mtch[0] != {}:
				count += 1
				if not transformed:
					print_message("refining process " + str(id))
				transformed = True
				# update nr of trans
				ntrans += int(RSYS[index][1][0][1]) - int(RSYS[index][0][0][1])
				# delete states
				for i in range(int(((RSYS[index][0])[0])[0]), int(((RSYS[index][0])[0])[2])):
					smap = mtch[0].get(str(i))
					nstates -= 1
					deleted.append(smap)
					(transformed_procs[id]).pop(smap)
				# add new states
				for i in range(int(((RSYS[index][1])[0])[0]), int(((RSYS[index][1])[0])[2])):
					if len(deleted) > 1:
						newmtch[str(i)] = deleted.popleft()
					else:
						newmtch[str(i)] = str(statecounter)
						statecounter += 1
					nstates += 1
					rlt = mapdict.get(str(i))
					if rlt == None:
						rlt = set([])
					rlt.add(newmtch[str(i)])
					mapdict[str(i)] = rlt
				for i in range(int(((RSYS[index][1])[0])[0]), int(((RSYS[index][1])[0])[2])):
					smap = newmtch[str(i)]
					newtrans = {}
					trans = ((RSYS[index][1])[1]).get(str(i))
					if trans != None:
						for a, v in trans.items():
							newv = set([])
							for tgt in v:
								if int(tgt) < int(RSYS[index][1][0][0]):
									newv.add(mtch[0].get(tgt))
								else:
									newv.add(newmtch.get(tgt))
							concrete_a = actionRename(a, mtch[1])
							newtrans[concrete_a] = newv
							(transformed_procs[id])[smap] = newtrans
				# apply changes to glue state transitions
				for i in range(0, int(((RSYS[index][0])[0])[0])):
					smap = mtch[0].get(str(i))
					out = (transformed_procs[id]).get(smap)
					outpattern = ((RSYS[index][0])[1]).get(str(i))
					if out == None:
						out = {}
					if outpattern == None:
						outpattern = {}
					# remove the transitions in the left hand pattern
					for a, v in outpattern.items():
						# remove the equivalent transitions in the process LTS
						concrete_a = actionRename(a, mtch[1])
						outwithlabel = out.get(concrete_a)
						for tgt in v:
							tmap = mtch[0].get(tgt)
							outwithlabel.remove(tmap)
						out[concrete_a] = outwithlabel
					# add the transitions in the right hand pattern
					outpattern = ((RSYS[index][1])[1]).get(str(i))
					if outpattern != None:
						for a, v in outpattern.items():
							concrete_a = actionRename(a, mtch[1])
							outwithlabel = out.get(concrete_a)
							if outwithlabel == None:
								outwithlabel = set([])
							newv = set([])
							for tgt in v:
								if int(tgt) < int(((RSYS[index][1])[0])[0]):
									tmap = mtch[0].get(tgt)
								else:
									tmap = newmtch.get(tgt)
								newv.add(tmap)
							out[concrete_a] = outwithlabel | newv
					(transformed_procs[id])[smap] = out
		newstatesmaps.append(mapdict)
	# all rules have been applied; shift the state indices and rewrite to remove gaps
	# also shift state indices in procndivsets[i] and remove deleted indices
	if transformed:
		imap = {}
		cnt = 0
		skip = '-1'
		newndivs = set([])
		for i in range(0,statecounter):
			if deleted != deque([]) and skip == '-1':
				skip = deleted.popleft()
			if str(i) != skip:
				imap[str(i)] = str(cnt)
				if opt_divergency:
					if str(i) in procndivsets[id]:
						newndivs.add(str(cnt))
				cnt += 1
			else:
				skip = '-1'
		if opt_divergency:
			procndivsets[id] = newndivs
		# use new indexing in procs structure
		nprocs = {}
		for src, v in transformed_procs[id].items():
			nprocs[imap[src]] = {}
			for a, t in v.items():
				(nprocs[imap[src]])[a] = set([])
				for tgt in t:
					((nprocs[imap[src]])[a]).add(imap[tgt])
		transformed_procs[id] = nprocs
		# store the new header for the LTS
		transformed_procsheaders[id] = ('0', str(ntrans), str(nstates))
		# use new indexing in newstatesmaps
		for i in range(0,len(procsrulematches[id])):
			nmaps = {}
			for src, v in newstatesmaps[i].items():
				nmaps[src] = set([])
				for sm in v:
					nmaps[src].add(imap[sm])
			newstatesmaps[i] = nmaps
		# store globally
		transformed_procs_new_states.append(newstatesmaps)

def transform():
	global procs, transformed_procs, transformed_procsheaders, procs_orig_size
	for i in range(0,len(procs)):
		transformed_procs.append({})
		transformed_procsheaders.append(())
	for i in range(0,len(procs)):
		transformProcess(i)

def getMatchesForProcess(processid):
	"""get the matches for the given process LTS according to the rule system"""
	global RS, RS_add, procsrulestatemaps, procsrulematches, rulematcheson
	rulestatemap = []
	print_message("determining matches in process " + str(processid))
	for rid in range(0,len(RS)):
		# ACTUALLY ADD?
		# determine that the set of sync actions of left pattern of rule rid is equal to the intersection of the set of sync actions
		# of the process LTS and set of actions of the left pattern. In other words, we only match a rule on a process iff the syncactions
		# in the pattern are actually also sync actions in the process.
		statemap = []
		if RS[rid][2] == set([]) or processid in RS[rid][2]:
			statemap = getAllMatches(processid, RS[rid])
			# the set of deleted states contains all states with an index >= first number in headers of rule
			statemap = removeDanglingMatches(processid, RS[rid], statemap)
			statemap = removeExclusiveglueViolations(processid, RS[rid], statemap)
		rulestatemap.append(statemap)
	# add special entries for add rules
	for rid in range(0,len(RS_add)):
		if processid in RS_add[rid][2]:
			rulestatemap.append([[{'0': '0'},{}]])
		else:
			rulestatemap.append([])
	# procsrulestatemaps is used to keep for all processes and all rules the mappings from left pattern states to process states
	# This is used later when updating the non-divergency sets (divergency of process states must be carried over to rule states)
	# procsrulematches is a list of matches for each process, each rule
	procsrulematches.append(rulestatemap)
	tmp = []
	for rid in range(0,len(RS)):
		mapinfo = groupMatches(rulestatemap[rid])
		# perform check 2 for confluence check: no two matches intersect
		if mapinfo != {}:
			checkMatchesIntersect(RS[rid], mapinfo, rulestatemap[rid])
		tmp.append(mapinfo)
	procsrulestatemaps.append(tmp)

def matchRulesystem():
	"""Determine all matches of rules on processes"""
	global procs, actprocs, revprocs, procsrulestatemaps, procsrulematches, rulematcheson, RS_add
	actprocs = []
	revprocs = []
	for i in range(0,len(procs)):
		rlt = getActTrans(procs[i])
		actprocs.append(rlt)
		rlt = getRevTrans(procs[i])
		revprocs.append(rlt)
	procsrulestatemaps = []
	procsrulematches = []
	for i in range(0,len(procs)):
		getMatchesForProcess(i)
	# build list indicating on which processes a rule is applicable
	rulematcheson = []
	for rid in range(0,len(RS)):
		tmp = set([])
		for pid in range(0,len(procsrulematches)):
			if procsrulematches[pid][rid] != []:
				tmp.add(pid)
		rulematcheson.append(tmp)
	for rid in range(0,len(RS_add)):
		rulematcheson.append(RS_add[rid][2])

def matchSyncrules():
	"""Match the syncrules (old and new) of the rulesystem on the network"""
	global RS, RS_vector, syncrules, newsyncrules, procsrulestatemaps, procsrulematches, syncactions, newsyncactions, RS_left_syncrules
	global RS_right_syncrules, RS_ownsaction, inputnetwork, rulematcheson, procsactprefixsets, RS_actprefixsets, procs_orig_size
	structureset = []
	#for i in range(0,len(inputnetwork)):
	#	if hasattr(inputnetwork[i], "__iter__") and not isinstance(inputnetwork[i], str):
	#		print inputnetwork
	#		structureset.append(set(inputnetwork[i]))
	#	else:
	#		structureset.append(set([inputnetwork[i]]))
	#posset = {}
	#for i in range(0,len(procsrulematches)):
	#	for j in range(0,len(structureset)):
	#		if i in structureset[j]:
	#			posset[i] = j
	# check compatibility between left syncrules and syncrules of network
	# iterate over action prefixes in left patterns of transformation rules
	seen = set([])
	for i in range(0,len(RS_vector)):
		r = RS_vector[i]
		if r not in seen:
			seen.add(r)
			for a in RS_actprefixsets[r][0]:
				# determine synchronisation possibilities of a
				synchs_a = set([])
				for (vects, rltact, interface) in RS_left_syncrules:
					tmplist = []
					for v in vects:
						if v[i] == a:
							for j in range(0,len(v)):
								if v[j] != '*':
									tmplist.append(v[j])
							break
					if tmplist != []:
						for ac in list(interface.elements()):
							for j in range(0,interface[ac]):
								tmplist.append(ac)
						tmplist.sort()
						synchs_a.add((tuple(tmplist), actionStripInterfacecounter(rltact)))
				# for all processes matched on by r, check that a has the same synching potential
				for pid in rulematcheson[r]:
					for j in range(0,len(syncrules)):
						for ac, v in syncrules[j].items():
							for (vect, rltact) in v:
								if vect[pid] == a:
									tmplist = []
									for k in range(0,len(vect)):
										if vect[k] != '*':
											tmplist.append(vect[k])
									if len(tmplist) > 1:
										tmplist.sort()
										if (tuple(tmplist),actionStripInterfacecounter(rltact)) not in synchs_a:
											print_error("action " + str(a) + " has different synching potential in process " + str(pid) + " than in rule " + str(r) + "!")
	# Determine how to translate the RS right syncrules to new concrete network rules
	# Construct concrete sync rules from the RS right syncrules
	f = {}
	for (vects, rltact, interface) in RS_right_syncrules:
		for v in vects:
			for i in range(0,len(v)):
				if v[i] != '*':
					if f.get(v[i]) == None:
						tmp = set([])
						rls = RS_ownsaction[1].get(v[i])
						for r in rls:
							tmp |= rulematcheson[r]
						f[v[i]] = tmp
		intset = set(interface.elements())
		for a in intset:
			if f.get(a) == None:
				tmp = set([])
				for i in range(0,len(syncactions)):
					# NOTE: If a does not appear in any syncactions[i], the given interface was not valid.
					# We decide not to act on this; if after verification, the RS is correct with respect
					# to a more restricting interface, then it will definitely be correct with a less restricting
					# one.
					if a in syncactions[i]:
						tmp.add(i)
				f[a] = tmp
	matched_rules = {}
	for (vects, rltact, interface) in RS_right_syncrules:
		intset = set(interface.elements())
		for v in vects:
			newvect = []
			for i in range(0,len(v)):
				if v[i] != '*':
					newvect.append(v[i])
			for a in intset:
				for i in range(0,interface[a]):
					newvect.append(a)
			tmplist = list(perm_pos_unique(newvect, f, len(procsrulematches)))
			inttmplist = [(t,(intset != set([]))) for t in tmplist]
			newact = actionStripInterfacecounter(rltact)
			if matched_rules.get(newact) == None:
				matched_rules[newact] = []
			rlt = matched_rules.get(newact)
			matched_rules[newact] = list(set(rlt) | set(inttmplist))
	# filter out invalid vectors, based on existing sync rules of input network, in cases that the interface is not empty
	for a, vects in matched_rules.items():
		newvects = []
		for (v, interface_is_not_empty) in vects:
			if interface_is_not_empty:
				# count the number of actions in the vector related to 'old' processes (i.e. not newly added ones)
				count = 0
				for i in range(0,procs_orig_size):
					if v[i] != '*':
						count += 1
				if count <= 1:
					# at most one old process involved. This is correct by definition. (No syncing within the old part of the system)
					found = True
				else:
					# construct list of sets of rules applicable on each process
					apprules = []
					for i in range(0,len(v)):
						if v[i] != '*':
							apprules.append(RS_getRulesWithAction(v[i],True))
						else:
							apprules.append(set([]))
					# We need to be able to find a matching rule in syncrules
					found = False
					for j in range(0,len(syncrules)):
						for sr_a, sr_v in syncrules[j].items():
							for (v2, rltact) in sr_v:
								for i in range(0,len(v)):
									if v[i] != '*':
										if v[i] in procsactprefixsets[i]:
											if v[i] != v2[i]:
												break
										else:
											# new action. determine whether a matching rule has v2[i] in left pattern
											rulefound = False
											for rule in apprules[i]:
												# if rule is an add rule, and v2[i] == '*', we have a match
												if rule > len(RS) and v2[i] == '*':
													rulefound = True
													break
												else:
													if v2[i] in RS_actprefixsets[rule][0]:
														rulefound = True
														break
											if not rulefound:
												break
									if i == len(v)-1:
										found = True
								if found:
									break
							if found:
								break
						if found:
							break
				if found:
					# match found. v is allowed for the network
					newvects.append(v)
			else:
				# interface is empty, rule can be safely added
				newvects.append(v)
		# add newvects to matched_rules
		matched_rules[a] = newvects
		# Possibly warn about filtering
		if newvects == []:
			print_message("WARNING!: a new sync right rule has been removed due to inapplicability!")
	# Construct the newsyncactions
	newsyncactions = copy.deepcopy(syncactions)
	for ac, vects in matched_rules.items():
		for v in vects:
			for i in range(0,len(v)):
				if v[i] != '*':
					newsyncactions[i].add(v[i])
	newsyncrules = copy.deepcopy(syncrules)
	newsyncrules = orderSyncrules(matched_rules, newsyncrules)

	# 2. CHECK MODEL DEPENDENT WELL-FORMEDNESS OF SYNC RULES
	# Check if existing sync actions occurring in right patterns also appear in corresponding left ones
	for i in range(0,len(RS)):
		pids = rulematcheson[i]
		for a in RS[i][1][2]:
			for pid in pids:
				if mustSync(a, pid, syncactions):
					# does the same action also exist in left pattern?
					if a not in RS[i][0][2]:
						print_error("well-formedness violated! Transformation rule " + str(i) + " applied on process " + str(pid) + " uses an existing sync action in the right pattern that is not in the left pattern!")
	# 3. CHECK CONSISTENT TRANSFORMATION OF SYNC ACTIONS
	# a data structure to track per process and per rule which concrete sync action labels are matched on
	procsruleconcretesyncactions = []
	for i in range(0,len(procs)):
		procsruleconcretesyncactions.append([])
		for j in range(0,len(RS)):
			procsruleconcretesyncactions[i].append(set([]))
	for j in range(0,len(RS)):
		for i in range(0,len(procs)):
			mapinfo = procsrulestatemaps[i][j]
			# for all syncing actions in process i occurring in the left-hand pattern, check the matches
			if mapinfo != {}:
				# obtain action ordered left pattern
				actleft = getActTrans((RS[j][0])[1])
				#for a in procssyncactsets[i]:
					#print procssyncactsets[i]
				for a in RS[j][0][2]:
					#if a in (rule[0])[2]:
						#print actprocs[i]
					a_stripped = actionStripParams(a)
					if a_stripped in syncactions[i]:
						# obtain a set of concrete action labels on which the pattern action label a applies (interprets placeholders)
						concrete_aset = getMatchedActions(a, procsrulematches[i][j], i)
						# add this set to procsruleconcretesyncactions
						procsruleconcretesyncactions[i][j] |= concrete_aset
						for ca in concrete_aset:
							a_trans = actprocs[i].get(ca)
							if a_trans != None:
								a_rule = actleft.get(a)
								for src, t in a_trans.items():
									found = False
									for srcm, tm in a_rule.items():
										if src in mapinfo.get(srcm):
											found = True
											break
									if not found:
										print_error("well-formedness violated! A syncing transition labelled " + str(ca) + " in process " + str(i) + " is not transformed by rule " + str(rule) + "!")
									else:
										for tgt in t:
											found = False
											for tgtm in tm:
												if tgt in mapinfo.get(tgtm):
													found = True
													break
											if not found:
												print_error("well-formedness violated! A syncing transition labelled " + str(ca) + " in process " + str(i) + " is not transformed by rule " + str(rule) + "!")

def flipMatch(mtch):
	"""produce the reverse of the given match to reason from process LTS states to rule pattern states"""
	f = {}
	for k, v in mtch[0].items():
		f[v] = k
	return f

def removeDanglingMatches(processid, rule, statemap):
	"""Remove matches that result in dangling edges"""
	global procs, revprocs
	newstatemap = []
	for mtch in statemap:
		if mtch != [{},{}]:
			next = False 
			nr_of_deleted_states = int(((rule[0])[0])[2]) - int(((rule[0])[0])[0])
			if nr_of_deleted_states > 0:
				flipmtch = flipMatch(mtch)
				for s in range(int(((rule[0])[0])[0]), int(((rule[0])[0])[2])):
					if next:
						break
					s_process = mtch[0].get(str(s))
					outgoing = (procs[processid]).get(s_process)
					if outgoing != None:
						for a, t in outgoing.items():
							if next:
								break
							for tgt in t:
								if next:
									break
								# is the tgt matched on by mtch?
								tgt_mtch = flipmtch.get(tgt)
								if tgt_mtch == None:
									next = True
									break
								else:
									out = ((rule[0])[1]).get(str(s))
									if out == None:
										next = True
										break
									else:
										outwithlabel = out.get(a)
										if outwithlabel == None:
											next = True
											break
										else:
											if tgt_mtch not in outwithlabel:
												next = True
												break
					incoming = (revprocs[processid]).get(s_process)
					if incoming != None:
						for a, t in incoming.items():
							if next:
								break
							for src in t:
								if next:
									break
								# is the tgt matched on by mtch?
								src_mtch = flipmtch.get(src)
								if src_mtch == None:
									next = True
									break
								else:
									out = ((rule[0])[1]).get(src_mtch)
									if out == None:
										next = True
										break
									else:
										outwithlabel = out.get(a)
										if outwithlabel == None:
											next = True
											break
										else:
											if str(s) not in outwithlabel:
												next = True
												break
			if not next:
				newstatemap.append(mtch)
	return newstatemap

def removeExclusiveglueViolations(processid, rule, statemap):
	"""Remove matches that violate exclusive glue conditions"""
	global procs, revprocs
	newstatemap = []
	for mtch in statemap:
		if mtch != [{},{}]:
			# obtain the reverse of the match to reason in the other direction
			fmtch = flipMatch(mtch)
			next = False
			if int(rule[0][0][3]) > 0:
				flipmtch = flipMatch(mtch)
				for i in range(0,int(rule[0][0][3])):
					outgoing = procs[processid].get(mtch[0].get(str(i)))
					if outgoing == None:
						outgoing = {}
					for a, t in outgoing.items():
						for tgt in t:
							if fmtch.get(tgt) == None:
								next = True
								break
						if next:
							break
					if next:
						break
			if not next:
				newstatemap.append(mtch)
	return newstatemap

def getAllMatches(processid, rule):
	"""get all the matches of the given rule on the given process LTS"""
	if len((rule[0])[1]) == 0:
		if rule[0][0][2] == '1':
			return [[{'0': '0'},{}]]
		else:
			return []
	else:
		return extendMatches(processid, rule, [[{},{}]])

def groupMatches(statemap):
	"""merge the info in the matches in statemap"""
	result = {}
	for mtch in statemap:
		for s, smap in mtch[0].items():
			rlt = result.get(s)
			if rlt == None:
				rlt = set([])
			rlt.add(smap)
			result[s] = rlt
	return result

def updateMatch(processid, tpl, mtch):
	"""check for a mapping of transition (k,a,tgt) on process and return all possible extensions of mtch"""
	# a match is a tuple containing:
	# 1) a mapping between states, and
	# 2) a placeholder dict
	(src,a,tgt) = tpl
	newstatemap = []
	#t = a.split(' ')
	#prefix = t[0]
	#i = 1
	#for i in range(1,len(t)):
	#	if t[i][0] != '#':
	#		prefix += ' ' + t[i]
	#	else:
	#		break
	params = actionGetPlaceholder(a)
	hasplaceholder = (params != '')
	prefix = actionStripPlaceholder(a)
	if prefix[0] == '\'':
		# prefix is a regular expression
		regexpr = True
	else:
		regexpr = False
	prefixparams = ''
	if hasplaceholder:
		# is the value of the placeholder known? If so, construct the concrete action label
		if mtch[1].get(params) != None:
			prefixparams = mtch[1].get(params)
			placeholderfound = True
		else:
			placeholderfound = False
	proc_a_trans = []
	if not regexpr and (not hasplaceholder or placeholderfound):
		# no placeholder there. We can directly fetch transitions
		proc_a_trans = (actprocs[processid]).get('"' + prefix + prefixparams + '"')
		if proc_a_trans != None:
			srcmap = mtch[0].get(src)
			tgtmap = mtch[0].get(tgt)
			for src1, v in proc_a_trans.items():
				for tgt1 in v:
					addsrc = False
					addtgt = False
					if srcmap == None:
						addsrc = True
					if srcmap == src1 or addsrc:
						if tgtmap == None:
							addtgt = True
						if tgtmap == tgt1 or addtgt:
							newmtch = copy.deepcopy(mtch)
							if addsrc:
								newmtch[0][src] = src1
							if addtgt:
								newmtch[0][tgt] = tgt1
							newstatemap.append(newmtch)
	else:
		for ac, rlts in actprocs[processid].items():
			if actionMatches(a, ac, mtch[1]):
				srcmap = mtch[0].get(src)
				tgtmap = mtch[0].get(tgt)
				for src1, v in rlts.items():
					for tgt1 in v:
						addsrc = False
						addtgt = False
						if srcmap == None:
							addsrc = True
						if srcmap == src1 or addsrc:
							if tgtmap == None:
								addtgt = True
							if tgtmap == tgt1 or addtgt:
								newmtch = copy.deepcopy(mtch)
								if addsrc:
									newmtch[0][src] = src1
								if addtgt:
									newmtch[0][tgt] = tgt1
								concrete_a = actionGetPlaceholdervalue(a, ac)
								if concrete_a != "":
									newmtch[1][params] = concrete_a
								newstatemap.append(newmtch)
	return newstatemap

def extendMatchesTransition(processid, tpl, statemap):
	(src,a,tgt) = tpl
	newstatemap = []
	for mtch in statemap:
		newstatemap = newstatemap + updateMatch(processid, (src,a,tgt), mtch)
	return newstatemap

def extendMatches(processid, rule, statemap):
	skip = False
	for src, v in ((rule[0])[1]).items():
		for a, t in v.items():
			for tgt in t:
				# was +=!
				statemap = extendMatchesTransition(processid, (src,a,tgt), statemap)
				if statemap == []:
					skip = True
					break
			if skip:
				break
		if skip:
			break
	return statemap

def isEmptyVector(v):
	"""is the vector v empty?"""
	for ve in v:
		if v != -1:
			return False
	return True

def constructChecks():
	"""construct the required checks for property preservation"""
	global RS, RS_vector, checks, RS_left_syncrules, RS_right_syncrules, exclsets
	tocheck = set([i for i in range(0,len(RS_vector))])
	# while we construct checks, we also construct and compare the relevant interfaces. Before and after
	# transformation, these should be the same for a given set of dependent transformation rules
	checks = []
	while tocheck != set([]):
		ri = tocheck.pop()
		c = [set([]),set([])]
		if RS_vector[ri] < len(RS):
			c[0].add(ri)
		else:
			c[1].add(ri)
		cset = set([ri])
		interfaces = [Counter(), Counter()]
		eset = set([])
		next = True
		while next:
			next = False
			for side in [0,1]:
				if side == 0:
					RS_syncrules = RS_left_syncrules
				else:
					RS_syncrules = copy.deepcopy(RS_left_syncrules)
					RS_syncrules += RS_right_syncrules
				for (vects, rltact, interface) in RS_syncrules:
					add_interface = False
					for v in vects:
						# is the vector relevant?
						found = False
						for i in range(0,len(v)):
							if v[i] != '*' and i in cset:
								found = True
								break
						if found:
							for i in range(0,len(v)):
								if v[i] != '*':
									if RS_vector[i] < len(RS):
										if i not in c[0]:
											next = True
											c[0].add(i)
											cset.add(i)
									else:
										if i not in c[1]:
											next = True
											c[1].add(i)
											cset.add(i)
							add_interface = True
					if add_interface:
						# extend the left interface of the check
						for ia in set(interface.elements()):
							if interfaces[side][ia] < interface[ia]:
								interfaces[side][ia] = interface[ia]
						if len(set(interface.elements())) > 0:
							# add rltact to eset
							eset.add('"' + rltact + '"')
		# we have obtained a new check. Remove those rules from tocheck
		for i in range(0,len(c)):
			if c[i] != -1:
				tocheck.discard(i)
		# compare interfaces
		if interfaces[0] != interfaces[1]:
			rset = set([])
			for c in cset:
				rset.add(RS_vector[c])
			print_error("inconsistency in possible synchronisation with the environment! Concerns rules " + str(rset))
		checks.append(c)
		exclsets.append(eset)

def powerset(iterable):
		"powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
		s = list(iterable)
		return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s)+1))

def checkPropPresSingle(chk):
	"""Perform single property preservation check (for multi-threading)"""
	global checks, RS, RS_vector, procndivsets, systemtrans, procsrulestatemaps, revhashtable, rulematcheson, exclsets
	ccl = []
	ccr = []
	ccl_headers = []
	ccr_headers = []
	subtrset = set(chk[2])
	for i in range(0,len(RS_vector)):
		if i in subtrset:
			r = copy.deepcopy(RS[RS_vector[i]][0][1])
			ccl.append(r)
			r = copy.deepcopy(RS[RS_vector[i]][1][1])
			ccr.append(r)
			header = tuple(['0', RS[RS_vector[i]][0][0][1], RS[RS_vector[i]][0][0][2]])
			ccl_headers.append(header)
			header = tuple(['0', RS[RS_vector[i]][1][0][1], RS[RS_vector[i]][1][0][2]])
			ccr_headers.append(header)
		else:
			if i in checks[chk[1]][1]:
				# add the rules introducing new processes
				r = copy.deepcopy(RS_add[RS_vector[i]-len(RS)][0][1])
				ccl.append(r)
				r = copy.deepcopy(RS_add[RS_vector[i]-len(RS)][1][1])
				ccr.append(r)
				header = tuple(['0', RS_add[RS_vector[i]-len(RS)][0][0][1], RS_add[RS_vector[i]-len(RS)][0][0][2]])
				ccl_headers.append(header)
				header = tuple(['0', RS_add[RS_vector[i]-len(RS)][1][0][1], RS_add[RS_vector[i]-len(RS)][1][0][2]])
				ccr_headers.append(header)
			else:
				ccl.append({})
				ccr.append({})
				ccl_headers.append(('0','0','0'))
				ccr_headers.append(('0','0','0'))

	# extend the rule patterns by rewriting kappa-selfloops
	for i in range(0,len(ccl)):
		if ccl[i] != {}:
			for src, v in ccl[i].items():
				rlt = v.get('kappa')
				if rlt != None:
					v['\"kappa_' + str(i) + "_" + src + "\""] = rlt
					v.pop('kappa')
					v1 = ccr[i].get(src)
					rlt = v1.get('kappa')
					v1['\"kappa_' + str(i) + "_" + src + "\""] = rlt
					v1.pop('kappa')
					ccr[i][src] = v1
	# generate LTSs of networks of rule patterns
	generateLTS(ccl_headers, ccl, [], RS_gen_left_syncrules, RS_left_syncactions, os.path.join(config.rs_folder, "rleft" + str(chk[0])), opt_quiet, config.outputobject, opt_genbound, list(range(0,len(ccl_headers))), set([]), False, False)
	generateLTS(ccr_headers, ccr, [], RS_gen_right_syncrules, RS_right_syncactions, os.path.join(config.rs_folder, "rright" + str(chk[0])), opt_quiet, config.outputobject, opt_genbound, list(range(0,len(ccr_headers))), set([]), False, False)

	if opt_divergency:
		# 1. CHECK THAT NO DIVERGENCY IS LOST ON GLUE-STATES WHEN refining
		# ONLY NEEDED WHEN OPT_DIVERGENCY IS TRUE, OTHERWISE THIS IS PART
		# OF THE BISIM CHECK AUTOMATICALLY
		#print_message("checking for divergency preservation")
		[local_header, local_trans, local_actset] = readAut(config.rs_folder, "rleft" + str(chk[0]) + ".aut")			
		local_map = readMap(config.rs_folder, "rleft" + str(chk[0]))
		# determine set of diverging states
		divset = computeDivergencySet(local_header, local_trans)
		# construct the set of state vectors that consist only of glue-states
		ldivvectors = set([])
		for s in divset:
			svec = local_map.get(s)
			mustadd = True
			for i in range(0,len(ccl)):
				if ccl[i] != {}:
					if int(s[i]) >= ccl_headers[i][0]:
						mustadd = False
			if mustadd:
				ldivvectors.add(tuple(svec))
		[local_header, local_trans, local_actset] = readAut(config.rs_folder, "rright" + str(chk[0]) + ".aut")			
		local_map = readMap(config.rs_folder, "rright" + str(chk[0]))
		# determine set of diverging states
		divset = computeDivergencySet(local_header, local_trans)
		# construct the set of state vectors
		rdivvectors = set([])
		for s in divset:
			svec = local_map.get(s)
			rdivvectors.add(tuple(svec))
		# compare the divvector sets
		if not ldivvectors <= rdivvectors:
			print_error("Transformation may remove diverging behaviour!")
		# 2. ADD DIVERGENCY INFO TO RULE NETWORKS AND COMPARE
		# add tau-selfloops to diverging states in left and right patterns
		for i in range(0,len(ccl)):
			if ccl[i] != {}:
				# We have to skip the source state if we previously added a new initial state)
				skip_src = False
				if int(RS[RS_vector[i]][0][0][0]) > 1:
					skip_src = True
				for src, v in ccl[i].items():
					if skip_src:
						srcmin = str(int(src)-1)
					else:
						srcmin = src
					if int(srcmin) > -1:
						addtau = True
						for pid in rulematcheson[RS_vector[i]]:
							mapset = procsrulestatemaps[pid][RS_vector[i]].get(srcmin)
							if mapset == None:
								mapset = set([])
							if mapset & procndivsets[i] != set([]):
								addtau = False
								break
						if addtau:
							rlt = v.get('"tau"')
							if rlt == None:
								rlt = set([])
							if src not in rlt:
								rlt.add(src)
								ccl_headers[i] = (ccl_headers[i][0], str(int(ccl_headers[i][1])+1), ccl_headers[i][2])
								v['"tau"'] = rlt
							# if src is a glue-state, also add this tau-loop to the right network
							if int(srcmin) < int(RS[RS_vector[i]][0][0][0]):
								vr = ccr[i].get(src)
								rlt = vr.get('"tau"')
								if rlt == None:
									rlt = set([])
								if src not in rlt:
									rlt.add(src)
									ccr_headers[i] = (ccr_headers[i][0], str(int(ccr_headers[i][1])+1), ccr_headers[i][2])
									vr['"tau"'] = rlt
									ccr[i][src] = vr
		# generate LTSs of networks of rule patterns, now with divergency information
		generateLTS(ccl_headers, ccl, [], RS_gen_left_syncrules, RS_left_syncactions, os.path.join(config.rs_folder, "rleft" + str(chk[0])), opt_quiet, config.outputobject, opt_genbound, list(range(0,len(ccl_headers))), set([]), False, False)
		generateLTS(ccr_headers, ccr, [], RS_gen_right_syncrules, RS_right_syncactions, os.path.join(config.rs_folder, "rright" + str(chk[0])), opt_quiet, config.outputobject, opt_genbound, list(range(0,len(ccr_headers))), set([]), False, False)

	# hide the necessary actions
	try:
		inFile = open(os.path.join(config.rs_folder, "rleft" + str(chk[0]) + ".aut"),'r')
	except IOError:
		exit(1)
	hideActions(config.rs_folder, "rleft" + str(chk[0]), exclsets[chk[1]], "")
	try:
		inFile = open(os.path.join(config.rs_folder, "rright" + str(chk[0]) + ".aut"),'r')
	except IOError:
		exit(1)
	hideActions(config.rs_folder, "rright" + str(chk[0]), exclsets[chk[1]], "")

	# perform bisim check
	if opt_fairness:
		#print_message("checking branching bisimilarity of left and right rule network LTSs")
		#rlt = subprocess.check_call(["ltscompare", "--equivalence=branching-bisim", os.path.join(config.network_folder, "rleft" + str(chk[0]) + "_hidden.aut"), os.path.join(config.network_folder, "rright" + str(chk[0]) + "_hidden.aut")], stderr=subprocess.PIPE)
		rlt = subprocess.Popen(["ltscompare", "--equivalence=branching-bisim", os.path.join(config.rs_folder, "rleft" + str(chk[0]) + "_hidden.aut"), os.path.join(config.rs_folder, "rright" + str(chk[0]) + "_hidden.aut")], stderr=subprocess.PIPE)
	else:
		#print_message("checking divergence-sensitive branching bisimilarity of left and right rule network LTSs")
		rlt = subprocess.Popen(["ltscompare", "--equivalence=dpbranching-bisim", os.path.join(config.rs_folder, "rleft" + str(chk[0]) + "_hidden.aut"), os.path.join(config.rs_folder, "rright" + str(chk[0]) + "_hidden.aut")], stderr=subprocess.PIPE)

	out, err = rlt.communicate()
	i = str(err).find('not')
	if i == -1:
		# clean up temporary files
		return 0
	#	if opt_fairness:
			#print_message("LTSs are branching bisimilar")
	#	else:
			#print_message("LTSs are divergence-sensitive branching bisimilar")
	else:
		return chk[0]
		#if opt_fairness:
			#print_message("check " + str(checknr) + ": LTSs are not branching bisimilar")
		#else:
			#print_message("check " + str(checknr) + ": LTSs are not divergence-sensitive branching bisimilar")

def extendRule(rule):
	"""Extend the given rule for property preservation checking"""
	# add a new root state connected to all glue-states in the rule patterns via kappa-transitions.
	# all glue-states have an incoming kappa-transition from this state, and all non-exclusive glue-states
	# have an outgoing kappa-transition to this state.
	# The kappa labels at this point are simply called 'kappa', later on they
	# will be altered by adding ids.

	for i in [0,1]:
		ntrans = int(rule[i][0][1])
		newr = {}
		for j in range(0,int(rule[i][0][2])):
			v = rule[i][1].get(str(j))
			newv = {}
			if v != None:
				for a, t in v.items():
					newv[a] = set([str(int(x)+1) for x in t])
			# add outgoing kappa-transition?
			if j < int(rule[i][0][0]) and (j >= int(rule[i][0][3]) or int(rule[i][0][3]) == 0):
				newv['kappa'] = set(['0'])
				ntrans += 1
			newr[str(j+1)] = newv
		# add new root state, with kappa transitions to the glue-states
		newr['0'] = {'kappa': set([str(x) for x in range(1,int(rule[i][0][0])+1)])}
		ntrans += int(rule[i][0][0])
		rule[i][1] = newr
		rule[i][0] = (rule[i][0][0], str(ntrans), str(int(rule[i][0][2])+1), rule[i][0][3])

def	checkPropPreservation():
	"""Check property preservation of RS"""
	#NEW CHECK: ALSO ADD DIVERGENCY LOOPS IN RIGHT PATTERNS, CHECK FOR DIV PRESERVATION BY CHECKING
	#DIVERGENCY SETS LEFT AND RIGHT, AND THAT RIGHT IS SUPERSET OF LEFT
	global checks, RS, RS_vector, procndivsets, systemtrans, procsrulestatemaps, revhashtable
	# add a new root state connected to all glue-states in the rule patterns via tau-transitions,
	# and add kappa-selfloops to glue-states (at this point simply called 'kappa', later on they
	# will be altered by adding ids). Do not add a state if there is only one glue-state, and do not
	# add kappa-transitions to exclusive glue-states.
	checks_done = set([])

	for rule in RS:
		extendRule(rule)
	# ADD RULES DO NOT NEED EXTENSIONS!
	#for rule in RS_add:
	#	extendRule(rule)
	# perform the checks
	nrofchecks = 0
	checknr = 0
	# list of actual checks to perform
	bisim_checks = []
	for ci in range(0,len(checks)):
		# consider any subset of transformation rules
		for pp in powerset(checks[ci][0]):
			# obtain the check configurations
			# skip the empty config
			if len(pp) > 0:
				checksig = tuple([RS_vector[i] for i in pp])
				if checksig not in checks_done:
					checknr += 1
					bisim_checks.append((checknr, ci, pp))
					checks_done.add(checksig)
					nrofchecks += 1
	pool = multiprocessing.Pool(min(multiprocessing.cpu_count(), config.nrthreads))
	counterexamples = []
	r = pool.map_async(checkPropPresSingle, bisim_checks, callback=counterexamples.append)
	r.wait()
	# if relevant, construct counter-example
	error = False
	print(counterexamples)
	for ce in counterexamples[0]:
		if ce == -1:
			print_error("Transformation may remove divergent behaviour!")
		else:
			if ce > 0:
				if opt_divergency:
					print_message("PROPERTY NOT PRESERVED: preservation of " + config.property_name + " for network " + config.network_name + " violated!")
				else:
					if config.property_name != "":
						print_message("PROPERTY NOT PRESERVED: preservation of " + config.property_name + " violated!")
					else:
						print_message("SEMANTICS NOT PRESERVED: semantics preservation violated!")
				print_message('---')
				if not opt_divergency:
					print_message("try to run the check again with the 'fairness' option, or do a model-dependent check")
				print_message("relevant rule networks and LTSs are written to disk as rleft.1.(exp/aut) and rright.1.(exp/aut)")
				os.rename(os.path.join(config.rs_folder, "rleft" + str(ce) + ".aut"), os.path.join(config.rs_folder, "rleft.1.aut"))
				os.rename(os.path.join(config.rs_folder, "rright" + str(ce) + ".aut"), os.path.join(config.rs_folder, "rright.1.aut"))
				os.rename(os.path.join(config.rs_folder, "rleft" + str(ce) + "_hidden.aut"), os.path.join(config.rs_folder, "rleft_hidden.1.aut"))
				os.rename(os.path.join(config.rs_folder, "rright" + str(ce) + "_hidden.aut"), os.path.join(config.rs_folder, "rright_hidden.1.aut"))
				# write the two rule networks to disk
				for i in range(0,len(bisim_checks)):
					if bisim_checks[i][0] == ce:
						ci_ce_ad = checks[bisim_checks[i][1]][1]
						pp_ce = bisim_checks[i][2]
						break
				ccl = []
				ccr = []
				ccl_headers = []
				ccr_headers = []
				for i in range(0,len(RS_vector)):
					if i in pp_ce:
						r = copy.deepcopy(RS[RS_vector[i]][0][1])
						ccl.append(r)
						r = copy.deepcopy(RS[RS_vector[i]][1][1])
						ccr.append(r)
						header = tuple(['0', RS[RS_vector[i]][0][0][1], RS[RS_vector[i]][0][0][2]])
						ccl_headers.append(header)
						header = tuple(['0', RS[RS_vector[i]][1][0][1], RS[RS_vector[i]][1][0][2]])
						ccr_headers.append(header)
					else:
						if i in ci_ce_ad:
							r = copy.deepcopy(RS_add[RS_vector[i]-len(RS)][0][1])
							ccl.append(r)
							r = copy.deepcopy(RS_add[RS_vector[i]-len(RS)][1][1])
							ccr.append(r)
							header = tuple(['0', RS_add[RS_vector[i]-len(RS)][0][0][1], RS_add[RS_vector[i]-len(RS)][0][0][2]])
							ccl_headers.append(header)
							header = tuple(['0', RS_add[RS_vector[i]-len(RS)][1][0][1], RS_add[RS_vector[i]-len(RS)][1][0][2]])
							ccr_headers.append(header)
						else:
							ccl.append({})
							ccr.append({})
							ccl_headers.append(('0','0','0'))
							ccr_headers.append(('0','0','0'))
				# extend the rule patterns by rewriting kappa-selfloops
				for i in range(0,len(ccl)):
					if ccl[i] != {}:
						for src, v in ccl[i].items():
							rlt = v.get('kappa')
							if rlt != None:
								v['\"kappa_' + str(i) + "_" + src + "\""] = rlt
								v.pop('kappa')
								v1 = ccr[i].get(src)
								rlt = v1.get('kappa')
								v1['\"kappa_' + str(i) + "_" + src + "\""] = rlt
								v1.pop('kappa')
								ccr[i][src] = v1
				rfiles = []
				for i in range(0,len(ccl)):
					if ccl[i] == {}:
						rfiles.append('empty.aut')
					else:
						rfiles.append('rlpat_' + str(i) + '.aut')
				writeNetwork(config.rs_folder, ccl, ccl_headers, rfiles, RS_gen_left_syncrules, "rleft", False)
				for i in range(0,len(ccl)):
					if rfiles[i] != 'empty.aut' or ccr[i] != {}:
						rfiles[i] = 'rrpat_' + str(i) + '.aut'					
				writeNetwork(config.rs_folder, ccr, ccr_headers, rfiles, RS_gen_right_syncrules, "rright", False)
				error = True
				break
	# done
	if not error:
		#print_message('---')
		if opt_divergency:
			print_message("PROPERTY PRESERVED: preservation of " + config.property_name + " for network " + config.network_name + " determined after " + str(nrofchecks) + " checks")
		else:
			if opt_fairness:
				print_message("PROPERTY PRESERVED UNDER FAIRNESS: preservation of " + config.property_name + " determined after " + str(nrofchecks) + " checks")
			elif config.property_name != "":
				print_message("PROPERTY PRESERVED: preservation of " + config.property_name + " determined after " + str(nrofchecks) + " checks")
			else:
				print_message("SEMANTICS PRESERVED: semantics preservation determined after " + str(nrofchecks) + " checks")
		print_message('---')
	#clean up temporary files
	for i in range(1,len(bisim_checks)+1):
		#if os.path.isfile(os.path.join(config.rs_folder, "rleft" + str(i) + ".aut")):
		#	os.remove(os.path.join(config.rs_folder, "rleft" + str(i) + ".aut"))
		#if os.path.isfile(os.path.join(config.rs_folder, "rright" + str(i) + ".aut")):
		#	os.remove(os.path.join(config.rs_folder, "rright" + str(i) + ".aut"))
		if os.path.isfile(os.path.join(config.rs_folder, "rleft" + str(i) + ".map")):
			os.remove(os.path.join(config.rs_folder, "rleft" + str(i) + ".map"))
		if os.path.isfile(os.path.join(config.rs_folder, "rright" + str(i) + ".map")):
			os.remove(os.path.join(config.rs_folder, "rright" + str(i) + ".map"))
		if os.path.isfile(os.path.join(config.rs_folder, "rleft" + str(i) + "_hidden.aut")):
			os.remove(os.path.join(config.rs_folder, "rleft" + str(i) + "_hidden.aut"))
		if os.path.isfile(os.path.join(config.rs_folder, "rright" + str(i) + "_hidden.aut")):
			os.remove(os.path.join(config.rs_folder, "rright" + str(i) + "_hidden.aut"))
		if os.path.isfile(os.path.join(config.rs_folder, "rleft" + str(i) + "_hidden.map")):
			os.remove(os.path.join(config.rs_folder, "rleft" + str(i) + "_hidden.map"))
		if os.path.isfile(os.path.join(config.rs_folder, "rright" + str(i) + "_hidden.map")):
			os.remove(os.path.join(config.rs_folder, "rright" + str(i) + "_hidden.map"))

# END functions for transformation rules / performing transformations on a network

def isIn(a, actions, startindex):
	"""Check if a is a superstring of an element in actions from startindex. If a contains '*' characters, do a more thorough scan"""
	for i in range(startindex,len(actions)):
		# two types actions: 1) strings enclosed by ", 2) regexps enclosed by '
		if actions[i][0] == '\'':
			exp = actions[i][1:][:-1]
			p = re.compile('"' + exp + '"')
			if p.match(a) != None:
				return True
		else:
			if actions[i] in a:
				return True
	return False

def writeHideActions():
	"""Write the hide actions to opt_constructhidingset.hs"""
	global hideactions
	outfile = open(os.path.join(config.property_folder, config.property_name + ".hs"), 'w')
	first = True
	firstact = True
	for i in range(0,len(hideactions)):
		if hideactions[i] == 'none':
			outfile.write(hideactions[i])
			outfile.close()
			return
		if hideactions[i] == 'all' or hideactions[i] == 'but':
			if not first:
				outfile.write(' ')
			else:
				first = False
			outfile.write(hideactions[i])
		else:
			if not firstact:
				outfile.write(', ')
			else:
				outfile.write(' ')
				firstact = False
			outfile.write(hideactions[i])
	outfile.close()

def readHideActions(hideFile):
	"""Read an action hide file"""
	global hideactions
	lines = hideFile.readlines()
	line = ""
	for l in lines:
		line += l.rstrip()
	tmp = line.split()
	hideactions = []
	hasall = False
	hasbut = False
	if tmp[0] == 'none':
		hideactions.append(tmp[0])
		return
	if tmp[0] == 'all':
		hideactions.append(tmp[0])
		hasall = True
	if len(tmp) > 1:
		if tmp[1] == 'but':
			hideactions.append(tmp[1])
			hasbut = True
		if len(tmp) > 2:
			if hasbut:
				line = line[8:]
			elif hasall:
				line = line[4:]
			tmp = (line.rstrip()).split(', ')
			for x in tmp:
				#if x[0] == '"' or x[0] == '\'':
				#	x = x[1:][:-1]
				hideactions.append(x)

def hideAction(a, exclset):
	"""Hide action a according to hideactions. Actions in exclset should not be hidden"""
	global hiddenactions, hideactions
	if a.find('kappa') != -1:
		return a
	a_prefix = '"' + actionStripParams(a) + '"'
	if a_prefix in exclset:
		return a
	if hideactions == []:
		return a
	if hideactions[0] == 'none':
		return a
	if a == 'i' or a == '"tau"':
		return '"tau"'
	if len(hideactions) == 1 and hideactions[0]=='all':
		return '"tau"'
	elif len(hideactions) > 1 and hideactions[0]=='all' and hideactions[1]=='but':
		if len(hideactions) == 2:
			return '"tau"'
		elif a in hiddenactions:
			return a
		elif isIn(a, hideactions, 2):
			hiddenactions.add(a)
			return a
		else:
			return '"tau"'
	elif a in hiddenactions:
		return '"tau"'
	elif isIn(a, hideactions, 0):
		hiddenactions.add(a)
		return '"tau"'
	else:
		return a

def hideActions(folder, autbase, exclset, outputfiletail):
	"""Hide the actions in hideactions in the given .aut file, and write the result. Actions in exclset should not be hidden"""
	global hiddenactions

	outFile = open(os.path.join(folder, autbase + "_hidden" + outputfiletail + ".aut"), 'w')
	try:
		inFile = open(os.path.join(folder, autbase + ".aut"),'r')
	except IOError:
		exit(1)
	# reset hiddenactions
	hiddenactions = set([])
	#lines = inFile.readlines()
	#inFile.close()
	# first copy header
	line = inFile.readline()
	outFile.write(line)
	line = inFile.readline()
	while line != "":
		linelist = line.split(', ')
		# if ', ' appears in transition label, concatenate label back together
		if len(linelist) > 3:
			linelist[1] = ''.join([linelist[j] + ', ' for j in range(1, len(linelist)-1)])
			linelist[1] = linelist[1][:-2]
			linelist[2] = linelist[len(linelist)-1]
		# add quotes if not present
		if linelist[1][0] != '"':
			linelist[1] = '"' + linelist[1] + '"'
		# remove spaces
		linelist[1] = linelist[1].replace(' ', '')
		linelist[1] = hideAction(linelist[1], exclset)
		outFile.write(linelist[0] + ', ' + linelist[1] + ', ' + linelist[2])
		line = inFile.readline()
	inFile.close()
	outFile.close()

# def hideActionsinDict():
# 	"""hide the actions in systemtrans according to the hideactions list"""
# 	global systemtrans
# 	hiddentrans = {}
# 	for src, v in systemtrans.items():
# 		hiddentrans[src] = {}
# 		outtau = v.get('"tau"')
# 		if outtau == None:
# 			outtau = set([])
# 		for a, t in v.items():
# 			if hideAction(a) == '"tau"':
# 				outtau = outtau | t
# 			else:
# 				(hiddentrans[src])[a] = t
# 		if outtau != set([]):
# 			(hiddentrans[src])['"tau"'] = outtau
# 	systemtrans = hiddentrans

def mustSync(action, i, syncactions):
	"""Must the action sync in process i?"""
#	for ac in syncactions[i]:
#		if actionMatches(action, ac):
#			return True
#	return False
	action_stripped = actionStripParams(action)
	return action_stripped in syncactions[i]

def peek(stack):
	"""Return but do not pop top element of stack"""
	return stack[len(stack)-1]

def setdiv(stack):
	"""Set divergency flag of top element on stack to true"""
	(stack[len(stack)-1])[2] = True

def computeDivergencySet(header, trans):
	"""Compute the set of tau-diverging states in the (hidden) system LTS that are in the given set of states"""
	# the resulting set of diverging states
	divset = set([])

	# DFS stack
	stack = []
	stackset = set([])
	# visited set
	V = set([])
	# iterate over all states
	for i in range(0,int(header[2])):
		s = str(i)
		# current divergency status on path in stack
		divreach = False
		if s not in V:
			targets = set([])
			outgoing = trans.get(s)
			if outgoing != None:
				for a, v in outgoing.items():
					if hideAction(a, set([])) == '"tau"':
						targets = targets | v
			stack.append([s,targets,False])
			stackset.add(s)

		while len(stack) > 0:
			[s,targets,isdiv] = peek(stack)
			# are there targets left to explore?
			nextfound = False
			while (not nextfound) and len(targets) > 0:
				next = targets.pop()
				if next in stackset:
					# tau-cycle found
					isdiv = True
				else:
					if next not in V:
						nextfound = True
					else:
						if next in divset:
							isdiv = True
			divreach = divreach or isdiv
			if len(targets) == 0 and not nextfound:
				if divreach:
					divset.add(s)
				V.add(s)
				stack.pop()
				stackset.remove(s)
			else:
				if divreach:
					setdiv(stack)
				divreach = False
				targets = set([])
				outgoing = trans.get(next)
				if outgoing != None:
					for a, v in outgoing.items():
						if hideAction(a, set([])) == '"tau"':
							targets = targets | v
				stack.append([next,targets,False])
				stackset.add(next)
	return divset

def computeNonDivergingProcessStates():
	"""Compute the non-diverging process states"""
	# fill the sets of non tau-diverging process states
	global revhashtable, procs, procndivsets, divergencyset
	hel = revhashtable['0']
	#!
	for i in range(0,len(hel)):
		procndivsets.append(set([]))
	for i in range(0,int(systemLTSheader[2])):
		if str(i) not in divergencyset:
			hel = revhashtable.get(str(i))
			for j in range(0,len(hel)):
				(procndivsets[j]).add(hel[j])

def updateNonDivergingProcessStates():
	"""Update the sets of non-diverging states after a transformation"""
	# Pre-condition: the RS is property preserving
	global transformed_procs_new_states, checks, newsyncrules, newsyncactions, RS, RS_vector, procndivsets, systemtrans, divergencyset, revhashtable, procsrulestatemaps, systemLTSheader, hashtable
	RS_nrgluestates = []
	rulendivsets = []
	for i in range(0,len(transformed_procs_new_states)):
		newl = []
		for j in range(0,len(RS)):
			newl.append(set([]))
		for j in range(0,len(RS_add)):
			newl.append(set([]))
		rulendivsets.append(newl)
	# add a new root state connected to all glue-states in the right rule patterns via tau-transitions
	for i in [0,1]:
		if i == 0:
			RS_rules = RS
		else:
			RS_rules = RS_add
		for rule in RS_rules:
			if int(rule[1][0][0]) > 1:
				ntrans = int(rule[1][0][1])
				newr = {}
				for j in range(0,int(rule[1][0][2])):
					v = rule[1][1].get(str(j))
					newv = {}
					if v != None:
						for a, t in v.items():
							newv[a] = set([str(int(x)+1) for x in t])
					newr[str(j+1)] = newv
				newr['0'] = {'"tau"': set([str(x) for x in range(1,int(rule[1][0][0])+1)])}
				ntrans += int(rule[1][0][0])
				rule[1][1] = newr
				RS_nrgluestates.append(rule[1][0][0])
				rule[1][0] = ('0', str(ntrans), str(int(rule[1][0][2])+1), rule[1][0][3])
			else:
				RS_nrgluestates.append(rule[1][0][0])
				rule[1][0] = ('0', rule[1][0][1], rule[1][0][2], rule[1][0][3])
	# for each check, compute the rule network LTSs, and update the new ndiv sets based on divergency checks in those LTSs
	for ci in range(0,len(checks)):
		# consider any subset of transformation rules
		for pp in powerset(checks[ci][0]):
			# obtain the check configurations
			# skip the empty config
			if len(pp) > 0:
				ccr = []
				ccr_headers = []
				for i in range(0,len(RS_vector)):
					if i in pp:
						r = copy.deepcopy(RS[RS_vector[i]][1][1])
						ccr.append(r)
						header = tuple(['0', RS[RS_vector[i]][1][0][1], RS[RS_vector[i]][1][0][2]])
						ccr_headers.append(header)
					else:
						if i in checks[ci][1]:
							# add the rules introducing new processes
							r = copy.deepcopy(RS_add[RS_vector[i]-len(RS)][1][1])
							ccr.append(r)
							header = tuple(['0', RS_add[RS_vector[i]-len(RS)][1][0][1], RS_add[RS_vector[i]-len(RS)][1][0][2]])
							ccr_headers.append(header)
						else:
							ccr.append({})
							ccr_headers.append(('0','0','0'))
				# extend the rule patterns by adding tau-selfloops to diverging states in right patterns
				for i in range(0,len(ccr)):
					if ccr[i] != {}:
						for src in range(0,int(RS_nrgluestates[RS_vector[i]])):
							src = str(src)
							addtau = True
							for pid in rulematcheson[RS_vector[i]]:
								mapset = procsrulestatemaps[pid][RS_vector[i]].get(src)
								if mapset == None:
									mapset = set([])
								if mapset & procndivsets[i] != set([]):
									addtau = False
									break
							if addtau:
								v = ccr[i].get(src)
								rlt = v.get('"tau"')
								if rlt == None:
									rlt = set([])
								if src not in rlt:
									rlt.add(src)
									ccr_headers[i] = (ccr_headers[i][0], str(int(ccr_headers[i][1])+1), ccr_headers[i][2])
									v['"tau"'] = rlt
				# generate right rule network LTS
				generateLTS(ccr_headers, ccr, [], RS_gen_right_syncrules, RS_right_syncactions, os.path.join(config.rs_folder, "rright"), opt_quiet, config.outputobject, opt_genbound, list(range(0,len(ccr_headers))), set([]), False, False)
				[systemLTSheader, systemtrans, systemactset] = readAut(config.rs_folder, "rright.aut")
				revhashtable = readMap(config.rs_folder, "rright")
				# compute divergency set for this network
				divergencyset = computeDivergencySet(systemLTSheader, systemtrans)
				for i in range(0,int(systemLTSheader[2])):
					if str(i) not in divergencyset:
						# obtain the state vector
						sv = revhashtable.get(str(i))
						# for each element in the vector, add it to the corresponding rulendivset of the matching procs
						for j in range(0,len(RS_vector)):
							if j in pp or j in checks[ci][1]:
								for pid in rulematcheson[RS_vector[j]]:
									rulendivsets[pid][RS_vector[j]].add(sv[j])
	# All states in rulendivsets relate to new non-diverging states in the process
	for i in range(0,len(transformed_procs_new_states)):
		newndivstates = set([])
		length = len(RS) + len(RS_add)
		for j in range(0,length):
			for s in rulendivsets[i][j]:
				tmp = transformed_procs_new_states[i][j].get(s)
				if tmp == None:
					tmp = set([])
				newndivstates = newndivstates | tmp
		procndivsets[i] = procndivsets[i] | newndivstates

# --- functions for compositional model checking ---

def cmc_constructDependency():
	global prop_dependency, procs, syncrules, syncactions
	tocheck = set([i for i in range(0,len(procs)-1)])
	# start from property id
	lastdepset = set([len(procs)-1])
	while tocheck != set([]):
		newdepset = set([])
		for i in range(0,len(syncrules)):
			for a, v in syncrules[i].items():
				for (vect, act) in v:
					for id in lastdepset:
						if vect[id] != '*':
							# suitable vector. find dependent processes
							new_tocheck = copy.deepcopy(tocheck)
							for pid in tocheck:
								if vect[pid] != '*':
									# found. move process to newdepset
									newdepset.add(pid)
									new_tocheck.remove(pid)
							# update tocheck
							tocheck = new_tocheck
		# next set computed. add it to prop_dependency
		prop_dependency.append(newdepset)
		lastdepset = newdepset

# hide actions of process i w.r.t. given LTS ids
# "accept" is considered a special keyword indicating accepting states.
# hiding this action is not allowed!
def cmc_createHidden_LTS(i, ids, outputtail):
	global procfiles, procsactsets, hideactions
	# set hiding set
	hideactions = copy.deepcopy(procsactsets[i])
	for id in ids:
		hideactions -= procsactsets[id]
	hideactions = list(hideactions)
	# hide actions in file
	hideActions(config.network_folder, procfiles[i][:-4], set(['"accept"']), outputtail)

def cmc_minimiseLTS(procfile):
	# minimise LTS
	rlt = subprocess.Popen(["ltsconvert", "--equivalence=branching-bisim-gjkw", "--tau=i", os.path.join(config.network_folder, procfile + ".aut"), os.path.join(config.network_folder, procfile + "_min.aut")], stderr=subprocess.PIPE)
	out, err = rlt.communicate()
	#print(out)
	#print(err)

# convention:
# - state 0 is initial state
# - state 1 is an accepting sink state (allowing all behaviour)
# - state 2 is a non-accepting state, with transitions for all behaviour going to state 1
def cmc_createInitialConstraintsLTS(i):
	"""create initial constraints LTS for process i"""
	global procs, procsactsets, procfiles, initial_constraintsLTSs
	
	cLTSheader = ['0', str(3*len(procsactsets[i])+2), '3']
	cLTStrans = {}
	first_outgoing = {}
	second_outgoing = {}
	third_outgoing = {}
	for a in procsactsets[i]:
		first_outgoing[a] = set(['1'])
		second_outgoing[a] = set(['1'])
		third_outgoing[a] = set(['1'])
	first_outgoing['"accept"'] = set(['0'])
	second_outgoing['"accept"'] = set(['1'])
	cLTStrans['0'] = first_outgoing
	cLTStrans['1'] = second_outgoing
	cLTStrans['2'] = third_outgoing
	initial_constraintsLTSs[i] = (cLTSheader, cLTStrans)

def cmc_createInitialMinConstraintsLTS(i):
	"""create minimal initial constraints LTS for process i"""
	global procs, procsactsets, procfiles, initial_min_constraintsLTSs
	
	cLTSheader = ['0', str(len(procsactsets[i])+1), '1']
	cLTStrans = {}
	outgoing = {}
	for a in procsactsets[i]:
		outgoing[a] = set(['0'])
	outgoing['"accept"'] = set(['0'])
	cLTStrans['0'] = outgoing
	initial_min_constraintsLTSs[i] = (cLTSheader, cLTStrans)

def cmc_minimiseInitialConstraintsLTSOnFile(i, filename):
	"""minimise LTS in .aut file using strong bisimulation, store result in min_constraintsLTSs"""
	global initial_min_constraintsLTSs
	rlt = subprocess.Popen(["ltsconvert", "--equivalence=bisim-gjkw", os.path.join(config.network_folder, filename + ".aut"), os.path.join(config.network_folder, "constraints_stmin.aut")], stderr=subprocess.PIPE)
	out, err = rlt.communicate()
#	print(out)
#	print(err)
	[h,v,a] = readAut(config.network_folder, "constraints_stmin.aut")
	initial_min_constraintsLTSs[i] = (h,v)

def cmc_minimiseConstraintsLTS(i):
	"""minimise constraints LTS of process i using strong bisimulation"""
	global constraintsLTSs, min_constraintsLTSs
	writeAut(config.network_folder, constraintsLTSs[i][0], constraintsLTSs[i][1], "constraints")
	# minimise LTS
	rlt = subprocess.Popen(["ltsconvert", "--equivalence=bisim-gjkw", os.path.join(config.network_folder, "constraints.aut"), os.path.join(config.network_folder, "constraints_stmin.aut")], stderr=subprocess.PIPE)
	out, err = rlt.communicate()
#	print(out)
#	print(err)
	[h,v,a] = readAut(config.network_folder, "constraints_stmin.aut")
	min_constraintsLTSs[i] = (h,v)

def cmc_extendConstraintsLTS(i, t, allow):
	"""extend the constraintsLTS of process i with the given trace t. There are two options: disable t and extensions (allow==False), or still allow extensions of t (allow=True)"""
	# t is a list of actions indicating a trace, constraints LTS i is a deterministic LTS
	global constraintsLTSs, procsactsets
	
	#print(str(i) + ": " + str(t) + " " + str(allow))
	# get constraintsLTS info
	(header, trans) = constraintsLTSs[i]
	nrstates = int(header[2])
	nrtrans = int(header[1])
	stop = False
	# current state
	current = '0'
	for aid in range(0,len(t)-1):
		outgoing = trans.get(current,{})
		a_tgts = outgoing.get(t[aid], set([]))
		# if there is no suitable target state, this trace should not be accepted, and we can stop
		if a_tgts == set([]):
			stop = True
			break
		tgt = next(iter(a_tgts))
		if int(tgt) <= 2:
			# redirect transition to a new state.
			outgoing[t[aid]] = set([str(nrstates)])
			trans[current] = outgoing
			# add transitions for all actions to the accepting state 1
			tgt_outgoing = {}
			for a in procsactsets[i]:
				tgt_outgoing[a] = set(['1'])
				nrtrans += 1
			# add an accepting self-loop only if the original tgt was accepting
			if tgt == '1':
				tgt_outgoing['"accept"'] = set([str(nrstates)])
				nrtrans += 1
			trans[str(nrstates)] = tgt_outgoing
			# move state
			current = str(nrstates)
			nrstates += 1
		else:
			# move state
			current = tgt
	# final step of t. depending on allow, either disable final step or not but make it go to a non-accepting state
	if not stop:
		outgoing = trans.get(current,{})
		if len(t) == 0:
			# special case: trace has length 0. Remove the 'accept' selfloop of the initial state
			del outgoing['"accept"']
			trans[current] = outgoing
			nrtrans -= 1
		else:
			a = t[len(t)-1]
			a_tgts = outgoing.get(a,set([]))
			if a_tgts != set([]):
				tgt = next(iter(a_tgts))
				if not allow:
					# completely disable this action
					del outgoing[a]
					trans[current] = outgoing
					nrtrans -= 1
				elif tgt != '2':
					# redirect transition to state 2
					outgoing[a] = set(['2'])
					trans[current] = outgoing
	# update header
	constraintsLTSs[i] = ([header[0], str(nrtrans), str(nrstates)], trans)

# construct environment LTS for process i, consisting of LTSs in ids (ids is a sorted list!)
def cmc_constructEnvironmentLTS(i, ids, outputtail):
	global procfiles, procsactsets, syncactions, syncrules
	env_procs = []
	env_procsheaders = []
	env_syncrules = []
	env_syncactions = []
	# first create an abstract LTS for each process LTS in environment
	for j in ids:
		cmc_createHidden_LTS(j, set([i]), outputtail)
		cmc_minimiseLTS(procfiles[j][:-4] + "_hidden" + outputtail)
		# read and store new process LTS
		[h,v,a] = readAut(config.network_folder, procfiles[j][:-4] + "_hidden" + outputtail + "_min.aut")
		env_procs.append(v)
		env_procsheaders.append(h)
		env_syncrules.append({})
		env_syncactions.append(set([]))
	# create new syncrules
	for j in range(0,len(syncrules)):
		for a, v in syncrules[j].items():
			if a in syncactions[i]:
				for (vect, act) in v:
					if vect[i] != '*':
						newvect = []
						# firstid contains first id of LTS with an entry != '*'
						firstid = len(syncrules)
						l = 0
						countproc = 0
						for k in ids:
							newvect.append(vect[k])
							if vect[k] != '*':
								countproc += 1
								if l < firstid:
									firstid = l
							l += 1
						# only add rule if countproc > 1
						if countproc > 1:
							alist = env_syncrules[firstid].get(a,[])
							alist.append((newvect,a))
							env_syncrules[firstid][a] = alist
							# add action to env_syncactions sets
							for k in range(0,len(ids)):
								if newvect[k] != '*':
									env_syncactions[k].add(a)
	# generate environment LTS
	generateLTS(env_procsheaders, env_procs, [], env_syncrules, env_syncactions, os.path.join(config.network_folder, procfiles[i][:-4] + "_env" + outputtail), opt_quiet, config.outputobject, False, list(range(0,len(env_procsheaders))), set([]), False, False)
	# finally, minimise result
	cmc_minimiseLTS(procfiles[i][:-4] + "_env" + outputtail)

def cmc_update_tree(i):
	"""update the constraints tree from position i in the current check configuration"""
	global cmc_checktree, cmc_checklist, constraintsLTSs, procfiles, constraintsLTS_is_extended

	seqname = "cons_"
	children1 = cmc_checktree
	for j in range(0,len(cmc_checklist)-1):
		cid = cmc_checklist[j][0]
		seqname += str(cid) + "_"
		children2 = children1[0].get(cid,({},set([])))
		if children2[0] == {}:
			children1[0][cid] = ({},set([]))
		if j >= i and constraintsLTS_is_extended[cid]:
			children1[0][cid] = (children2[0],children2[1].add(cid))
			# possibly add initial constraints LTS to root of tree
			if j == 0:
				cmc_checktree[1].add(cid)
			writeAut(config.network_folder, constraintsLTSs[cmc_checklist[j][0]][0], constraintsLTSs[cmc_checklist[j][0]][1], seqname + procfiles[cmc_checklist[j][0]][:-4])
			# no longer needed to keep track of extension
			constraintsLTS_is_extended[cid] = False
		children1 = children2

def cmc_add_to_constraints_updates(i, stems_from_shift):
	"""add an update to the constraints updates list"""
	global cmc_checklist, cmc_constraints_updates, procs
	
	if stems_from_shift:
		# indicate at current position that a shift took place
		(t,f) = cmc_constraints_updates[cmc_checklist[i][0]]
		if t > i:
			cmc_constraints_updates[cmc_checklist[i][0]] = (i,len(procs))
	elif i < len(cmc_checklist)-1:
		(t,f) = cmc_constraints_updates[cmc_checklist[i+1][0]]
		if f > i:
			cmc_constraints_updates[cmc_checklist[i+1][0]] = (t,i)
			

def cmc_move_constraints_updates(i):
	"""move the constraints updates recorded for position i in the current check configuration one place to the right (unless the end has been reached)"""
	global cmc_checklist, cmc_constraints_updates, procs

	cid = cmc_checklist[i][0]
	if i < len(cmc_checklist)-1:
		cidp = cmc_checklist[i+1][0]
		(t1,f1) = cmc_constraints_updates[cid]
		(t2,f2) = cmc_constraints_updates[cidp]
		cmc_constraints_updates[cidp] = (min(t1,t2),min(f1,f2))
	cmc_constraints_updates[cid] = (len(procs),len(procs))

def cmc_find_new_initial_constraints_LTS(i, remove_pos):
	"""find a new initial constraints LTS for the process at position i in the current check configuration.
	   Remove any files containing constraints that are no longer valid (i.e. refer to constraints at a position higher than
	   or equal to remove_pos)"""
	global cmc_checktree, cmc_checklist, initial_constraintsLTSs, constraintsLTSs, constraintsLTS_is_extended, cmc_constraintsdep, procfiles

	cid = cmc_checklist[i][0]
	seqname = "cons_"
	last_found = "cons_"
	last_found_pos = -1
	children1 = cmc_checktree
	k = 0
	while children1[0] != {} and k < i:
		ckd = cmc_checklist[k][0]
		children2 = children1[0].get(ckd,({},set([])))
		seqname += str(ckd) + "_"
		if cid in children2[1]:
			if k < remove_pos:
				last_found = copy.deepcopy(seqname)
				last_found_pos = k
			else:
				# this part of the check configuration has been reset. stored file is no longer relevant
				os.remove(os.path.join(config.network_folder, seqname + procfiles[cid][:-4]))
				S = children2[1]
				S.remove(i)
				children2 = (children2[0],S)
		children1 = children2
		k += 1
	if last_found == "cons_":
		# check for stored initial constraints LTS
		if cid in cmc_checktree[1]:
			last_found += str(cid) + "_"
		else:
			cmc_createInitialConstraintsLTS(cid)
			cmc_createInitialMinConstraintsLTS(cid)
			cmc_constraintsdep[cid] = -1
	if last_found != "cons_":
		# read
		filename = last_found + procfiles[cid][:-4]
		[h,v,a] = readAut(config.network_folder, filename + ".aut")
		initial_constraintsLTSs[cid] = (h,v)
		cmc_minimiseInitialConstraintsLTSOnFile(cid, os.path.join(config.network_folder, filename + ".aut"))
		cmc_constraintsdep[cid] = last_found_pos

def cmc_process_remaining_constraints_updates(i):
	"""before a shift is done, process any remaining updates performed in the current check configuration, starting at position i"""
	global cmc_checklist, initial_min_constraintsLTSs, initial_constraintsLTSs, constraintsLTSs, min_constraintsLTSs, constraintsLTS_is_extended, cmc_constraintsdep, cmc_constraints_updates, procs

	for j in range(i,len(cmc_checklist)):
		cid = cmc_checklist[j][0]
		(t,f) = cmc_constraints_updates[cid]
		if f != len(procs):
			# possibly find new initial constraints LTS
			if f <= cmc_constraintsdep[cid]:
				cmc_find_new_initial_constraints_LTS(j,f)
			# update current constraints LTS
			if constraintsLTS_is_extended[cid]:
				constraintsLTSs[cid] = copy.deepcopy(initial_constraintsLTSs[cid])
				min_constraintsLTSs[cid] = copy.deepcopy(initial_min_constraintsLTSs[cid])
				# constraints LTS is no longer extended
				constraintsLTS_is_extended[cid] = False
			# update the constraints updates list
			cmc_move_constraints_updates(j)

def cmc_update_constraints_LTSs(i):
	"""if necessary, find new initial constraints LTS for process at position i in current check configuration, and/or update current constraints LTS"""
	global cmc_checktree, cmc_checklist, initial_min_constraintsLTSs, initial_constraintsLTSs, constraintsLTSs, min_constraintsLTSs, constraintsLTS_is_extended, cmc_constraintsdep, cmc_constraints_updates, procs

	cid = cmc_checklist[i][0]
	(t,f) = cmc_constraints_updates[cid]
	if t != len(procs) or f != len(procs):
		# condition for updating the initial constraints LTS
		if t < len(procs) or f <= cmc_constraintsdep[cid]:
			# shift happened somewhere in the checklist. check for new initial constraints LTS
			cmc_find_new_initial_constraints_LTS(i,f)
		if t == i or f < len(procs):
			# on this position a shift took place. Set current constraints LTS to the initial constraints LTS
			constraintsLTSs[cid] = copy.deepcopy(initial_constraintsLTSs[cid])
			min_constraintsLTSs[cid] = copy.deepcopy(initial_min_constraintsLTSs[cid])
			# constraints LTS is no longer extended
			constraintsLTS_is_extended[cid] = False
		# update the constraints updates list
		cmc_move_constraints_updates(i)

def cmc_checklist_insert_abstract_property_LTS(i):
	"""insert an abstract property LTS for process at position i in current check configuration"""
	global cmc_checklist, propertyLTSs, procfiles, checkprocs, checkheaders, procs
	# insert an abstracted, minimised version of the property LTS. if it does not yet exist on disk, create it first

	(pid,c) = cmc_checklist[i]
	if propertyLTSs[pid] == ():
		propproclist = sorted([j for (j,c) in cmc_checklist[:i+1]])
		propprocliststring = ""
		for j in range(0,len(propproclist)):
			propprocliststring += str(propproclist[j]) + "_"
		propprocliststring = propprocliststring[:-1]
		if not os.path.isfile(os.path.join(config.network_folder, procfiles[len(procs)-1][:-4] + "_hidden" + propprocliststring + "_min.aut")):
			S = set(propproclist)
			cmc_createHidden_LTS(len(procs)-1, S, propprocliststring)
			cmc_minimiseLTS(procfiles[len(procs)-1][:-4] + "_hidden" + propprocliststring)
		[h,v,a] = readAut(config.network_folder, procfiles[len(procs)-1][:-4] + "_hidden" + propprocliststring + "_min.aut")
		propertyLTSs[pid] = (h,v)
	checkprocs[len(procs)-1] = propertyLTSs[pid][1]
	checkheaders[len(procs)-1] = propertyLTSs[pid][0]

def cmc_check():
	"""apply compositional model checking to verify LTS network"""
	global procfiles, procsheaders, procs, syncrules, syncactions, procsactprefixsets, procsactsets, initial_constraintsLTSs, constraintsLTSs, initial_min_constraintsLTSs, min_constraintsLTSs, propertyLTSs, cmc_syncrules, cmc_syncactions, cmc_constraintsdep, cmc_checktree, cmc_checklist, cmc_constraints_updates, checkprocs, checkheaders, constraintsLTS_is_extended
	
	# variables used to keep track of memory statistics
	max_states_in_check = 0
	states_in_check = 0
	# count the number of checks performed
	nrchecks = 0
	
	# are the constraints LTSs extended?
	constraintsLTS_is_extended = []
	
	# upperbound to number of traces considered for a process before it is moved in the checklist
	checkbound = 10
	# bound on number of shifts per process
	shiftbound = 10

	# create extended syncrules for each process in the network. Extensions are needed to allow for
	# correct inclusion of a constraints LTS
	cmc_syncrules = []
	cmc_syncactions = []
	for i in range(0,len(procs)-1):
		cmc_syncrules.append([])
		for j in range(0,len(syncrules)):
			cmc_syncrules[i].append({})
			for a, v in syncrules[j].items():
				newv = []
				for (vect, act) in v:
					newvect = copy.deepcopy(vect)
					if vect[i] != '*':
						# extend with action
						newvect.append(vect[i])
					else:
						newvect.append('*')
					newv.append((newvect, act))
				cmc_syncrules[i][j][a] = newv
		# add empty dict for constraints LTS
		cmc_syncrules[i].append({})
		# finally, add syncrules for all actions of process i not already syncing. Let them sync with the constraints LTS
		for a in (procsactprefixsets[i] - syncactions[i]):
			newvect = []
			for j in range(0,len(procs)):
				if j != i:
					newvect.append('*')
				else:
					newvect.append(a)
			newvect.append(a)
			cmc_syncrules[i][i][a] = [(newvect, a)]
		# create suitable syncactions entry for these syncrules
		copy_syncactions = copy.deepcopy(syncactions)
		# extend by one for constraints LTS
		copy_syncactions.append(set([]))
		copy_syncactions[i] = procsactprefixsets[i]
		copy_syncactions[len(copy_syncactions)-1] = procsactprefixsets[i]
		cmc_syncactions.append(copy_syncactions)
	# add one additional entry in procsactsets for constraints LTS
	procsactsets.append(set([]))
	
	# we use a list of tuples (i,c), with i a process, and c a counter to indicate how many traces
	# from process i have been considered in one sequence in the current check
	cmc_checklist = []
	# processes for check network
	checkprocs = []
	checkheaders = []
	# tree to keep track of constraintsLTSs for different check configurations
	cmc_checktree = ({},set([]))
	# constraints dependency function: on which current position in check does the constraints LTS of process i depend?
	cmc_constraintsdep = [-1 for i in range(0,len(procs)-1)]

	# construct initial constraintsLTSs, property LTSs, and updates list
	initial_constraintsLTSs = []
	constraintsLTSs = []
	initial_min_constraintsLTSs = []
	min_constraintsLTSs = []
	propertyLTSs = []
	constraintsLTS_is_extended = []
	cmc_constraints_updates = []
	for i in range(0,len(procs)-1):
		initial_constraintsLTSs.append(())
		constraintsLTSs.append(())
		initial_min_constraintsLTSs.append(())
		min_constraintsLTSs.append(())
		propertyLTSs.append(())
		cmc_createInitialConstraintsLTS(i)
		constraintsLTSs[i] = copy.deepcopy(initial_constraintsLTSs[i])
		cmc_createInitialMinConstraintsLTS(i)
		min_constraintsLTSs[i] = copy.deepcopy(initial_min_constraintsLTSs[i])
		constraintsLTS_is_extended.append(False)
		cmc_constraints_updates.append((len(procs),len(procs)))

	# based on property dependencies, fix an initial check order for processes
	for i in range(0,len(prop_dependency)):
		while prop_dependency[i] != set([]):
			p = prop_dependency[i].pop()
			cmc_checklist.append((p,0))
			checkprocs.append({})
			checkheaders.append(['0','0','0'])
	# two extra procs and headers for property LTS and constraints LTS
	checkprocs.append({})
	checkheaders.append(['0','0','0'])
	checkprocs.append({})
	checkheaders.append(['0','0','0'])

	# list to keep track of the number of shifts done per process. if this hits the upperbound, we shift from the start of the list
	shifts_done = [0 for j in range(0,len(cmc_checklist))]

	# index of currently considered position in cmc_checklist
	check_current = 0
	while check_current < len(cmc_checklist):
	#while check_current < 1:
		(i,c) = cmc_checklist[check_current]
		if (c == checkbound and check_current < len(cmc_checklist)-1) or (shifts_done[i] == shiftbound):
			# move all processes beyond this point one step to the right in cmc_checklist. move final process to current position in cmc_checklist
			# forget current abstracted, minimised version of the property LTS for the involved processes
			# store the current constraints LTSs on disk, taking into account the current order of processes in cmc_checklist
			# and try to read new constraints LTSs from disk, taking into account the new order
			# increase shift count
			if shifts_done[i] == shiftbound:
				check_current = 0
				shifts_done = [0 for j in range(0,len(cmc_checklist))]
			else:
				shifts_done[i] += 1

			for j in range(check_current,len(cmc_checklist)):
				propertyLTSs[j] = ()
			# process any constraints updates still in the pipeline
			cmc_process_remaining_constraints_updates(check_current)
			# store current constraints LTSs
			cmc_update_tree(check_current)
			# move
			tmp = cmc_checklist[len(cmc_checklist)-1][0]
			j = len(cmc_checklist)-2
			while j >= check_current:
				cmc_checklist[j+1] = (cmc_checklist[j][0],0)
				checkprocs[cmc_checklist[j+1][0]] = {}
				checkheaders[cmc_checklist[j+1][0]] = ['0','0','0']
				j -= 1
			cmc_checklist[check_current] = (tmp,0)
			checkprocs[cmc_checklist[check_current][0]] = {}
			checkheaders[cmc_checklist[check_current][0]] = ['0','0','0']
			
			# update the update list
			cmc_add_to_constraints_updates(check_current,True)
		else:
			# new check
			nrchecks += 1
			# possibly update constraints LTSs
			cmc_update_constraints_LTSs(check_current)
			# keep track of number of states in check
			states_in_check += int(min_constraintsLTSs[i][0][2])
			# increment counter
			cmc_checklist[check_current] = (i,c+1)
			# create check network for this process, and perform the check
			# add current process and its constraints LTS to the network
			checkprocs[i] = procs[i]
			checkheaders[i] = procsheaders[i]
			checkprocs[len(procs)] = min_constraintsLTSs[i][1]
			checkheaders[len(procs)] = min_constraintsLTSs[i][0]
			# add property LTS
			cmc_checklist_insert_abstract_property_LTS(check_current)
			# check
			if os.path.isfile(os.path.join(config.network_folder, "check-ce.aut")):
				os.remove(os.path.join(config.network_folder, "check-ce.aut"))
			if os.path.isfile(os.path.join(config.network_folder, "check-extrace")):
				os.remove(os.path.join(config.network_folder, "check-extrace"))
			generateLTS(checkheaders, checkprocs, procsactsets, cmc_syncrules[i], cmc_syncactions[i], os.path.join(config.network_folder, "check"), opt_quiet, config.outputobject, 0, [j for (j,c) in cmc_checklist[:check_current+1]] + [len(procs)-1,len(procs)], set([len(procs)-1,len(procs)]), True, check_current != 0)
			# for monitoring reasons: read size of generated LTS
			inFile = open(os.path.join(config.network_folder, "check.aut"), 'r')
			# scan the first line
			line = inFile.readline()
			global token, next
			results, remainder = autscanner.scan(line)
			inFile.close()
			# read the header
			# an aut file has three values in the header
			news = int(results[3][1])
			#newt = int(results[2][1])
			if (states_in_check+news) > max_states_in_check:
				max_states_in_check = states_in_check+news
				#max_transitions_generated = newt
			# is there a counter-example? if so, read it
			if os.path.isfile(os.path.join(config.network_folder, "check-ce.aut")):
				[h,v,a] = readAut(config.network_folder, "check-ce.aut")
				# extract from this LTS a smaller LTS consisting of only behaviour from process i
				newv = {}
				trace_src = h[0]
				proc_srcs = set([procsheaders[i][0]])
				newstate = 0
				while trace_src != h[2]:
					outgoing = v.get(trace_src,{})
					if outgoing == {}:
						trace_src = h[2]
					else:
						newoutgoing = {}
						for a, tgts in outgoing.items():
							proc_tgts = set([])
							for src in proc_srcs:
								proc_outgoing = procs[i].get(src,{})
								proc_tgts |= proc_outgoing.get(a,set([]))
							if proc_tgts != set([]):
								newoutgoing[a] = set([str(newstate+1)])
								newv[str(newstate)] = newoutgoing
								newstate += 1
							if a == '"tau"' or a == '"i"':
								proc_tgts |= proc_srcs
							if proc_tgts != set([]):
								proc_srcs = proc_tgts
							trace_src = list(tgts)[0]
				newh = ['0',str(newstate),str(newstate+1)]
				# place this counter-example in check network, and move forward in checking
				checkprocs[i] = newv
				checkheaders[i] = newh
				check_current += 1
			else:
				# no counter-example. if we are looking at the first process, we have established that no counter-example exists at all.
				# else, take explored trace into account and move back to previous process in cmc_checklist
				if os.path.isfile(os.path.join(config.network_folder, "check-extrace")):
					# read trace
					t = []
					for tline in open(os.path.join(config.network_folder, "check-extrace")):
						t.append(tline)
					# strip newlines
					t = [tr.strip('\n') for tr in t]
					# extend the constraints LTS of the process for which the constraint was reported
					prev_pos = int(t[0])
					# number of states in check
					states_in_check -= int(min_constraintsLTSs[cmc_checklist[prev_pos][0]][0][2])
					cmc_extendConstraintsLTS(cmc_checklist[prev_pos][0], t[1:-1], t[len(t)-1] == 'True')
					cmc_minimiseConstraintsLTS(cmc_checklist[prev_pos][0])
					constraintsLTS_is_extended[cmc_checklist[prev_pos][0]] = True
					#number of states in check
					states_in_check += int(min_constraintsLTSs[cmc_checklist[prev_pos][0]][0][2])
					# record change
					cmc_add_to_constraints_updates(prev_pos,False)
					# move back
					while check_current > prev_pos:
						# update number of states in check
						states_in_check -= int(min_constraintsLTSs[cmc_checklist[check_current][0]][0][2])
						# clear LTS from check
						checkheaders[cmc_checklist[check_current][0]] = ['0','0','0']
						checkprocs[cmc_checklist[check_current][0]] = {}
						check_current -= 1
					states_in_check -= int(min_constraintsLTSs[cmc_checklist[prev_pos][0]][0][2])
				else:
					break
	# report
	if check_current == 0:
		print_message('Network satisfies property!')
	else:
		print_message('Network does NOT satisfy property!')
		for i in range(0,len(procs)-1):
			print("process " + str(i) + ":")
			for j in range(0,int(checkheaders[i][2])):
				outgoing = checkprocs[i].get(str(j),{})
				print(str(j) + ": " + str(outgoing))
	print_message('Largest number of states in a check: ' + str(max_states_in_check))
	print_message('Number of checks performed: ' + str(nrchecks))
	# remove constraints files on disk
	for fl in glob.glob(os.path.join(config.network_folder, "cons_*")):
		os.remove(fl)

# --- the main functions for verification / transformation ---

def refiner_readNetwork(quiet):
	global inputsync, syncrules, syncactions, procs, procsactsets, procsheaders, inputnetwork, procs_orig_size, procfiles, tmpsynclist, tmpsynccounter
	global tmplists, tmpcounter, lastsyncaction, acttmplist, tmprule, proccounter, revprocs, actprocs, procsactprefixsets
	proccounter = 0
	procfiles = []
	procsheaders = []
	procs = []
	procs_orig_size = 0
	revprocs = []
	actprocs = []
	procsactsets = []
	procsactprefixsets = []
	inputsync = {}
	syncrules = {}
	syncactions = []
	tmpsynclist = []
	tmpsynccounter = -1
	inputnetwork = []
	tmplists = [[]]
	tmpcounter = 0
	lastsyncaction = ""
	acttmplist = []
	tmprule = ""
	if not quiet:
		print_message('reading network ' + config.network_name)
		print_separator()
	readNetwork(config.network_name, quiet)
	if not quiet:
		print_message('finished reading network ' + config.network_name)
		print_message('---')

def refiner_readRulesystem():
	print_message('reading rule system ' + config.rs_name)
	print_separator()
	readRulesystem(os.path.join(config.rs_folder, config.rs_name))
	print_message('finished reading rule system ' + config.rs_name)
	print_message('---')

def refiner_constructhidingset():
	# derive the action hiding set from a property
	global hideactions
	if config.property_name == "":
		hideactions = ['none']
	else:
		hideactions = []
		print_message('constructing hiding set')
		print_separator()
		constructHidingSet()
		print_message('finished reading property ' + config.property_name)
		print_message('---')

def refiner_generateLTS():
	global opt_safety_onthefly, opt_genbound, procsheaders, procs, syncrules, syncactions
	print_message('generating system LTS')
	if opt_genbound > 0:
		print_message('generation bound is ' + str(opt_genbound))
	if opt_safety_onthefly:
		print_message('checking safety property on-the-fly')
	print_separator()
	if opt_safety_onthefly:
		generateLTS(procsheaders, procs, [], syncrules, syncactions, os.path.join(config.network_folder, config.network_name), opt_quiet, config.outputobject, opt_genbound, list(range(0,len(procsheaders))), set([len(procsheaders)-1]), False, False)	
	else:
		generateLTS(procsheaders, procs, [], syncrules, syncactions, os.path.join(config.network_folder, config.network_name), opt_quiet, config.outputobject, opt_genbound, list(range(0,len(procsheaders))), set([]), False, False)

def refiner_preparedivinfo():
	# check if the sets of process nondiverging states is already available on disk
	global revhashtable, procsheaders, procs, syncrules, syncactions
	opt_divergency = True
	if not os.path.isfile(os.path.join(config.network_folder, config.network_name + "_" + config.property_name + ".ndv")):
		# check if the network is present on disk.
		if not os.path.isfile(os.path.join(config.network_folder, config.network_name + ".aut")):
			print_message("system .aut file does not exist, it will be generated!")
			print_separator()
			print_message('generating system LTS')
			print_separator()
			generateLTS(procsheaders, procs, [], syncrules, syncactions, os.path.join(config.network_folder, config.network_name), opt_quiet, config.outputobject, opt_genbound, list(range(0,len(procsheaders))), set([]), False, False)
		print_separator()
		print_message("reading system LTS")
		print_separator()
		[systemLTSheader, systemtrans, systemactset] = readAut(config.network_folder, config.network_name + ".aut")
		print_message("reading state mapping")
		print_separator()
		revhashtable = readMap(config.network_folder, config.network_name)
		if hideactions == []:
			print_message('reading hiding set')
			print_separator()
			try:
				inFile = open(os.path.join(config.property_folder, config.property_name + ".hs"), 'r')
			except IOError:
				print_error("action hiding file does not exist!")
			readHideActions(inFile)
			inFile.close()
		# compute the set of tau-diverging states
		print_message('computing set of tau-diverging system states')
		divergencyset = computeDivergencySet(systemLTSheader, systemtrans)
		print_separator()
		print_message('computing the non-diverging process states')
		computeNonDivergingProcessStates()
		print_separator()
		print_message('writing non-diverging process states to disk')
		writeProcndivsets(config.network_name, config.property_name)
		print_separator()

def refiner_transform(div):
	global opt_divergency, transformed_procs_new_states, transformed_procs, transformed_procsheaders, procfiles, newprocfiles, procs, procsheaders
	transformed_procs_new_states = []
	transformed_procs = []
	transformed_procsheaders = []
	opt_divergency = div
	if opt_divergency:
		refiner_preparedivinfo()
	preprocessRulesystem()
	print_message('checking confluence of rule system ' + config.rs_name)
	print_separator()
	checkConfluence()
	# check if divergency info needs to be updated when transforming
	if config.property_name != "":
		try:
			inFile = open(os.path.join(config.network_folder, config.network_name + "_" + config.property_name + ".ndv"), 'r')
		except IOError:
			opt_divergency = False
			inFile = 0
		if inFile != 0:
			opt_divergency = True
	else:
		opt_divergency = False
	if opt_divergency:
		print_message('reading non-diverging process states list')
		print_separator()
		readProcndivsets(config.network_name, config.property_name)
	print_message('determining matches of rules in ' + config.rs_name + ' on process LTSs of ' + config.network_name)
	print_separator()
	matchRulesystem()
	print_separator()
	print_message('checking applicability of synchronisation rules of ' + config.rs_name + ' on those of ' + config.network_name)
	print_separator()
	matchSyncrules()
	print_message('refining the process LTSs of ' + config.network_name)
	print_separator()
	transform()
	print_separator()
	if opt_divergency:
		print_message('updating the lists of non-diverging states of ' + config.network_name)
		print_separator()
		updateNonDivergingProcessStates()
		print_message('writing updated non-diverging process states')
		print_separator()
		writeNewProcndivsets(config.network_name, config.property_name)
	print_message('writing the new network')
	print_separator()
	writeNetwork(config.network_folder, transformed_procs, transformed_procsheaders, procfiles, newsyncrules, config.network_name, opt_expopenoutput)
	print_message('TRANSFORMATION FINISHED: network ' + config.network_name + ' has been transformed and saved as network ' + config.transformed_network_name)
	print_message('---')
	# transformed network becomes new input network
	config.network_name = config.transformed_network_name
	refiner_readNetwork(True)

def refiner_verify(div, fair):
	global opt_divergency, opt_fairness
	opt_divergency = div
	opt_fairness = fair
	if opt_divergency:
		refiner_preparedivinfo()
	preprocessRulesystem()
	print_message('checking confluence of rule system ' + config.rs_name)
	print_separator()
	checkConfluence()
	if opt_divergency:
		print_message('reading non-diverging process states list of ' + config.network_name)
		print_separator()
		readProcndivsets(config.network_name, config.property_name)
		print_message('determining matches of rules of ' + config.rs_name + ' on process LTSs of ' + config.network_name)
		print_separator()
		matchRulesystem()
		print_separator()
		print_message('checking applicability of synchronisation rules of ' + config.rs_name)
		print_separator()
		matchSyncrules()
	print_message('constructing preservation checks for rule system ' + config.rs_name)
	print_separator()
	constructChecks()
	if hideactions == []:
		print_message('reading hiding set')
		print_separator()
		try:
			inFile = open(os.path.join(config.property_folder, config.property_name + ".hs"), 'r')
		except IOError:
			print_error("action hiding file does not exist!")
		readHideActions(inFile)
		inFile.close()
	if config.property_name == '':
		config.property_name = 'semantics'
	print_message('checking preservation of ' + config.property_name + ' for ' + config.rs_name)
	print_separator()
	checkPropPreservation()

# IMPORTANT: for compositional mc, we currently assume that the network is structured in such a way, that for all actions a,
# a process can only fire a if all other processes that have a in their alphabet can also fire a in the current state.
# Support for other networks is future work.
def refiner_compositional_mc():
	global prop_dependency
	# first we partition the process LTSs based on their dependency w.r.t. the property (final LTS in the network)
	print_message('determining dependency of process LTSs on property')
	print_separator()
	cmc_constructDependency()
	print_message('checking')
	cmc_check()

# --- END main functions

def main(args):
	"""The main function"""
	global opt_prop, opt_constructhidingset, opt_generate, opt_transform, opt_check, opt_divergency, opt_fairness, opt_compositional_mc, opt_safety_onthefly
	global opt_expopenoutput, opt_quiet, opt_network, hideactions, opt_genbound
	for i in range(0,len(args)):
		if args[i] == '-h':
			print("Usage: pypy/python3 refiner.py [-gprcdfeqtCn]")
			print("")
			print("Refine a given network of LTSs according to a given rule system,")
			print("or analyse either such a network or such a rule system")
			print("")
			print(" -g                    generate the LTS of the given network")
			print(" -o                    generate the LTS, and check on-the-fly safety property included in network")
			print(" -p <prop>             determine property preservation of mu-calculus formula <prop>.mcl")
			print(" -r <rule system>      refine a given network according to rule system <rule system>.rs")
			print(" -c <rule system>      check property preservation for <rule system>.rs. Select a property with -p")
			print(" -d                    take divergency in processes into account")
			print(" -f                    check property preservation under fairness (all divergency is ignored)")
			print(" -e                    when refining, write resulting network in exp.open form (i.e. add one-party sync. rules)")
			print(" -q                    quiet mode, no messages to screen")
			print(" -t <n>                use n threads in parallel for property preservation checking")
			print(" -icc                  apply compositional model checking with Incremental Counter-Example Construction (assumes presence of a property LTS)")
			print(" -n <network>          the input network <network>.exp")
			print(" -b <bound>            perform a bounded search (state space generation)")
		elif args[i] == '-p':
			i += 1
			opt_prop = args[i]
		elif args[i] == '-g':
			opt_generate = True
		elif args[i] == '-o':
			opt_safety_onthefly = True
		elif args[i] == '-r':
			i += 1
			opt_transform = args[i]
		elif args[i] == '-c':
			i += 1
			opt_check = args[i]
		elif args[i] == '-d':
			opt_divergency = True
		elif args[i] == '-f':
			opt_fairness = True
		elif args[i] == '-e':
			opt_expopenoutput = True
		elif args[i] == '-q':
			opt_quiet = True
		elif args[i] == '-t':
			i += 1
			config.nrthreads = int(args[i])
		elif args[i] == '-icc':
			i += 1
			opt_compositional_mc = True
		elif args[i] == '-n':
			i += 1
			opt_network = args[i]
		elif args[i] == '-b':
			i += 1
			opt_genbound = int(args[i])

	# check that we have all required input
	if opt_generate or opt_transform != "" or opt_divergency or opt_compositional_mc:
		if opt_network == "":
			print_error("no network specified!")
	if opt_check != "":
		if opt_prop == "":
			#print_error("no property specified!")
			hideactions = ['none']

	if opt_compositional_mc and (opt_generate or opt_transform or opt_divergency or opt_check):
		print_error("combination of options is invalid!")

	if opt_network != "":
		config.network_folder, config.network_name = os.path.split(opt_network)
		refiner_readNetwork(False)

	if opt_generate or opt_safety_onthefly:
		# generate the LTS of the given network
		refiner_generateLTS()

	if opt_prop:
		config.property_folder, config.property_name = os.path.split(opt_prop)
		if not os.path.isfile(os.path.join(config.property_folder, config.property_name + ".hs")):
				opt_constructhidingset = True

	if opt_constructhidingset:
		refiner_constructhidingset()

	if opt_transform != "":
		config.rs_folder, config.rs_name = os.path.split(opt_transform)
		refiner_readRulesystem()
		refiner_transform(opt_divergency)

	if opt_check != "":
		config.rs_folder, config.rs_name = os.path.split(opt_check)
		refiner_readRulesystem()
		refiner_verify(opt_divergency, opt_fairness)

	if opt_compositional_mc:
		refiner_compositional_mc()

if __name__ == '__main__':
	args = []
	for i in range(1,len(sys.argv)):
		args.append(sys.argv[i])
	main(args)
