class Process:
  def __init__( self ):
    self.next_id = 0
    self.states = {}
    self.glue_states = {}
    self.exclusive_glue_states = {}
    self.transitions = {}

  def add_state( self, x, y ):
    self.states[ self.next_id ] = { "id": self.next_id, "x": x, "y": y }
    self.next_id += 1

  def add_glue_state( self, x, y ):
    self.glue_states[ self.next_id ] = { "id": self.next_id, "x": x, "y": y }
    self.next_id += 1
    self.relabel()

  def add_exclusive_glue_state( self, x, y ):
    self.exclusive_glue_states[ self.next_id ] = { "id": self.next_id, "x": x, "y": y }
    self.next_id += 1
    self.relabel()

  def get_related_loop_transitions( self, state ):
    return [ index for index in self.transitions if self.transitions[ index ][ "source" ][ "id" ] == state and self.transitions[ index ][ "target" ][ "id" ] == state ]
    
  def get_related_line_transitions( self, state ):
    return [ index for index in self.transitions
             if ( self.transitions[ index ][ "source" ][ "id" ] == state or self.transitions[ index ][ "target" ][ "id" ] == state ) and
                ( self.transitions[ index ][ "source" ] != self.transitions[ index ][ "target" ] )
           ]
  
  def get_related_transitions( self, state ):
    return [ index for index in self.transitions if self.transitions[ index ][ "source" ][ "id" ] == state or self.transitions[ index ][ "target" ][ "id" ] == state ]

  def delete_related_transitions( self, index ):
    delete = []
    for transition_index in self.transitions:
      if self.transitions[ transition_index ][ "source" ][ "id" ] == index:
        delete.append( transition_index )
      if self.transitions[ transition_index ][ "target" ][ "id" ] == index:
        delete.append( transition_index )
    for transition_index in delete:
      del self.transitions[ transition_index ]
    
  def delete( self, index ):
    if index in self.states:
      self.delete_related_transitions( index )
      del self.states[ index ]
    if index in self.glue_states:
      self.delete_related_transitions( index )
      del self.glue_states[ index ]
    if index in self.exclusive_glue_states:
      self.delete_related_transitions( index )
      del self.exclusive_glue_states[ index ]
    if index in self.transitions:
      del self.transitions[ index ]
    self.relabel()

  def get_state( self, id ):
    if id in self.states:
      return self.states[ id ]
    if id in self.glue_states:
      return self.glue_states[ id ]
    if id in self.exclusive_glue_states:
      return self.exclusive_glue_states[ id ]
      
  def add_transition( self, source_id, target_id, label, c1x, c1y, c2x, c2y, lx, ly ):
    source = self.get_state( source_id )
    target = self.get_state( target_id )
    self.transitions[ self.next_id ] = { "id": self.next_id, "source": source, "target": target, "label": label, "c1x": c1x, "c1y": c1y, "c2x": c2x, "c2y": c2y, "lx": lx, "ly": ly }
    self.next_id += 1
    
  def relabel( self ):
    new_exclusive_glue_states = {}
    new_glue_states = {}
    new_states = {}
    new_transitions = {}
    index = 0
    for exclusive_glue_state in self.exclusive_glue_states:
      new_exclusive_glue_states[ index ] = self.exclusive_glue_states[ exclusive_glue_state ]
      new_exclusive_glue_states[ index ][ "id" ] = index
      index += 1
    for glue_state in self.glue_states:
      new_glue_states[ index ] = self.glue_states[ glue_state ]
      new_glue_states[ index ][ "id" ] = index
      index += 1
    for state in self.states:
      new_states[ index ] = self.states[ state ]
      new_states[ index ][ "id" ] = index
      index += 1
    for transition in self.transitions:
      new_transitions[ index ] = self.transitions[ transition ]
      new_transitions[ index ][ "id" ] = index
      index += 1
    self.exclusive_glue_states = new_exclusive_glue_states
    self.glue_states = new_glue_states
    self.states = new_states
    self.transitions = new_transitions
    self.next_id = index

class Rule:
  def __init__( self ):
    self.left = Process()
    self.right = Process()

  def setPatterns( self, lpr, rpr ):
    self.left = lpr
    self.right = rpr
