from tkinter import *
import cmath, math
import tkinter.simpledialog as simpledialog

class _Buttons:
  def __init__( self, frame, parent ):
    self.parent = parent

    self.state_button = Button( frame, text="State", command=self.select_state_tool, state=DISABLED )
    self.glue_state_button = Button( frame, text="Glue state", command=self.select_glue_state_tool, state=DISABLED )
    self.exclusive_glue_state_button = Button( frame, text="Exclusive glue state", command=self.select_exclusive_glue_state_tool, state=DISABLED, width="16" )
    self.transition_button = Button( frame, text="Transition", command=self.select_transition_tool, state=DISABLED )
    self.delete_button = Button( frame, text="Delete", command=self.select_delete_tool, state=DISABLED )
    self.state_button.pack( fill=X )
    self.glue_state_button.pack( fill=X )
    self.exclusive_glue_state_button.pack( fill=X )
    self.transition_button.pack( fill=X )
    self.delete_button.pack( fill=X )

  def select_state_tool( self ):
    self.parent.set_tool( 'state' )

  def select_glue_state_tool( self ):
    self.parent.set_tool( 'glue_state' )

  def select_exclusive_glue_state_tool( self ):
    self.parent.set_tool( 'exclusive_glue_state' )

  def select_transition_tool( self ):
    self.parent.set_tool( 'transition' )

  def select_delete_tool( self ):
    self.parent.set_tool( 'delete' )

  def update( self, item_type ):
    if item_type == "none":
      self.state_button.config( state=DISABLED )
      self.glue_state_button.config( state=DISABLED )
      self.exclusive_glue_state_button.config( state=DISABLED )
      self.transition_button.config( state=DISABLED )
      self.delete_button.config( state=DISABLED )
    elif item_type == "rule":
      self.state_button.config( state=NORMAL )
      self.glue_state_button.config( state=NORMAL )
      self.exclusive_glue_state_button.config( text="Exclusive glue state", state=NORMAL )
      self.transition_button.config( state=NORMAL )
      self.delete_button.config( state=NORMAL )
    elif item_type == "process":
      self.state_button.config( state=NORMAL )
      self.glue_state_button.config( state=DISABLED )
      self.exclusive_glue_state_button.config( text="Initial state", state=NORMAL )
      self.transition_button.config( state=NORMAL )
      self.delete_button.config( state=NORMAL )








class _ProcessCanvas( Canvas ):
  def __init__( self, containing_frame, edit_frame, name, **kargs ):
    Canvas.__init__( self, containing_frame, **kargs )
    self.edit_frame = edit_frame
    self.name = name
    self.mouse_pressed = False
    self.mouse_moved_while_pressed = False

    self.process = None

    self.items = {}
    self.item = None
    self.item_index = None
    self.item_type = "none"
    self.point = None
    
    self.show_points = False

    self.bind( "<ButtonPress-1>", self.mouse_button_pressed )
    self.bind( "<ButtonRelease-1>", self.mouse_button_released )

  def draw_state( self, state ):
    x = state[ "x" ]
    y = state[ "y" ]
    id = self.create_oval( x - 15, y - 15, x + 15, y + 15, width="2.0", fill="white", tags="dragable selectable state" )
    self.create_text( x, y, text=state[ "id" ] );
    self.items[ id ] = state[ "id" ]

  def draw_glue_state( self, glue_state ):
    x = glue_state[ "x" ]
    y = glue_state[ "y" ]
    id = self.create_oval( x - 15, y - 15, x + 15, y + 15, width="2.0", fill="black", tags="dragable selectable glue_state" )
    self.create_text( x, y, text=glue_state[ "id" ], fill="white" );
    self.items[ id ] = glue_state[ "id" ]

  def rotate_point_vector( self, x, y, dx, dy, cx, cy ):
    cangle = complex( dx, dy )
    if cangle != 0:
      cangle = cangle / abs( cangle )

    center = complex( cx, cy )
    
    c = complex( x, y )
    c = ( c - center ) * cangle + center

    return ( c.real, c.imag )

  def draw_exclusive_glue_state( self, exclusive_glue_state ):
    x = exclusive_glue_state[ "x" ]
    y = exclusive_glue_state[ "y" ]
    
    ( x1, y1 ) = ( x - 15, y - 15 )
    ( x2, y2 ) = ( x + 15, y - 15 )
    ( x3, y3 ) = ( x + 15, y + 15 )
    ( x4, y4 ) = ( x - 15, y + 15 )
    
    id = self.create_polygon(
      x1, y1, x2, y2, x3, y3, x4, y4, fill="black",
      width="2.0", outline="black", tags="dragable selectable exclusive_glue_state"
    )
    self.create_text( x, y, text=exclusive_glue_state[ "id" ], fill="white" );
    self.items[ id ] = exclusive_glue_state[ "id" ]

  def draw_arrowhead( self, xs, ys, xt, yt ):
    arrowhead_factor = self.edit_frame.calculate_factor( xs, ys, xt, yt, 15 )
    ( arrowhead_x, arrowhead_y ) = self.edit_frame.calculate_coordinates_on_line( xs, ys, xt, yt, arrowhead_factor )
    
    ( rx, ry ) = ( arrowhead_x + 14, arrowhead_y - 6 )
    ( lx, ly ) = ( arrowhead_x + 14, arrowhead_y + 6 )
    ( tx, ty ) = ( arrowhead_x, arrowhead_y )

    ( dx, dy ) = ( xs - arrowhead_x, ys - arrowhead_y )

    ( cx, cy ) = ( arrowhead_x, arrowhead_y )

    ( rx, ry ) = self.rotate_point_vector( rx, ry, dx, dy, cx, cy )
    ( lx, ly ) = self.rotate_point_vector( lx, ly, dx, dy, cx, cy )
    ( tx, ty ) = self.rotate_point_vector( tx, ty, dx, dy, cx, cy )

    self.create_polygon( rx, ry, lx, ly, tx, ty, fill="black", outline="black" )

  def draw_line_transition( self, transition, points ):
    xs = transition[ "source" ][ "x" ]
    ys = transition[ "source" ][ "y" ]
    xt = transition[ "target" ][ "x" ]
    yt = transition[ "target" ][ "y" ]

    c1x = transition[ "c1x" ]
    c1y = transition[ "c1y" ]
    c2x = transition[ "c2x" ]
    c2y = transition[ "c2y" ]
    
    lx = transition[ "lx" ]
    ly = transition[ "ly" ]

    self.draw_arrowhead( c2x, c2y, xt, yt )
    id1 = self.create_line( xs, ys, c1x, c1y, c2x, c2y, xt, yt, smooth=True, fill="black", width="2.0", tags="selectable transition" )
    if points == True:
      id2 = self.create_oval( c1x - 4, c1y - 4, c1x + 4, c1y + 4, width="1.0", fill="grey", outline="grey", tags="selectable c1" )
      id3 = self.create_oval( c2x - 4, c2y - 4, c2x + 4, c2y + 4, width="1.0", fill="grey", outline="grey", tags="selectable c2" )
    id4 = self.create_text( lx, ly, text=transition[ "label" ], anchor=SW, tags="selectable label" )
    self.items[ id1 ] = transition[ "id" ]
    if points == True:
      self.items[ id2 ] = transition[ "id" ]
      self.items[ id3 ] = transition[ "id" ]
    self.items[ id4 ] = transition[ "id" ]

  def draw_loop_transition( self, transition, points ):
    ( x, y ) = ( transition[ "source" ][ "x" ], transition[ "source" ][ "y" ] )

    ( c1x, c1y ) = ( transition[ "c1x" ], transition[ "c1y" ] )
    ( c2x, c2y ) = ( transition[ "c2x" ], transition[ "c2y" ] )

    ( lx, ly ) = ( transition[ "lx" ], transition[ "ly" ] )

    ( mx, my ) = self.edit_frame.calculate_coordinates_on_line( c1x, c1y, c2x, c2y, 1/2 )

    self.draw_arrowhead( c2x, c2y, x, y )
    id1 = self.create_line( x, y, c1x, c1y, mx, my, fill="black", width="2.0", smooth=True, tags="selectable transition" )
    id2 = self.create_line( x, y, c2x, c2y, mx, my, fill="black", width="2.0", smooth=True, tags="selectable transition" )
    if points == True:
      id3 = self.create_oval( c1x - 4, c1y - 4, c1x + 4, c1y + 4, width="1.0", fill="grey", outline="grey", tags="selectable c1" )
      id4 = self.create_oval( c2x - 4, c2y - 4, c2x + 4, c2y + 4, width="1.0", fill="grey", outline="grey", tags="selectable c2" )
    id5 = self.create_text( lx, ly, text=transition[ "label" ], anchor=SW, tags="selectable label" )
    self.items[ id1 ] = transition[ "id" ]
    self.items[ id2 ] = transition[ "id" ]
    if points == True:
      self.items[ id3 ] = transition[ "id" ]
      self.items[ id4 ] = transition[ "id" ]
    self.items[ id5 ] = transition[ "id" ]
      
  def draw_process( self ):
    self.delete( ALL )
    if self.process != None:
      if self.show_points == True:
        points = True
      else:
        points = False
      self.items = {}
      self.delete( ALL )
      loop_transitions = [ self.process.transitions[ index ] for index in self.process.transitions if self.process.transitions[ index ][ "source" ] == self.process.transitions[ index ][ "target" ] ]
      line_transitions = [ self.process.transitions[ index ] for index in self.process.transitions if self.process.transitions[ index ][ "source" ] != self.process.transitions[ index ][ "target" ] ]
      for transition in line_transitions:
        self.draw_line_transition( transition, points )
      for transition in loop_transitions:
        self.draw_loop_transition( transition, points )
      for exclusive_glue_state in self.process.exclusive_glue_states:
        self.draw_exclusive_glue_state( self.process.exclusive_glue_states[ exclusive_glue_state ] )
      for glue_state in self.process.glue_states:
        self.draw_glue_state( self.process.glue_states[ glue_state ] )
      for state in self.process.states:
        self.draw_state( self.process.states[ state ] )

  def mouse_button_pressed( self, event ):
    self.mouse_pressed = True
    self.select_item( event )
    self.bind( "<B1-Motion>", self.mouse_motion )

  def select_item( self, event ):
    ( x, y ) = ( event.x, event.y )
    widget = event.widget
    
    items = widget.find_overlapping( x - 1, y - 1, x + 1, y + 1 )
    items = [ item for item in items if 'selectable' in widget.gettags( item ) ]

    states = [ item for item in items if 'state' in widget.gettags( item ) ]
    glue_states = [ item for item in items if 'glue_state' in widget.gettags( item ) ]
    exclusive_glue_states = [ item for item in items if 'exclusive_glue_state' in widget.gettags( item ) ]
    transitions = [ item for item in items if 'transition' in widget.gettags( item ) ]
    c1s = [ item for item in items if 'c1' in widget.gettags( item ) ]
    c2s = [ item for item in items if 'c2' in widget.gettags( item ) ]
    labels = [ item for item in items if 'label' in widget.gettags( item ) ]
    if len( states ) > 0:
      self.item_index = self.items[ states[ 0 ] ]
      self.item = self.process.states[ self.item_index ]
      self.item_type = 'state'
      self.point = None
    elif len( glue_states ) > 0:
      self.item_index = self.items[ glue_states[ 0 ] ]
      self.item = self.process.glue_states[ self.item_index ]
      self.item_type = 'glue_state'
      self.point = None
    elif len( exclusive_glue_states ) > 0:
      self.item_index = self.items[ exclusive_glue_states[ 0 ] ]
      self.item = self.process.exclusive_glue_states[ self.item_index ]
      self.item_type = 'exclusive_glue_state'
      self.point = None
    elif len( c1s ) > 0:
      self.item_index = self.items[ c1s[ 0 ] ]
      self.item = self.process.transitions[ self.item_index ]
      self.item_type = 'transition'
      self.point = 'c1'
    elif len( c2s ) > 0:
      self.item_index = self.items[ c2s[ 0 ] ]
      self.item = self.process.transitions[ self.item_index ]
      self.item_type = 'transition'
      self.point = 'c2'
    elif len( labels ) > 0:
      self.item_index = self.items[ labels[ 0 ] ]
      self.item = self.process.transitions[ self.item_index ]
      self.item_type = 'transition'
      self.point = 'label'
    elif len( transitions ) > 0:
      self.item_index = self.items[ transitions[ 0 ] ]
      self.item = self.process.transitions[ self.item_index ]
      self.item_type = 'transition'
      self.point = None
    else:
      self.item = None
      self.item_index = None
      self.item_type = "none"
      self.point = None
        
  def mouse_button_released( self, event ):
    self.mouse_pressed = False
    self.unbind( "<B1-Motion>" )
    if self.mouse_moved_while_pressed == False:
      self.edit_frame.perform_action( event.x, event.y, self.item_index, self.item_type, self.item, self.name )
    self.mouse_moved_while_pressed = False

  def mouse_motion( self, event ):
    if self.mouse_pressed == True:
      self.mouse_moved_while_pressed = True
    self.move_selected_item( event )

  def move_selected_item( self, event ):
    ( x, y ) = ( event.x, event.y )
    process = self.process

    if self.item_type == "state":
      state = self.process.states[ self.item_index ]
      ( dx, dy ) = ( x - state[ "x" ], y - state[ "y" ] )
      ( state[ "x" ], state[ "y" ] ) = ( x, y )
      self.move_related_transitions( self.item_index, dx, dy )
    elif self.item_type == "glue_state":
      state = self.process.glue_states[ self.item_index ]
      ( dx, dy ) = ( x - state[ "x" ], y - state[ "y" ] )
      ( state[ "x" ], state[ "y" ] ) = ( x, y )
      self.move_related_transitions( self.item_index, dx, dy )
    elif self.item_type == "exclusive_glue_state":
      state = self.process.exclusive_glue_states[ self.item_index ]
      ( dx, dy ) = ( x - state[ "x" ], y - state[ "y" ] )
      ( state[ "x" ], state[ "y" ] ) = ( x, y )
      self.move_related_transitions( self.item_index, dx, dy )
    elif self.item_type == "transition":
      transition = self.process.transitions[ self.item_index ]
      if self.point == "c1":
        ( osx, osy ) = ( transition[ "c1x" ], transition[ "c1y" ] )
        ( nsx, nsy ) = ( x, y )
        ( cx, cy ) = ( transition[ "c2x" ], transition[ "c2y" ] )
        ( lx, ly ) = ( transition[ "lx" ], transition[ "ly" ] )

        a = self.angle_between_vectors( nsx - cx, nsy - cy, osx - cx, osy - cy  )
        
        ( transition[ "c1x" ], transition[ "c1y" ] ) = ( x, y )
        ( transition[ "lx" ], transition[ "ly" ] ) = self.rotate_point_angle( lx, ly, a, cx, cy )
      elif self.point == "c2":
        ( osx, osy ) = ( transition[ "c2x" ], transition[ "c2y" ] )
        ( nsx, nsy ) = ( x, y )
        ( cx, cy ) = ( transition[ "c1x" ], transition[ "c1y" ] )
        ( lx, ly ) = ( transition[ "lx" ], transition[ "ly" ] )

        a = self.angle_between_vectors( nsx - cx, nsy - cy, osx - cx, osy - cy )
        
        ( transition[ "c2x" ], transition[ "c2y" ] ) = ( x, y )
        ( transition[ "lx" ], transition[ "ly" ] ) = self.rotate_point_angle( lx, ly, a, cx, cy )
      elif self.point == "label":
        ( transition[ "lx" ], transition[ "ly" ] ) = ( x, y )
    self.draw_process()

  def rotate_point_angle( self, x, y, a, cx, cy ):
    cangle = cmath.exp( a * 1j )

    center = complex( cx, cy )
    
    c = complex( x, y )
    c = ( c - center ) * cangle + center

    return ( c.real, c.imag )
    
  def angle_between_vectors( self, x1, y1, x2, y2 ):
      return math.atan2( y1, x1 ) - math.atan2( y2, x2 )

  def move_related_transitions( self, state_index, dx, dy ):
    loops = self.process.get_related_loop_transitions( state_index )
    for transition_index in loops:
      transition = self.process.transitions[ transition_index ]
      transition[ "lx" ] += dx
      transition[ "ly" ] += dy
      transition[ "c1x" ] += dx
      transition[ "c1y" ] += dy
      transition[ "c2x" ] += dx
      transition[ "c2y" ] += dy
    lines = self.process.get_related_line_transitions( state_index )
    for transition_index in lines:
      transition = self.process.transitions[ transition_index ]
      if transition[ "source" ][ "id" ] == state_index:
        state = transition[ "source" ]
        center = transition[ "target" ]
      else:
        state = transition[ "target" ]
        center = transition[ "source" ]
      ( sx_n, sy_n ) = ( state[ "x" ], state[ "y" ] )
      ( sx_o, sy_o ) = ( sx_n - dx, sy_n - dy )
      ( cx, cy ) = ( center[ "x" ], center[ "y" ] )
      ( dx_n, dy_n ) = ( sx_n - cx, sy_n - cy )
      ( dx_o, dy_o ) = ( sx_o - cx, sy_o - cy )
      if ( dx_o != 0.0 and dy_o != 0.0 and dx_o != -0.0 and dy_o != -0.0 ):
        ( sx, sy ) = ( float( dx_n ) / dx_o, float( dy_n ) / dy_o )
      else:
        ( sx, sy ) = ( 1, 1 )
      
      print( "sx " + str( sx ) + "sy " + str( sy ) )
      
      ( ldx_o, ldy_o ) = ( transition[ "lx" ] - cx, transition[ "ly" ] - cy )
      ( c1dx_o, c1dy_o ) = ( transition[ "c1x" ] - cx, transition[ "c1y" ] - cy )
      ( c2dx_o, c2dy_o ) = ( transition[ "c2x" ] - cx, transition[ "c2y" ] - cy )
      
      if (
        ldx_o != 0.0 and ldx_o != -0.0 and ldy_o != 0.0 and ldy_o != -0.0 and
        c1dx_o != 0.0 and c1dx_o != -0.0 and c1dy_o != 0.0 and c1dy_o != -0.0 and
        c2dx_o != 0.0 and c2dx_o != -0.0 and c2dy_o != 0.0 and c2dy_o != -0.0
      ):
        ( lx_n, ly_n ) = ( cx + ( ldx_o * sx ), cy + ( ldy_o * sy ) )
        ( c1x_n, c1y_n ) = ( cx + ( c1dx_o * sx ), cy + ( c1dy_o * sy ) )
        ( c2x_n, c2y_n ) = ( cx + ( c2dx_o * sx ), cy + ( c2dy_o * sy ) )
      
      #a = self.angle_between_vectors( osx - cx, osy - cy, nsx - cx, nsy - cy )
      #( transition[ "lx" ], transition[ "ly" ] ) = self.rotate_point_angle( transition[ "lx" ], transition[ "ly" ], a, cx, cy )
      #( transition[ "c1x" ], transition[ "c1y" ] ) = self.rotate_point_angle( transition[ "c1x" ], transition[ "c1y" ], a, cx, cy )
      #( transition[ "c2x" ], transition[ "c2y" ] ) = self.rotate_point_angle( transition[ "c2x" ], transition[ "c2y" ], a, cx, cy )
      
        ( transition[ "lx" ], transition[ "ly" ] ) = ( lx_n, ly_n )
        ( transition[ "c1x" ], transition[ "c1y" ] ) = ( c1x_n, c1y_n )
        ( transition[ "c2x" ], transition[ "c2y" ] ) = ( c2x_n, c2y_n )
      
        print( "lx " + str( lx_n ) + "ly " + str( ly_n ) )


    
class RuleAndProcessEditFrame( Frame ):
  def __init__( self, parent ):
    Frame.__init__( self, parent )
    self.parent = parent
    
    self.tool = None
    self.source_state = None
    self.source_state_index = None
    self.source_state_x = 0
    self.source_state_y = 0

    left_canvas_frame = Frame( self, relief=RAISED, borderwidth=1 )
    button_frame = Frame( self, borderwidth=1 )
    right_canvas_frame = Frame( self, relief=RAISED, borderwidth=1 )

    self.left_canvas = _ProcessCanvas( left_canvas_frame, self, "left", width=400, height=400, state=DISABLED, bg="light grey" )
    self.right_canvas = _ProcessCanvas( right_canvas_frame, self, "right", width=400, height=400, state=DISABLED, bg="light grey" )

    self.buttons = _Buttons( button_frame, self )
    self.left_canvas.pack()
    self.right_canvas.pack()
    left_canvas_frame.pack( side=LEFT )
    button_frame.pack( side=LEFT, padx=10 )
    right_canvas_frame.pack( side=LEFT )

  def set_tool( self, tool_type ):
    self.tool = tool_type
    if self.tool == "transition":
      self.left_canvas.show_points = True
      self.right_canvas.show_points = True
    else:
      self.left_canvas.show_points = False
      self.right_canvas.show_points = False
    self.left_canvas.draw_process()
    self.right_canvas.draw_process()
    
  def update( self, selected_item, item_type ):
    self.buttons.update( item_type )
    if item_type == "none":
      self.left_canvas.process = None
      self.right_canvas.process = None
      self.left_canvas.config( state=DISABLED, bg="light grey" )
      self.right_canvas.config( state=DISABLED, bg="light grey" )
    elif item_type == "process":
      self.left_canvas.process = selected_item
      self.right_canvas.process = None
      self.left_canvas.config( state=NORMAL, bg="white" )
      self.right_canvas.config( state=DISABLED, bg="light grey" )
    elif item_type == "rule":
      self.left_canvas.process = selected_item.left
      self.right_canvas.process = selected_item.right
      self.left_canvas.draw_process()
      self.right_canvas.draw_process()
      self.left_canvas.config( state=NORMAL, bg="white" )
      self.right_canvas.config( state=NORMAL, bg="white" )
    self.left_canvas.draw_process()
    self.right_canvas.draw_process()

  def calculate_coordinates_on_line( self, xs, ys, xt, yt, f ):
    x = xt + ( ( xs - xt ) * f )
    y = yt + ( ( ys - yt ) * f )
    return ( x, y )

  def calculate_factor( self, x1, y1, x2, y2, r ):
    div = x2**2 - ( 2 * x1 * x2 ) + x1**2 + y2**2 - ( 2 * y1 * y2 ) + y1**2
    if div == 0:
      factor = 1
    else:
      factor = math.sqrt( r**2 / div )
    return factor

  def calculate_label_coordinates( self, xs, ys, xt, yt ):
    middle_x = xs + ( ( xt - xs ) / 2 )
    middle_y = ys + ( ( yt - ys ) / 2 )
    d_x = xt - xs
    d_y = yt - ys
    if d_y > 0:
      label_factor = self.calculate_factor( middle_x + d_y, middle_y - d_x, middle_x, middle_y, 10 )
      ( label_x, label_y ) = self.calculate_coordinates_on_line( middle_x + d_y, middle_y - d_x, middle_x, middle_y, label_factor )
    else:
      label_factor = self.calculate_factor( middle_x - d_y, middle_y + d_x, middle_x, middle_y, 10 )
      ( label_x, label_y ) = self.calculate_coordinates_on_line( middle_x - d_y, middle_y + d_x, middle_x, middle_y, label_factor )
    return ( label_x, label_y )
    
  def calculate_control_points( self, source, target, xs, ys, xt, yt ):
    if source == target:
      f = 40
      ( c1x, c1y ) = ( xs + f, ys - f )
      ( c2x, c2y ) = ( xs + f, ys + f )
    else:
      ( c1x, c1y ) = self.calculate_coordinates_on_line( xs, ys, xt, yt, 2/3 )
      ( c2x, c2y ) = self.calculate_coordinates_on_line( xs, ys, xt, yt, 1/3 )
    return ( c1x, c1y, c2x, c2y )

  def perform_action( self, x, y, item_index, item_type, item, name ):
    if self.tool == "state":
      self.source_state = None
      self.parent.add_state( name, x, y )
    elif self.tool == "glue_state":
      self.source_state = None
      self.parent.add_glue_state( name, x, y )
    elif self.tool == "exclusive_glue_state":
      self.source_state = None
      self.parent.add_exclusive_glue_state( name, x, y )
    elif self.tool == "delete":
      self.source_state = None
      self.parent.delete( name, item_index, item_type )
    elif self.tool == "transition":
      if item_type == "state" or item_type == "glue_state" or item_type == "exclusive_glue_state":
        if self.source_state == None:
          self.source_state = item
          self.source_state_index = item_index
          self.source_state_x = item[ "x" ]
          self.source_state_y = item[ "y" ]
        else:
          if item != None:
            label = simpledialog.askstring( "Transition label", "Please enter a label for this transition:", parent=self )
            if label != None:
              ( c1x, c1y, c2x, c2y ) = self.calculate_control_points( self.source_state_index, item_index, self.source_state_x, self.source_state_y, item[ "x" ], item[ "y" ] )
              ( lx, ly ) = self.calculate_label_coordinates( c1x, c1y, c2x, c2y )
              self.parent.add_transition( name, self.source_state_index, item_index, label, c1x, c1y, c2x, c2y, lx, ly )
            self.source_state = None
