from tkinter import *
import tkinter.filedialog as filedialog
import config

class _Buttons:
  def __init__( self, frame, parent ):
    self.select_label = Label( frame, text="Rule selection", font=parent.customfont )
    self.show_and_edit_button = Button( frame, text="Show and edit", command=parent.show_and_edit_item, state=DISABLED, font=parent.customfont )
    self.clone_button = Button( frame, text="Clone", command=parent.clone_item, state=DISABLED, font=parent.customfont )
    self.delete_button = Button( frame, text="Delete", command=parent.delete_item, state=DISABLED, font=parent.customfont )
    self.add_rule_button = Button( frame, text="New Process", command=parent.add_process, font=parent.customfont )
    self.add_process_button = Button( frame, text="New Transformation", command=parent.add_rule, font=parent.customfont )
    self.select_label.pack( fill=X )
    self.show_and_edit_button.pack( fill=X )
    self.clone_button.pack( fill=X )
    self.delete_button.pack( fill=X )
    self.add_process_button.pack( fill=X )
    self.add_rule_button.pack( fill=X )
  
  def update( self, no_rules_and_processes ):
    if no_rules_and_processes == True:
      self.show_and_edit_button.config( state=DISABLED )
      self.clone_button.config( state=DISABLED )
      self.delete_button.config( state=DISABLED )
    else:
      self.show_and_edit_button.config( state=NORMAL )
      self.clone_button.config( state=NORMAL )
      self.delete_button.config( state=NORMAL )

class RuleAndProcessSelectionFrame( Frame ):
  def __init__( self, frame, parent, customfont ):
    Frame.__init__( self, frame )
    self.parent = parent
    self.customfont = customfont

    self.next_id = 0

    list_box_frame = Frame( self )
    button_frame = Frame( self, highlightthickness=0 )
    scrollbar = Scrollbar( list_box_frame )
    self.list_box = Listbox( list_box_frame, yscrollcommand=scrollbar.set, background="white", font=self.customfont )
    scrollbar.config( command=self.list_box.yview )
    self.buttons = _Buttons( button_frame, self )

    hack_label = Label( self, text="" )

    scrollbar.pack( side=RIGHT, fill=Y )
    self.list_box.pack( side=LEFT, fill=BOTH )
    
    hack_label.grid( row=0, column=0, columnspan=2 )
    list_box_frame.grid( row=1, column=0, padx=5, pady=5 )
    button_frame.grid( row=1, column=1, padx=5, pady=5, sticky=N )

    self.list_box.bind( "<Double-Button-1>", self.show_and_edit_item )

  def list_box_is_empty( self ):
    return self.list_box.size() == 0

  def item_is_selected( self ):
    if self.list_box_is_empty() == True:
      return False
    else:
      return len( self.list_box.curselection() ) > 0

  def update( self, no_rules_and_processes ):
    self.list_box.delete( 0, END )
    for name in sorted(config.rules_and_processes.keys()):
      c = config.rules_and_processes[name]
      self.list_box.insert( END, name )
    self.buttons.update( no_rules_and_processes )
    
  def show_and_edit_item( self, event=None ):
    if self.item_is_selected():
      selected_item_index = int( self.list_box.curselection()[ 0 ] )
      string = self.list_box.get( selected_item_index )[0:7]
      selected_item = self.list_box.get( selected_item_index )
      if string == 'Process':
        item_type = "process"
      else:
        item_type = "rule"
      self.parent.show_item( selected_item, item_type )

  def clone_item( self ):
    if self.item_is_selected():
      selected_item_index = int( self.list_box.curselection()[ 0 ] )
      string = self.list_box.get( selected_item_index )[0:7]
      selected_item = self.list_box.get( selected_item_index )
      if string == 'Process':
        new_item = "Process " + str( self.next_id )
      else:
        new_item = "Transformation " + str( self.next_id )
      self.list_box.insert( END, new_item )
      self.parent.clone_item( new_item, selected_item )
      self.next_id += 1

  def delete_item( self ):
    if self.item_is_selected():
      selected_item_index = int( self.list_box.curselection()[ 0 ] )
      selected_item = self.list_box.get( selected_item_index )
      self.list_box.delete( selected_item_index )
      self.parent.delete_item( selected_item )

  def add_rule( self ):
    new_item = "Transformation " + str( self.next_id )
    self.parent.add_rule( new_item )
    self.list_box.insert( END, new_item )
    self.list_box.selection_clear( 0, END )
    self.list_box.selection_set( END )
    self.parent.show_item( new_item, "rule" )
    self.next_id += 1

  def add_process( self ):
    new_item = "Process " + str( self.next_id )
    self.parent.add_process( new_item )
    self.list_box.insert( END, new_item )
    self.list_box.selection_clear( 0, END )
    self.list_box.selection_set( END )
    self.parent.show_item( new_item, "process" )
    self.next_id += 1

  def set_nextid( self, id ):
    self.next_id = id
