from tkinter import *
from refiner import *
import config

class _Buttons:
  def __init__( self, frame, parent ):
    self.parent = parent

    self.verify_button = Button( frame, text="Verify", command=self.verify, state=DISABLED, font=parent.customfont )
    self.transform_button = Button( frame, text="Transform", command=self.transform, state=DISABLED, font=parent.customfont )
    self.divergency_check_button = Checkbutton( frame, text="Divergency", variable=parent.divergency, command=self.switchdivergency, onvalue=True, offvalue=False, state=DISABLED, font=parent.customfont )
    self.fairness_check_button = Checkbutton( frame, text="Fairness", variable=parent.fairness, command=self.switchfairness, onvalue=True, offvalue=False, state=DISABLED, font=parent.customfont )
    
    self.transform_button.pack( fill=X )
    self.verify_button.pack( fill=X )
    self.divergency_check_button.pack( anchor=W, ipadx=5 )
    self.fairness_check_button.pack( anchor=W, ipadx=5 )
    
  def switchdivergency( self ):
    if self.fairness_check_button['state'] == NORMAL:
  	  self.fairness_check_button.config( state=DISABLED )
    else:
      self.fairness_check_button.config( state=NORMAL )

  def switchfairness( self ):
    if self.parent.fairness.get():
      self.parent.divergency.set(False)
      self.divergency_check_button.config( state=DISABLED )
    elif config.network_name != "":
      self.divergency_check_button.config( state=NORMAL )

  def update( self, model, rule_system, property ):
    if rule_system:
      self.verify_button.config( state=NORMAL )
      if model != "":
        self.transform_button.config( state=NORMAL )
        if property != "":
          self.divergency_check_button.config( state=NORMAL )
          self.fairness_check_button.config( state=NORMAL )
        else:
          self.divergency_check_button.config( state=DISABLED )
          self.fairness_check_button.config( state=DISABLED )
      else:
        self.transform_button.config( state=DISABLED )
        self.divergency_check_button.config( state=DISABLED )
        if property != "":
          self.fairness_check_button.config( state=NORMAL )
        else:
          self.fairness_check_button.config( state=DISABLED )

  def verify( self ):
    self.parent.verify()

  def transform( self ):
    self.parent.transform()

class VerifyAndTransformFrame( Frame ):
  def __init__( self, frame, parent, customfont ):
    Frame.__init__( self, frame )
    self.parent = parent
    self.customfont = customfont
    
    self.divergency = BooleanVar()
    self.fairness = BooleanVar()
    self.button_frame = Frame( self )

    self.buttons = _Buttons( self.button_frame, self )
    self.button_frame.pack()
    
  def update( self, model, rule_system, property ):
    self.buttons.update( model, rule_system, property )

  def verify( self ):
    self.parent.verify( (self.divergency.get() == 1), (self.fairness.get() == 1) )

  def transform( self ):
    self.parent.transform( (self.divergency.get() == 1) )
